#include "Standard.h"

DEFINE_ALLOCATOR(ScheduledValue<MessageChannel *>);
DEFINE_ALLOCATOR(ScheduledValue<Component *>);

bool
Domain	::	BindParameters(
	Domain	*This,
	TupleSpace	*theTupleSpace,
	char		*aName)
	{

	theTupleSpace->StartSubSpace(DefaultPass(aName, "ClockingDomain"));
	SetComponentName(This, theTupleSpace);
	theTupleSpace->Bind(Parameter(This, TotalEffect), (bool)false);
	theTupleSpace->EndSubSpace();
	return(true);
	}

int
Domain	::	AsyncSimulateOneCycle (  )
	{
	int	Index;
	MessageChannel	*theChannel;
	Component		*theComponent;
	bool			Change = true;

	if (  ! CurrentCycle )
		{
		// call simulate for sync components
		Index = 0;
		while ( Index < SyncSimulateComponents.GetSize ( ) )
			{
			if ( SyncSimulateComponents[Index]->SyncSimulateOneCycle() < 0 )
				SyncSimulateComponents.RemoveAt(Index);
			else
				++Index;
			}

		// call simulate for all components
		Index = 0;
		while ( Index < SimulateComponents.GetSize ( ) )
			{
			if ( SimulateComponents[Index]->AsyncSimulateOneCycle() < 0 )
				SimulateComponents.RemoveAt(Index);
			else
				++Index;
			}
		}
	else
		{
		for ( Index = 0; Index < SyncSimulateComponents.GetSize(); Index++)
			SyncSimulateComponents[Index]->SyncSimulateOneCycle();
		for ( Index = 0; Index < SimulateComponents.GetSize(); Index++)
			SimulateComponents[Index]->AsyncSimulateOneCycle();
		}

	if (! TotalEffect )
		{
		while ( LiveChannels.Peek ( CurrentCycle ) )
			{
			theChannel = LiveChannels.Pop ( CurrentCycle );
			theChannel->Deliver();
			}
		}
	else
		{
		while ( Change )
			{
			Change = false;
			while ( LiveChannels.Peek ( CurrentCycle ) )
				{
				Change = true;
				theChannel = LiveChannels.Pop ( CurrentCycle );
				theChannel->Deliver();
				}

			while ( LiveComponents.Peek ( CurrentCycle ) )
				{
				Change = true;
				theComponent = LiveComponents.Pop ( CurrentCycle );
				theComponent->AsyncSimulateOneCycle();
				}
			}
		}


	CurrentCycle++;
	StatCurrentCycle++;
	return(1);
	}

void
Domain	::	AddComponent ( Component * aChild )
	{	
	if ( aChild == this )	return;
	SimulateComponents.Add(aChild);
	SyncSimulateComponents.Add(aChild);
	if(aChild->IsMessageChannel())
		AddChannel((MessageChannel*)aChild);
	else	
		Components.Add ( aChild );
	}

void
Domain	::	AddChannel ( MessageChannel * aChannel )
	{
	Channels.Add ( aChannel );
	}

bool
Domain	::	ResetStatistics()
	{
	StatCurrentCycle = 0;
	int	Index = 0;
	while ( Index < Components.GetSize() )
		{
		Components[Index]->ResetStatistics();
		++Index;
		}
	Index = 0;
	while ( Index < Channels.GetSize() )
		{
		Channels[Index]->ResetStatistics();
		++Index;
		}
	return(true);
	}

