
// it is expected that a DEFINE_ALLOCATOR(ScheduledValue<T>) will exist for each SchedluedQueue class type
template<class	T>
class	ScheduledValue
	{
	public:

		T		theValue;
		uint64	theCycle;

		void	Set(T	aValue, uint64	aCycle)	{ theValue = aValue; theCycle = aCycle; }

		inline	bool	operator	<	(ScheduledValue<T>	&v)	{ return theCycle < v.theCycle; }
		inline	bool	operator	>	(ScheduledValue<T>	&v)	{ return theCycle > v.theCycle; }
		inline	bool	operator	==	(ScheduledValue<T>	&v)	{ return theCycle == v.theCycle; }
	};
     

template < class T > 
class	ScheduledQueue
	{
	public:
		Array <T>		*NearFuture;
		int				NearFutureLength;
		int				PresentLocation;
		uint64			CurrentCycle;
		int				PQAccesses;
		int				FastAccesses;
		int				Size;

		PQueueOfPtr	<ScheduledValue<T>*>	PastAndFuture;

		ScheduledQueue (uint64	aCurrentCycle = 0, int aNearFutureLength = 16)
			{
			CurrentCycle = aCurrentCycle;
			NearFutureLength = aNearFutureLength;
			NearFuture = new Array<T>[NearFutureLength];
			PresentLocation = 0;
			FastAccesses = 0;
			PQAccesses = 0;
			Size = 0;
			}

		inline	int	GetSize() { return Size; }


		inline	void	Push(T & Value, uint64 ReadyCycle)
			{
			int	Delta = ReadyCycle - CurrentCycle;
			if (Delta < 0 || Delta > NearFutureLength)
				{
				ScheduledValue<T>	*sv = NEW(ScheduledValue<T>);
				sv->Set(Value, ReadyCycle);
				PastAndFuture.Push(sv);
				}
			else
				{
				NearFuture[(PresentLocation + Delta) % NearFutureLength].Add(Value);
				}
			Size++;
			}

		inline	void	DemoteCycle(uint64	aCycle, Array<T> *aArray)
			{
			int	length = aArray->GetSize();
			for(int j = 0; j < length; j++)
				{
				ScheduledValue<T>	*sv = NEW(ScheduledValue<T>);
				sv->Set((*aArray)[j],aCycle);
				PastAndFuture.Push(sv);
				}
			aArray->Clear();
			}
     
		inline	void	AdvancePresent(uint64	aCycle)
			{
			int Delta = aCycle - CurrentCycle;

			if (Delta < 0 || Delta > NearFutureLength)
				{
				// If we move to far, or backwards we ditch the near future and
				// dump it all into the priority queue.  This is a bad thing from a
				// performance perspective.

				for (int i = 0; i < NearFutureLength; i++)
					{
					Array<T> *c = &NearFuture[(PresentLocation + i) % NearFutureLength];
					DemoteCycle(CurrentCycle + i, c);
					}
				CurrentCycle = aCycle;
				PresentLocation = 0;
				}
			 else
				{
				// near future
	 
				for (int i = 0; i < Delta; i++)
					{
					Array<T> *c = &NearFuture[PresentLocation];
					DemoteCycle(CurrentCycle, c);
					CurrentCycle++;
					PresentLocation = (PresentLocation + 1) % NearFutureLength;
					}
				}
			}


		inline	T	Pop(uint64	atCycle)
			{
			if (atCycle != CurrentCycle)
				{
				AdvancePresent(atCycle);
				}

			if (PastAndFuture.GetSize() != 0 && PastAndFuture.Peek()->theCycle <= atCycle)
				{
				ScheduledValue<T>	*v = PastAndFuture.Pop();
				T	r = v->theValue;
				PQAccesses++;
				DELETE(ScheduledValue<T>, v);
				Size--;
				return r;
				}
			else
				{
				T	r;
				FastAccesses++;
				if (NearFuture[PresentLocation].GetSize() != 0)
					{
					r = NearFuture[PresentLocation][0];
					NearFuture[PresentLocation].RemoveAt(0);
					Size--;
				return r;
					}
				else
					{
					return((T)0);
					}
				}
			}
     
		inline	T	Peek(uint64	atCycle)
			{
			if (atCycle != CurrentCycle)
				AdvancePresent(atCycle);
	  
			if (PastAndFuture.GetSize() != 0 && PastAndFuture.Peek()->theCycle <= atCycle)
				{
				ScheduledValue<T> *v = PastAndFuture.Peek();
				T	r = v->theValue;
				PQAccesses++;
				return r;
				}
			else
				{
				T	r;
				FastAccesses++;
				if (NearFuture[PresentLocation].GetSize() != 0)
					{
					r = NearFuture[PresentLocation][0];
					return r;
					}
				else
					return((T)0);
				}
			}
     
	};
