#	include	"Standard.h"



bool
SimpleCache	::	BindParameters(SimpleCache	*This, TupleSpace	*theTupleSpace,	char	*aName )
	{
	int	Factor = 1;

	if(aName)
	if ( !strcmp(aName, "L2Cache"))	Factor = 4;

	theTupleSpace->StartSubSpace(DefaultPass(aName, "Cache"));
	SetComponentName(This, theTupleSpace);
	theTupleSpace->Bind(Parameter(This, Latency), 1*Factor);
	theTupleSpace->Bind(Parameter(This, MaxOutstandingRequests), 4);
	theTupleSpace->Bind(Parameter(This, LineSize), 32);
	theTupleSpace->Bind(Parameter(This, BankSize), 32*Factor);
	theTupleSpace->Bind(Parameter(This, Associativity), 4);
	theTupleSpace->Bind(Parameter(This, ConcurrentRequests), 4);
	theTupleSpace->Bind(Parameter(This, WriteThrough), (bool)false);
	theTupleSpace->EndSubSpace();
	return(true);
	}

bool
SimpleCache	::	BindStructures(
			InputPort	*aAccessInputPort,
			OutputPort	*aAccessOutputPort,
			InputPort	*aFillInputPort,
			OutputPort	*aFillOutputPort)
	{
	RequestsInFlight = 0;
	FillRequestId = 0;
	theAccessOutputPort = aAccessOutputPort;
	theAccessInputPort = aAccessInputPort;
	theAccessInputPort->Connect(this);

	theFillOutputPort = aFillOutputPort;
	theFillInputPort = aFillInputPort;
	theFillInputPort->Connect(this);

	Overflow.Resize(MaxOutstandingRequests*4);
	ToFill.Resize(MaxOutstandingRequests*4);
	ToAccess.Resize(MaxOutstandingRequests*4);

	theCache = SetAssociativeMemory<int>::Create(LineSize, BankSize, Associativity, LRU);

	ResetStatistics();
	return(true);
	}

bool
SimpleCache	::ResetStatistics()
	{
	Hits = Misses = WriteBacks = 0;
	return(true);
	}

int
SimpleCache	::	WhenCanYouTakeThisMessage(
	Message	*m,
	Port	*p)
	{
	MemoryTransactionMessage	*mtm = (MemoryTransactionMessage*) m;

	if ( (int)mtm->Id3 == GetId() )	return(0);	// result messages we always can take

	if ( p != theAccessInputPort )	return(0);	// message will be dropped (correctly)
	if( (mtm->Command != Read) && (mtm->Command != Write) )	return(0);	// message will be dropped (correctly)

	if ( MaxOutstandingRequests )
	if ( RequestsInFlight >= MaxOutstandingRequests )
		{
		return(1);	// not really sure when, so just say next cycle
		}

	// mesage will be handled, take it
	return(0);
	}

bool
SimpleCache	::	ReceiveMessage (
	Message	*m,
	Port	*p)
	{
	MemoryTransactionMessage	*mtm = (MemoryTransactionMessage*) m;

#ifdef	ENABLE_MESSAGE_CHECK
	printf("%s: ReceiveMessage: ", GetName());mtm->Dump(stdout);
#endif

	if ( (int)mtm->Id3 == GetId() )
		{
		// we made this message.. this occurs (usually) when a cache miss happens
		// and we made a request for something.  That something has now returned to us
		// (possibly)
		if ( mtm->Command != ReadResponse )
			{
			// maybe not.. this is probably a Read or Write that we made to the upstream side
			// on a shared bus.
			DELETE(MemoryTransactionMessage, mtm);
			return(true);
			}
		// yes it did, what we do now is insert it into the cache...
		// then check our PendingMisses list to see if there is anyone we can now handle
		// (there should be...)
		uint64	Line1;
		Line1 = mtm->Address >> theCache->ShiftAmount;

		int	Index = PendingMisses.GetSize();
		while ( Index > 0 )
			{
			--Index;
			if ((PendingMisses[Index]->Address >> theCache->ShiftAmount) == Line1)
				{
				MemoryTransactionMessage	*mtm2 = PendingMisses[Index];
				if ( mtm2->Command == Read )
					{
					mtm2->Command = ReadResponse;
					ToAccess.Insert(mtm2);
					}
				else
					{
					// this was a write request
					--RequestsInFlight;
					if ( WriteThrough )
						{
						mtm2->Address = mtm->Address;
						mtm2->Width = LineSize;
						mtm2->Command = Read;
						mtm2->Id = FillRequestId++;
						mtm2->Id3 = GetId();
						++WriteBacks;
						++RequestsInFlight;
						ToFill.Insert(mtm2);
						}
					}
				PendingMisses.RemoveAt(Index);
				}
			}
		// place the new data into the cache
		uint		HitOrMiss;
		uint		ForcedWriteBack;
		Line<int>	theLine;
		theCache->Locate(mtm->Address, &HitOrMiss, &ForcedWriteBack, &theLine);
		DELETE(MemoryTransactionMessage, mtm);

		// if this caused a write back pend that to the outgoing queue
		if ( !WriteThrough )
		if ( ForcedWriteBack )
			{
			// ug.. need to send out an mtm
			MemoryTransactionMessage	*mtm2 = NEW(MemoryTransactionMessage);	
			mtm2->Address = theLine.RealAddress;
			mtm2->Width = LineSize;
			mtm2->Command = Write;
			mtm2->Id = FillRequestId++;
			mtm2->Id3 = GetId();
			++RequestsInFlight;
			ToFill.Insert(mtm2);
			++WriteBacks;
			}
		InDomain->ComponentIsLive(this, InDomain->GetCurrentCycle());
		return(true);
		}

	if ( p != theAccessInputPort )	// ignore requests on our back side
		{
		DELETE(MemoryTransactionMessage, mtm);
		return(true);
		}

	// we did not initiate this message.. we should try and handle it
	// if its appropriate
	if( (mtm->Command != Read) && (mtm->Command != Write) )
		{
		DELETE(MemoryTransactionMessage, mtm);
		return(true);
		}
	if ( MaxOutstandingRequests )
	if ( RequestsInFlight >= MaxOutstandingRequests )
		{
		return(false);
		}

	if ( ProcessedEvents.GetSize() >= ConcurrentRequests )
		{
		Overflow.Insert(mtm);
		}
	else
		{
		ProcessedEvents.Push(mtm, InDomain->GetCurrentCycle() + Latency);
		}
	++RequestsInFlight;
	InDomain->ComponentIsLive(this, InDomain->GetCurrentCycle());
	return(true);
	}

int
SimpleCache	::	AsyncSimulateOneCycle()
	{
	MemoryTransactionMessage	*mtm;

	while ( true )
		{
		mtm = ProcessedEvents.Peek(InDomain->GetCurrentCycle());
		if ( mtm )
			{
			ProcessedEvents.Pop(InDomain->GetCurrentCycle());
			if ( mtm->Command == Write )
				{

				// examine this mtm, would it be a cache it?
				if ( theCache->IsInCache(mtm->Address) )
					{
					// nifty,
					--RequestsInFlight;
					++Hits;
					uint	HitOrMiss, ForcedWriteBack;
					Line<int>	theLine;
					Line<int>	*Access = theCache->Locate(mtm->Address, &HitOrMiss, &ForcedWriteBack, &theLine);
					Access->Dirty = 1;
					DELETE(MemoryTransactionMessage, mtm);
					// we hit in the cache, are we writethrough?
					if ( WriteThrough )
						{
						MemoryTransactionMessage	*mtm2 = NEW(MemoryTransactionMessage);
						mtm2->Address = mtm->Address;
						mtm2->Width = mtm->Width;
						mtm2->Command = Write;
						mtm2->Id = FillRequestId++;
						mtm2->Id3 = GetId();
						++WriteBacks;
						++RequestsInFlight;
						ToFill.Insert(mtm2);
						}
					}
				else
					{
					// no pend this request and issue a read (i.e. write-allocate policy)
					++Misses;
					MemoryTransactionMessage	*mtm2 = NEW(MemoryTransactionMessage);
					mtm2->Address = mtm->Address >> theCache->ShiftAmount;
					mtm2->Address = mtm2->Address << theCache->ShiftAmount;
					mtm2->Width = LineSize;
					mtm2->Command = Read;
					mtm2->Id = FillRequestId++;
					mtm2->Id3 = GetId();
					bool	FoundMatchingRequest = false;
					int	Index = 0;
					while ( Index < PendingMisses.GetSize() )
						{
						if ((PendingMisses[Index]->Address >> theCache->ShiftAmount) == 
							(mtm->Address >> theCache->ShiftAmount) )
							{
							FoundMatchingRequest = true;
							break;
							}
						++Index;
						}
					if ( ! FoundMatchingRequest )
						{
						++RequestsInFlight;
						ToFill.Insert(mtm2);
						}
					else
						DELETE(MemoryTransactionMessage, mtm2);
					PendingMisses.Add(mtm);
					}

				continue;
				}
			else	
			if ( mtm->Command == Read )
				{

				if ( theCache->IsInCache(mtm->Address) )
					{
					// nifty,
					uint	HitOrMiss, ForcedWriteBack;
					Line<int>	theLine;
					theCache->Locate(mtm->Address, &HitOrMiss, &ForcedWriteBack, &theLine);
				
					mtm->Command = ReadResponse;
					++Hits;
					ToAccess.Insert(mtm);
					}
				else
					{
					// pend this request
					// no pend this request and issue a read

					++Misses;
					MemoryTransactionMessage	*mtm2 = NEW(MemoryTransactionMessage);
					mtm2->Address = mtm->Address >> theCache->ShiftAmount;
					mtm2->Address = mtm2->Address << theCache->ShiftAmount;
					mtm2->Width = LineSize;
					mtm2->Command = Read;
					mtm2->Id = FillRequestId++;
					mtm2->Id3 = GetId();
					bool	FoundMatchingRequest = false;
					int	Index = 0;
					while ( Index < PendingMisses.GetSize() )
						{
						if ((PendingMisses[Index]->Address >> theCache->ShiftAmount) == 
							(mtm->Address >> theCache->ShiftAmount) )
							{
							FoundMatchingRequest = true;
							break;
							}
						++Index;
						}
					if ( ! FoundMatchingRequest )
						{
						++RequestsInFlight;
						ToFill.Insert(mtm2);
						}
					else
						DELETE(MemoryTransactionMessage, mtm2);

					PendingMisses.Add(mtm);
					}
				}
			else
				FatalError("Hmmm...");
			continue;
			}

		if ( ProcessedEvents.GetSize() < ConcurrentRequests )
		if ( Overflow.GetSize() )
			{
			Overflow.Remove(&mtm);
			ProcessedEvents.Push(mtm, InDomain->GetCurrentCycle() + Latency);
			continue;
			}
		break;
		}

	// anything we have to send to the access port, ?
	while ( ToAccess.GetSize() )
		{
		ToAccess.Peek(&mtm);
		mtm->SetSourceComponentId(GetId());

		if ( ! theAccessOutputPort->SendMessage(mtm) )
			break;
#ifdef	ENABLE_MESSAGE_CHECK
	printf("%s: Sent - to Access: ", GetName());mtm->Dump(stdout);
#endif
		--RequestsInFlight;
		ToAccess.Remove(&mtm);
		}

	// anything we have to send to the fill port, do so now...
	while ( ToFill.GetSize() )
		{
		ToFill.Peek(&mtm);
		mtm->SetSourceComponentId(GetId());
		if ( ! theFillOutputPort->SendMessage(mtm) )
			break;
#ifdef	ENABLE_MESSAGE_CHECK
	printf("%s: Sent - to Fill: ", GetName());mtm->Dump(stdout);
#endif
		--RequestsInFlight;
		ToFill.Remove(&mtm);
		}

	return(1);
	}

bool
SimpleCache	::	PrintStatistics(
	OutputFormat	theFormat,
	FILE			*fp)
	{
	PrintValue(theFormat, fp, "Hits", Hits);
	PrintValue(theFormat, fp, "Misses", Misses);
	PrintPercent(theFormat, fp, "HitRatio", Hits, Misses);
	PrintValue(theFormat, fp, "WriteBacks", WriteBacks);
	return ( true );
	}	

