#	include	"Standard.h"


bool
SimpleCacheMemorySystem	::	BindParameters(SimpleCacheMemorySystem	*This, TupleSpace	*theTupleSpace, char	*aName )
	{
	theTupleSpace->StartSubSpace(DefaultPass(aName, "MemorySystem"));
	SetComponentName(This, theTupleSpace);
	SimpleClockedMemoryBus::BindParameters(StaticPass(This, L2ToSDRAMBus), theTupleSpace, "L2ToSDRAMBus");
	SimpleClockedMemoryBus::BindParameters(StaticPass(This, L1ToL2Bus), theTupleSpace, "L1ToL2Bus");
	SimpleClockedMemoryBus::BindParameters(StaticPass(This, L1InstructionToProcessorBus), theTupleSpace, "L1InstructionToProcessorBus");
	SimpleClockedMemoryBus::BindParameters(StaticPass(This, ProcessorToL1InstructionBus), theTupleSpace, "ProcessorToL1InstructionBus");
	SimpleClockedMemoryBus::BindParameters(StaticPass(This, L1DataToProcessorBus), theTupleSpace, "L1DataToProcessorBus");
	SimpleClockedMemoryBus::BindParameters(StaticPass(This, ProcessorToL1DataBus), theTupleSpace, "ProcessorToL1DataBus");
			
	SimpleCache::BindParameters(StaticPass(This, L2Cache), theTupleSpace, "L2Cache");
	SimpleCache::BindParameters(StaticPass(This, L1InstructionCache), theTupleSpace, "L1InstructionCache");
	SimpleCache::BindParameters(StaticPass(This, L1DataCache), theTupleSpace, "L1DataCache");
	SimpleSDRAMModel::BindParameters(StaticPass(This, SDRAMModel), theTupleSpace);
	theTupleSpace->EndSubSpace();
	return(true);
	}

bool
SimpleCacheMemorySystem	::	Create()
	{
	//////////////////////////////////////////////////////////////////////////////////////////
	// Allocate all of the objects
	L2ToSDRAMBus = new SimpleClockedMemoryBus;
	L2ToSDRAMInputPort = new InputPort;
	L2ToSDRAMOutputPort = new OutputPort;

	L1ToL2Bus = new SimpleClockedMemoryBus;
	L1ToL2InputPort = new InputPort;
	L1ToL2OutputPort = new OutputPort;

	L1InstructionToProcessorBus = new SimpleClockedMemoryBus;
	L1InstructionToProcessorInputPort = new InputPort;
	L1InstructionToProcessorOutputPort = new OutputPort;
	ProcessorToL1InstructionBus = new SimpleClockedMemoryBus;
	ProcessorToL1InstructionInputPort = new InputPort;
	ProcessorToL1InstructionOutputPort = new OutputPort;

	L1DataToProcessorBus = new SimpleClockedMemoryBus;
	L1DataToProcessorInputPort = new InputPort;
	L1DataToProcessorOutputPort = new OutputPort;
	ProcessorToL1DataBus = new SimpleClockedMemoryBus;
	ProcessorToL1DataInputPort = new InputPort;
	ProcessorToL1DataOutputPort = new OutputPort;

	SDRAMModel = new SimpleSDRAMModel;
	L2Cache = new SimpleCache;
	L1InstructionCache = new SimpleCache;
	L1DataCache = new SimpleCache;

	//////////////////////////////////////////////////////////////////////////////////////////
	// create them all
	L1ToL2Bus->Create();
	L1ToL2InputPort->Create();
	L1ToL2OutputPort->Create();
	L2ToSDRAMBus->Create();
	L2ToSDRAMInputPort->Create();
	L2ToSDRAMOutputPort->Create();

	L1InstructionToProcessorBus->Create();
	L1InstructionToProcessorInputPort->Create();
	L1InstructionToProcessorOutputPort->Create();
	ProcessorToL1InstructionBus->Create();
	ProcessorToL1InstructionInputPort->Create();
	ProcessorToL1InstructionOutputPort->Create();

	L1DataToProcessorBus->Create();
	L1DataToProcessorInputPort->Create();
	L1DataToProcessorOutputPort->Create();
	ProcessorToL1DataBus->Create();
	ProcessorToL1DataInputPort->Create();
	ProcessorToL1DataOutputPort->Create();

	L1InstructionCache->Create();
	L1DataCache->Create();	
	L2Cache->Create();

	SDRAMModel->Create();
		
					
/*	//////////////////////////////////////////////////////////////////////////////////////////
	// name the ports (useful for debugging)
	NameByPtr(L1InstructionToProcessorOutputPort);
	NameByPtr(L1InstructionToProcessorInputPort);
	NameByPtr(ProcessorToL1InstructionOutputPort);
	NameByPtr(ProcessorToL1InstructionInputPort);
	NameByPtr(L1DataToProcessorOutputPort);
	NameByPtr(L1DataToProcessorInputPort);
	NameByPtr(ProcessorToL1DataOutputPort);
	NameByPtr(ProcessorToL1DataInputPort);
	NameByPtr(L1ToL2InputPort);
	NameByPtr(L1ToL2OutputPort);
	NameByPtr(L2ToSDRAMInputPort);
	NameByPtr(L2ToSDRAMOutputPort);
	NameByPtr(L1ToL2Bus);
	NameByPtr(L2ToSDRAMBus);
	NameByPtr(L1InstructionToProcessorBus);
	NameByPtr(L1DataToProcessorBus);
	NameByPtr(ProcessorToL1InstructionBus);
	NameByPtr(ProcessorToL1DataBus);
	NameByPtr(SDRAMModel);
	NameByPtr(L2Cache);
	NameByPtr(L1InstructionCache);
	NameByPtr(L1DataCache);
*/

	//////////////////////////////////////////////////////////////////////////////////////////
	// connect busses to ports
	L1InstructionToProcessorBus->SetOutputPort(L1InstructionToProcessorOutputPort);
	L1InstructionToProcessorBus->SetInputPort(L1InstructionToProcessorInputPort);
	ProcessorToL1InstructionBus->SetOutputPort(ProcessorToL1InstructionOutputPort);
	ProcessorToL1InstructionBus->SetInputPort(ProcessorToL1InstructionInputPort);
	L1DataToProcessorBus->SetOutputPort(L1DataToProcessorOutputPort);
	L1DataToProcessorBus->SetInputPort(L1DataToProcessorInputPort);
	ProcessorToL1DataBus->SetOutputPort(ProcessorToL1DataOutputPort);
	ProcessorToL1DataBus->SetInputPort(ProcessorToL1DataInputPort);
	L1ToL2Bus->SetOutputPort(L1ToL2OutputPort);
	L1ToL2Bus->SetInputPort(L1ToL2InputPort);
	L2ToSDRAMBus->SetOutputPort(L2ToSDRAMOutputPort);
	L2ToSDRAMBus->SetInputPort(L2ToSDRAMInputPort);

	//////////////////////////////////////////////////////////////////////////////////////////
	// add them to the simulation pool
	AddComponent(L1InstructionToProcessorBus);
	AddComponent(L1InstructionToProcessorInputPort);
	AddComponent(L1InstructionToProcessorOutputPort);
	AddComponent(ProcessorToL1InstructionBus);
	AddComponent(ProcessorToL1InstructionInputPort);
	AddComponent(ProcessorToL1InstructionOutputPort);
	AddComponent(L1DataToProcessorBus);
	AddComponent(L1DataToProcessorInputPort);
	AddComponent(L1DataToProcessorOutputPort);
	AddComponent(ProcessorToL1DataBus);
	AddComponent(ProcessorToL1DataInputPort);
	AddComponent(ProcessorToL1DataOutputPort);
	AddComponent(L1ToL2Bus);
	AddComponent(L1ToL2InputPort);
	AddComponent(L1ToL2OutputPort);
	AddComponent(L2ToSDRAMBus);
	AddComponent(L2ToSDRAMInputPort);
	AddComponent(L2ToSDRAMOutputPort);
	AddComponent(L1InstructionCache);
	AddComponent(L1DataCache);
	AddComponent(L2Cache);
	AddComponent(SDRAMModel);
	return(true);
	}

bool
SimpleCacheMemorySystem	::	BindStructures(TupleSpace	*theTupleSpace)
	{
	L1InstructionToProcessorBus->BindStructures(theTupleSpace);
	ProcessorToL1InstructionBus->BindStructures(theTupleSpace);
	L1DataToProcessorBus->BindStructures(theTupleSpace);
	ProcessorToL1DataBus->BindStructures(theTupleSpace);
	L2ToSDRAMBus->BindStructures(theTupleSpace);
	L1ToL2Bus->BindStructures(theTupleSpace);
					
	L2Cache->BindStructures(	L1ToL2InputPort, L1ToL2OutputPort,				// access
								L2ToSDRAMInputPort, L2ToSDRAMOutputPort);		// fill
	L1InstructionCache->BindStructures(ProcessorToL1InstructionInputPort, L1InstructionToProcessorOutputPort,	// access
										L1ToL2InputPort, L1ToL2OutputPort);										// fill
	L1DataCache->BindStructures(ProcessorToL1DataInputPort, L1DataToProcessorOutputPort,	// access
										L1ToL2InputPort, L1ToL2OutputPort);					// fill
					
	SDRAMModel->BindStructures(theTupleSpace, L2ToSDRAMBus->GetInputPort(), L2ToSDRAMBus->GetOutputPort());
	return(true);
	}

bool
SimpleCacheMemorySystem	::	PrintStatistics(
	OutputFormat	theFormat,
	FILE			*fp)
	{
	L1InstructionToProcessorBus->PrintStatistics(theFormat, fp);
	ProcessorToL1InstructionBus->PrintStatistics(theFormat, fp);
	L1DataToProcessorBus->PrintStatistics(theFormat, fp);
	ProcessorToL1DataBus->PrintStatistics(theFormat, fp);
	L1ToL2Bus->PrintStatistics(theFormat, fp);
	L2ToSDRAMBus->PrintStatistics(theFormat, fp);
	L1InstructionCache->PrintStatistics(theFormat, fp);
	L1DataCache->PrintStatistics(theFormat, fp);
	L2Cache->PrintStatistics(theFormat, fp);
	SDRAMModel->PrintStatistics(theFormat, fp);
	return ( true );
	}

