#	include	"Standard.h"


bool
SimpleSDRAMModel	::	BindParameters(SimpleSDRAMModel	*This, TupleSpace	*theTupleSpace,	char	*aName )
	{
	theTupleSpace->StartSubSpace(DefaultPass(aName, "SDRAM"));
	SetComponentName(This, theTupleSpace);
	theTupleSpace->Bind(Parameter(This, Latency), 60);
	theTupleSpace->Bind(Parameter(This, MaxOutstandingRequests), 8);
	theTupleSpace->EndSubSpace();
	return(true);
	}

bool
SimpleSDRAMModel	::	BindStructures(
			TupleSpace	*theTupleSpace,
			InputPort	*aInputPort,
			OutputPort	*aOutputPort)
	{
	RequestsInFlight = 0;
	theOutputPort = aOutputPort;
	theInputPort = aInputPort;
	theInputPort->Connect(this);
	Overflow.Resize(MaxOutstandingRequests);
	OutputQueue.Resize(MaxOutstandingRequests);
	return(true);
	}

int
SimpleSDRAMModel	::	WhenCanYouTakeThisMessage (
	Message	*m,
	Port	*p)
	{
	MemoryTransactionMessage	*mtm = (MemoryTransactionMessage*) m;
	MemoryTransactionMessage	*mtm2;

	if( (mtm->Command != Read) && (mtm->Command != Write) )	return(0);	// message will be dropped correctly

	if ( MaxOutstandingRequests )
	if ( RequestsInFlight >= MaxOutstandingRequests )
		{
		mtm2 = ProcessedEvents.Peek(GetDomain()->GetCurrentCycle());
		if ( mtm2 )
			{
			if ( mtm2->Command == Write )
				return(0);	// now will work
			}
		return(1);	// try again later
		}

	return(0);	// now is fine
	}
	
bool
SimpleSDRAMModel	::	ReceiveMessage (
	Message	*m,
	Port	*p)
	{
	MemoryTransactionMessage	*mtm = (MemoryTransactionMessage*) m;
	MemoryTransactionMessage	*mtm2;

	if( (mtm->Command != Read) && (mtm->Command != Write) )
		{
		DELETE(MemoryTransactionMessage, mtm);
		return(true);
		}
	if ( MaxOutstandingRequests )
	if ( RequestsInFlight >= MaxOutstandingRequests )
		{
		mtm2 = ProcessedEvents.Peek(GetDomain()->GetCurrentCycle());
		if ( mtm2 )
			{
			if ( mtm2->Command == Write )
				{
				ProcessedEvents.Pop(GetDomain()->GetCurrentCycle());
				--RequestsInFlight;
				DELETE(MemoryTransactionMessage, mtm2);
				}
			else	
			if ( mtm2->Command == Read )
				{
				mtm2->Command = ReadResponse;
				mtm2->SetSourceComponentId(GetId());
				OutputQueue.Insert(mtm2);
				ProcessedEvents.Pop(GetDomain()->GetCurrentCycle());
				}
			else
				FatalError("Hmmm...");
			}
		if ( ! ProcessedEvents.GetSize() )
		if ( Overflow.GetSize() )
			{
			Overflow.Remove(&mtm);
			ProcessedEvents.Push(mtm, InDomain->GetCurrentCycle() + Latency);
			}	
		}

	if ( MaxOutstandingRequests )
	if ( RequestsInFlight >= MaxOutstandingRequests )
		{
		return ( false );
		}
	if ( ProcessedEvents.GetSize() )
		Overflow.Insert(mtm);
	else
		{
		ProcessedEvents.Push(mtm, InDomain->GetCurrentCycle() + Latency);
		}
	++RequestsInFlight;
	if ( OutputQueue.GetSize() )
		{
		OutputQueue.Peek(&mtm);

		if ( theOutputPort->SendMessage(mtm) )
			{
			--RequestsInFlight;
			OutputQueue.Remove(&mtm);
			}
		}
	
	return(true);
	}

int
SimpleSDRAMModel	::	AsyncSimulateOneCycle()
	{
	MemoryTransactionMessage	*mtm;

	while ( true )
		{
		mtm = ProcessedEvents.Peek(InDomain->GetCurrentCycle());
		if ( mtm )
			{
			if ( mtm->Command == Write )
				{
				--RequestsInFlight;
				ProcessedEvents.Pop(InDomain->GetCurrentCycle());
				DELETE(MemoryTransactionMessage, mtm);
				continue;
				}
			else	
			if ( mtm->Command == Read )
				{
				mtm->Command = ReadResponse;
				mtm->SetSourceComponentId(GetId());
				ProcessedEvents.Pop(InDomain->GetCurrentCycle());
				OutputQueue.Insert(mtm);
				continue;
				}
			else
				FatalError("Hmmm...");
			}
		if ( ! ProcessedEvents.GetSize() )
		if ( Overflow.GetSize() )
			{
			Overflow.Remove(&mtm);
			ProcessedEvents.Push(mtm, InDomain->GetCurrentCycle() + Latency);
			continue;
			}
		break;
		}

	while ( OutputQueue.GetSize() )
		{
		OutputQueue.Peek(&mtm);

		if ( theOutputPort->SendMessage(mtm) )
			{
			--RequestsInFlight;
			OutputQueue.Remove(&mtm);
			}
		else
			break;
		}
	return(1);
	}

