#	include	"Standard.h"

bool
BranchTargetBuffer	::	ResetStatistics()
	{
	CorrectPredictions = IncorrectPredictions = 0;
	return(true);
	}

bool
BranchTargetBuffer	::	PrintStatistics (
	OutputFormat	theFormat,
	FILE			*fp)
	{
	PrintValue(theFormat, fp, "CorrectPredictions", CorrectPredictions);
	PrintValue(theFormat, fp, "IncorrectPredictions", IncorrectPredictions);
	PrintPercent(theFormat, fp, "PredictionAccuracy", CorrectPredictions, IncorrectPredictions);
	return(true);
	}


bool
SimpleBranchTargetBuffer	::	BindParameters(SimpleBranchTargetBuffer	*This, TupleSpace	*theTupleSpace)
	{
	theTupleSpace->StartSubSpace("SimpleBranchTargetBuffer");
	SetComponentName(This, theTupleSpace);
	theTupleSpace->Bind(Parameter(This, ShiftRight), 2);
	theTupleSpace->Bind(Parameter(This, TableSize), 1024);
	theTupleSpace->EndSubSpace();
	return(true);
	}

bool
SimpleBranchTargetBuffer	::	BindStructures()
	{
	Mask = TableSize - 1;
	PredictedPCTable = new uint64[TableSize];
	for(int Index = 0;Index < TableSize; Index++)
		PredictedPCTable[Index] = 0;
	CorrectPredictions = IncorrectPredictions = 0;
	return ( true );	
	}

uint64
SimpleBranchTargetBuffer	::	AccessBranchTargetBuffer(
	uint64	BranchPC,
	Instruction	*theInstruction )
	{
	if ( PredictedPCTable[ (BranchPC>>ShiftRight) & Mask ] == theInstruction->NextProgramCounter )
		++CorrectPredictions;
	else
		++IncorrectPredictions;
	// perfect indirect branch prediction??
	//return ( theInstruction->NextProgramCounter);
	return ( PredictedPCTable[ (BranchPC>>ShiftRight) & Mask ] );
	}

bool
SimpleBranchTargetBuffer	::	UpdateBranchTargetBuffer (
	uint64	BranchPC,
	uint64	TargetPC,
	Instruction	*theInstruction )
	{
	PredictedPCTable [ (BranchPC>>ShiftRight) & Mask ] = TargetPC;
	return ( true );
	}

bool
SetAssociativeBranchTargetBuffer	::	BindParameters(SetAssociativeBranchTargetBuffer	*This, TupleSpace	*theTupleSpace)
	{
	theTupleSpace->StartSubSpace("SetAssociativeBranchTargetBuffer");
	SetComponentName(This, theTupleSpace);
	theTupleSpace->Bind(Parameter(This, BankSize), 1024);
	theTupleSpace->Bind(Parameter(This, Associativity), 4);
	theTupleSpace->Bind(Parameter(This, ReturnAddressStackSize), 8); 
	theTupleSpace->EndSubSpace();
	return(true);
	}

bool
SetAssociativeBranchTargetBuffer	::	BindStructures()
	{
	CorrectPredictions = IncorrectPredictions = 0;
	theMemory = SetAssociativeMemory<uint64>::Create(sizeof(uint32), BankSize, Associativity, LRU);
	ReturnAddressStackPointer = 0;
	ReturnAddressStackCurrentSize = 0;
	if ( ReturnAddressStackSize )
		ReturnAddressStack = new uint64[ReturnAddressStackSize];
	return ( true );	
	}

uint64
SetAssociativeBranchTargetBuffer	::	AccessBranchTargetBuffer(
	uint64	BranchPC,
	Instruction	*theInstruction )
	{
	uint			HitOrMiss;
	uint			ignore1;
	Line<uint64>	ignore2;

	if ( ReturnAddressStackSize )
	if ( (theInstruction->theAlphaInstruction.theOpCode == jsr )||
		 (theInstruction->theAlphaInstruction.theOpCode == bsr ) )
		{
		ReturnAddressStack [ ReturnAddressStackPointer ] = BranchPC + sizeof(uint32);
		ReturnAddressStackPointer =  ( ReturnAddressStackPointer + 1 ) % ReturnAddressStackSize;
		if ( ReturnAddressStackCurrentSize < ReturnAddressStackSize )
			++ReturnAddressStackCurrentSize;
		}
	if ( ReturnAddressStackSize )
	if ( theInstruction->theAlphaInstruction.theOpCode == ret )
	if ( ReturnAddressStackCurrentSize )
		{
		if ( ReturnAddressStackPointer )
			ReturnAddressStackPointer =  ( ReturnAddressStackPointer - 1 );
		else
			ReturnAddressStackPointer = ReturnAddressStackSize - 1;
		uint64	Prediction = ReturnAddressStack [ ReturnAddressStackPointer ];
		
		
		--ReturnAddressStackCurrentSize;
		if ( Prediction  == theInstruction->NextProgramCounter )
			{
			++CorrectPredictions;
			}
		else
			{
			++IncorrectPredictions;
			}

		return ( Prediction );
		}

	Line<uint64>	*theLine = theMemory->Locate(BranchPC, &HitOrMiss, &ignore1, &ignore2);

	if ( theLine->theObject  == theInstruction->NextProgramCounter )
		++CorrectPredictions;
	else
		++IncorrectPredictions;

	return ( theLine->theObject );
	}

bool
SetAssociativeBranchTargetBuffer	::	UpdateBranchTargetBuffer (
	uint64	BranchPC,
	uint64	TargetPC,
	Instruction	*theInstruction )
	{
	uint			HitOrMiss;
	uint			ignore1;
	Line<uint64>	ignore2;

	Line<uint64>	*theLine = theMemory->Locate(BranchPC, &HitOrMiss, &ignore1, &ignore2);

	theLine->theObject = TargetPC;
	return ( true );
	}


