#	include	"Standard.h"



bool
CompletionStage	::	BindParameters(CompletionStage	*This, TupleSpace	*theTupleSpace, char	*aName )
	{
	theTupleSpace->StartSubSpace(DefaultPass(aName, "CompletionStage"));
	SetComponentName(This, theTupleSpace);
	theTupleSpace->Bind(Parameter(This, CompletionBufferSize), "PhysicalRegisterFileSize", 128);
	theTupleSpace->Bind(Parameter(This, MaximumDispatchesPerCycle), "MachineWidth", 4);
	theTupleSpace->Bind(Parameter(This, MaximumCompletionsPerCycle), "MachineWidth", 4);
	theTupleSpace->EndSubSpace();
	return(true);
	}

bool
CompletionStage	::	BindStructures(
						FetchStage	*aFetchStage,
						RenameStage	*aRenameStage)
	{
	theFetchStage = aFetchStage;
	theRenameStage = aRenameStage;
	SequenceNumber = 0;
	for(int Index = 0;Index < CompletionBufferSize;Index++)
		CompletionBuffer.Add(NULL);

	memset(RetiredInstructions, 0, sizeof(uint64)*TotalOps);
	return(true);
	}
	
bool
CompletionStage	::	PrintDebuggingInformation(FILE	*fp, int	Level)
	{
	PrintValue(HumanFormat, fp, "SequenceNumber", SequenceNumber);
	uint	Count = 0;
	for(int Index = 0; Index <  CompletionBufferSize; Index++)
		{
		if ( CompletionBuffer[Index] )
			++Count;
		}
	PrintValue(HumanFormat, fp, "CompletionBufferSize", CompletionBufferSize);
	PrintValue(HumanFormat, fp, "InstructionsInFlight", Count);
	return ( true );
	}
	
bool
CompletionStage	::	DispatchTo (
	Instruction	*aInstruction )
	{
	if ( MaximumDispatchesPerCycle )
	if ( DispatchesThisCycle >= MaximumDispatchesPerCycle )
		return ( false );

	if ( aInstruction->Sequence >= (SequenceNumber + CompletionBufferSize))
		return ( false );	// avoiding live lock
	
	CompletionBuffer[aInstruction->Sequence % CompletionBufferSize ] = aInstruction;
	aInstruction->CurrentProcessorStage = this;
	++DispatchesThisCycle;
	return ( true );
	}

bool
CompletionStage	::	ReserveSpaceFor (
	Instruction	*aInstruction )
	{
	if ( aInstruction->Sequence >= (SequenceNumber + CompletionBufferSize))
		return ( false );	// nope, can't do it

	return ( true );
	}

int
CompletionStage	::	SimulateOneCycle ()
	{
	DispatchesThisCycle = 0;
	CompletionsThisCycle = 0;

	while ( true )
		{
		if ( MaximumCompletionsPerCycle )
		if ( CompletionsThisCycle >= MaximumCompletionsPerCycle )
			break;

		// not hit limit
		if ( ! CompletionBuffer[SequenceNumber % CompletionBufferSize] )
			break;	// nothing to complete (empty)

		Instruction	*theInstruction = CompletionBuffer[SequenceNumber % CompletionBufferSize];

		theRenameStage->AddToFreeList(theInstruction->AddToFreeList1);
		theRenameStage->AddToFreeList(theInstruction->AddToFreeList2);

		CompletionBuffer[SequenceNumber % CompletionBufferSize] = NULL;

		if ( !theInstruction->MisSpeculatedInstruction )
			++RetiredInstructions[theInstruction->Operation];
		++SequenceNumber;
		++CompletionsThisCycle;
		theFetchStage->FreeInstruction(theInstruction);
		}
	return ( true );
	}

bool
CompletionStage	::	FlushSpeculativeState()
	{
	SequenceNumber = nscSequenceNumber;
	for(int	Index = 0;Index < CompletionBufferSize;Index++)
		CompletionBuffer[Index] = NULL;
	return ( true );
	}

bool
CompletionStage	::	EnterSpeculativeState(
	uint64	MisSpeculatedInstruction)
	{
	nscSequenceNumber = MisSpeculatedInstruction;
	return ( true );
	}

bool
CompletionStage	::	PrintStatistics(
	OutputFormat	theFormat,
	FILE			*fp)
	{
	int	Index;
	char	s[128];

	uint64	Total = 0;
	for(Index =0; Index < TotalOps; Index++)
		{
		sprintf(s, "RetiredOperations.%s", OpNames[Index]);
		PrintValue(theFormat, fp, s, RetiredInstructions[Index]);
		Total += RetiredInstructions[Index];
		}
	PrintValue(theFormat, fp, "RetiredOperations.Total", Total);

	double	IPC = ((double)Total) / ((double)GetDomain()->GetStatCurrentCycle());

	PrintValue(theFormat, fp, "IPC", IPC);
	return(true);
	}

bool
CompletionStage	::	ResetStatistics()	
	{
	int	Index;

	for(Index =0; Index < TotalOps; Index++)
		RetiredInstructions[Index] = 0;
	
	return(true);
	}

