#	include	"Standard.h"


bool
DelayStage	::	Initialize (
						int		aDelayFor,
						int		aMaximumInstructionsInFlight,
						bool	aPullInstructions,
						bool	aPushInstructions )
		
	{
	DelayFor = aDelayFor;
	MaximumInstructionsInFlight = aMaximumInstructionsInFlight;
	PullInstructions = aPullInstructions;
	PushInstructions = aPushInstructions;

	CanLeave.Resize(MaximumInstructionsInFlight);

	Waiting.Resize(MaximumInstructionsInFlight);
	EnteredAt.Resize(MaximumInstructionsInFlight);
	CurrentTime = 0;
	InstructionsInFlight = 0;
	SetName("DelayStage");
	return(true);
	}

bool
DelayStage	::	AcquireFrom (
	Instruction	**aInstruction,
	int			OfType )
	{
	(*aInstruction) = NULL;
	if ( OfType == InstructionTypeAll )
		{
		if ( CanLeave.GetSize() )
			{
			CanLeave.Remove(aInstruction);
			--InstructionsInFlight;
			return ( true );
			}
		}

	if ( !CanLeave.GetSize() )
		return(false);

	CanLeave.PeekAt(aInstruction);
	if((*aInstruction)->theInstructionType != OfType)
		return(false);

	CanLeave.Remove(aInstruction);
	--InstructionsInFlight;
	return ( true );
	}

bool
DelayStage	::	PeekAt (
	Instruction	**aInstruction,
	int			OfType )
	{
	(*aInstruction) = NULL;
	if ( OfType == InstructionTypeAll )
		{
		if ( CanLeave.GetSize() )
			{
			CanLeave.PeekAt(aInstruction);
			return ( true );
			}
		}

	if ( !CanLeave.GetSize() )
		return(false);

	CanLeave.PeekAt(aInstruction);
	if((*aInstruction)->theInstructionType != OfType)
		return(false);

	return ( true );
	}

bool
DelayStage	::	DispatchTo (
	Instruction	*aInstruction )
	{
	if ( InstructionsInFlight >= MaximumInstructionsInFlight )
		return ( false );
	Waiting.Insert(aInstruction);
	EnteredAt.Insert(CurrentTime);
	++InstructionsInFlight;
	return ( true );
	}

int
DelayStage	::	SimulateOneCycle ()
	{
	// possibly migrate stuff from the Waiting for into the CanLeave
	if ( PullInstructions )
	if ( UpStream )
		{
		Instruction	*theInstruction;

		while ( UpStream->PeekAt(&theInstruction) && (InstructionsInFlight < MaximumInstructionsInFlight))
			{
			UpStream->AcquireFrom(&theInstruction);

			theInstruction->CurrentProcessorStage = this;
			Waiting.Insert(theInstruction);
			EnteredAt.Insert(CurrentTime);
			++InstructionsInFlight;
			}
		}
	++CurrentTime;
	while ( EnteredAt.GetSize() )
		{
		Counter	theCounter;

		EnteredAt.PeekAt(&theCounter);

		if ( (theCounter + DelayFor) <= CurrentTime )
			{
			Instruction		*theInstruction;

			EnteredAt.Remove(&theCounter);
			Waiting.Remove(&theInstruction);
			CanLeave.Insert(theInstruction);
			}
		else
			break;
		}
	if ( PushInstructions )
	if ( DownStream )
		{
		Instruction	*theInstruction;
		// "peek" at instruction and dispatch.. if succesful then vaporize it with a pull
		while ( PeekAt(&theInstruction) )
			{
			if ( DownStream->DispatchTo(theInstruction) )
				AcquireFrom(&theInstruction);
			else
				break;
			}
		}
	return(0);
	}

bool
DelayStage	::	FlushSpeculativeState ()
	{
	InstructionsInFlight = 0;
	Waiting.Clear();
	EnteredAt.Clear();
	CanLeave.Clear();
	return ( true );
	}

