#	include	"Standard.h"


bool
ExecutionStage	::	BindParameters(ExecutionStage	*This, TupleSpace	*theTupleSpace, char	*aName )
	{
	theTupleSpace->StartSubSpace(DefaultPass(aName, "ExecutionStage"));
	SetComponentName(This, theTupleSpace);
	theTupleSpace->Bind(Parameter(This, MaxExecutePerCycle), "MachineWidth", 4);
	theTupleSpace->Bind(Parameter(This, MaxIntegerExecutionCores), "MachineWidth", 4);
	theTupleSpace->Bind(Parameter(This, MaxFloatingPointExecutionCores), "MachineWidth", 4);
	theTupleSpace->EndSubSpace();
	return(true);	
	}

bool
ExecutionStage	::	BindStructures(
						FetchStage				*aFetchStage,
						RenameStage				*aRenameStage,
						RegisterAccessStage		*aRegisterAccessStage,
						StoreBuffer				*aStoreBuffer,
						CompletionStage			*aCompletionStage)
	{
	theFetchStage = aFetchStage;
	theRenameStage = aRenameStage;
	theRegisterAccessStage = aRegisterAccessStage;
	theStoreBuffer = aStoreBuffer;
	theCompletionStage = aCompletionStage;
	return ( true );
	}

bool
ExecutionStage	::	PrintDebuggingInformation(FILE	*fp, int	Level)
	{
	PrintValue(HumanFormat, fp, "MaxIntegerExecutionCores", MaxIntegerExecutionCores);
	PrintValue(HumanFormat, fp, "InUseIntegerExecutionCores", IntegerExecutionCores.GetSize());
	PrintValue(HumanFormat, fp, "MaxFloatingPointExecutionCores", MaxFloatingPointExecutionCores);
	PrintValue(HumanFormat, fp, "InUseFloatingPointExecutionCores", FloatingPointExecutionCores.GetSize());
	for(int	Index=0;Index<IntegerExecutionCores.GetSize();Index++)
		{
		fprintf(fp, "core%d: ", Index);
		Debugger::DumpInstruction(IntegerExecutionCores[Index], theRenameStage);
		Debugger::PrintDependencies(2, IntegerExecutionCores[Index], theFetchStage, theRenameStage);
		}
	return(true);
	}

int
ExecutionStage	::	SimulateOneCycle()
	{
	int	Index = 0;
	Instruction	*theInstruction;

	// go through the execution units and flush the ones complete
	Index = IntegerExecutionCores.GetSize() - 1;
	while ( Index >= 0 )
		{
		theInstruction = IntegerExecutionCores[Index];
		if(!theInstruction->Dispatched)
		if(theRenameStage->IsRegisterValid(theInstruction->Input1))
		if(theRenameStage->IsRegisterValid(theInstruction->Input2))
		if(theRenameStage->IsRegisterValid(theInstruction->Input3))
			{
			theInstruction->Dispatched = 1;
			theRenameStage->NotifyAboutRegisterUse(theInstruction->Input1);
			theRenameStage->NotifyAboutRegisterUse(theInstruction->Input2);
			theRenameStage->NotifyAboutRegisterUse(theInstruction->Input3);
			}
		if(theInstruction->Dispatched)
		if(theInstruction->Delay > 0)
			--theInstruction->Delay;
		if(theInstruction->Delay <= 0 )
			{
			// finish off this thing
			if ( theInstruction->theInstructionType == InstructionTypeMemory )
				{
				if ( theStoreBuffer->DispatchTo(theInstruction) )
					IntegerExecutionCores.RemoveAt(Index);
				}
			else
			if(DownStream->DispatchTo(theInstruction))
				{
				theRenameStage->WriteBackRegister(theInstruction->Output1);
				theRenameStage->WriteBackRegister(theInstruction->Output2);
				theRenameStage->NotifyAboutRegisterUse(theInstruction->Output1);
				theRenameStage->NotifyAboutRegisterUse(theInstruction->Output2);
				theRegisterAccessStage->WriteBackRegister(theInstruction->Output1);
				theRegisterAccessStage->WriteBackRegister(theInstruction->Output2);
				IntegerExecutionCores.RemoveAt(Index);
				theInstruction->Executed = 1;
				}
			}
		--Index;
		}

	Index = FloatingPointExecutionCores.GetSize() - 1;
	while ( Index >= 0 )
		{
		theInstruction = FloatingPointExecutionCores[Index];
		if(!theInstruction->Dispatched)
		if(theRenameStage->IsRegisterValid(theInstruction->Input1))
		if(theRenameStage->IsRegisterValid(theInstruction->Input2))
		if(theRenameStage->IsRegisterValid(theInstruction->Input3))
			{
			theInstruction->Dispatched = 1;
			theRenameStage->NotifyAboutRegisterUse(theInstruction->Input1);
			theRenameStage->NotifyAboutRegisterUse(theInstruction->Input2);
			theRenameStage->NotifyAboutRegisterUse(theInstruction->Input3);
			}
		if(theInstruction->Dispatched)
		if(theInstruction->Delay > 0)
			--theInstruction->Delay;
		if(theInstruction->Delay <= 0 )
		if(DownStream->DispatchTo(theInstruction))
			{
			theRenameStage->WriteBackRegister(theInstruction->Output1);
			theRenameStage->WriteBackRegister(theInstruction->Output2);
			theRenameStage->NotifyAboutRegisterUse(theInstruction->Output1);
			theRenameStage->NotifyAboutRegisterUse(theInstruction->Output2);
			theRegisterAccessStage->WriteBackRegister(theInstruction->Output1);
			theRegisterAccessStage->WriteBackRegister(theInstruction->Output2);
			FloatingPointExecutionCores.RemoveAt(Index);
			theInstruction->Executed = 1;
			}
		--Index;
		}

	Index = 0;
	while ( Index < MaxExecutePerCycle )
		{
		if ( IntegerExecutionCores.GetSize() >= MaxIntegerExecutionCores )
			break;	// there is no free space

		if(UpStream->AcquireFrom(&theInstruction, InstructionTypeMemory))
			{
			theInstruction->CurrentProcessorStage = this;
			IntegerExecutionCores.Add(theInstruction);
			if(theRenameStage->IsRegisterValid(theInstruction->Input1))
			if(theRenameStage->IsRegisterValid(theInstruction->Input2))
			if(theRenameStage->IsRegisterValid(theInstruction->Input3))
				{
				theInstruction->Dispatched = 1;
				theRenameStage->NotifyAboutRegisterUse(theInstruction->Input1);
				theRenameStage->NotifyAboutRegisterUse(theInstruction->Input2);
				theRenameStage->NotifyAboutRegisterUse(theInstruction->Input3);
				}
			++Index;
			}
		else
		if(UpStream->AcquireFrom(&theInstruction, InstructionTypeInteger))
			{
			theInstruction->CurrentProcessorStage = this;
			IntegerExecutionCores.Add(theInstruction);
			if(theRenameStage->IsRegisterValid(theInstruction->Input1))
			if(theRenameStage->IsRegisterValid(theInstruction->Input2))
			if(theRenameStage->IsRegisterValid(theInstruction->Input3))
				{
				theInstruction->Dispatched = 1;
				theRenameStage->NotifyAboutRegisterUse(theInstruction->Input1);
				theRenameStage->NotifyAboutRegisterUse(theInstruction->Input2);
				theRenameStage->NotifyAboutRegisterUse(theInstruction->Input3);
				}
			++Index;
			}
		else
			break;
		}

	while ( Index < MaxExecutePerCycle )
		{
		if ( FloatingPointExecutionCores.GetSize() >= MaxFloatingPointExecutionCores )
			break;	// there is no free space

		if(UpStream->AcquireFrom(&theInstruction, InstructionTypeFloatingPoint))
			{
			theInstruction->CurrentProcessorStage = this;
			FloatingPointExecutionCores.Add(theInstruction);
			if(theRenameStage->IsRegisterValid(theInstruction->Input1))
			if(theRenameStage->IsRegisterValid(theInstruction->Input2))
			if(theRenameStage->IsRegisterValid(theInstruction->Input3))
				{
				theInstruction->Dispatched = 1;
				theRenameStage->NotifyAboutRegisterUse(theInstruction->Input1);
				theRenameStage->NotifyAboutRegisterUse(theInstruction->Input2);
				theRenameStage->NotifyAboutRegisterUse(theInstruction->Input3);
				}
			++Index;
			}
		else
			break;
		}
	return ( 1 );
	}

bool
ExecutionStage	::	FlushSpeculativeState()
	{
	IntegerExecutionCores.Clear();
	FloatingPointExecutionCores.Clear();
	return ( true );
	}


