#	include	"Standard.h"



bool
FetchStage	::	BindParameters(FetchStage	*This, TupleSpace	*theTupleSpace, char	*aName )
	{
	theTupleSpace->StartSubSpace(DefaultPass(aName, "FetchStage"));
	SetComponentName(This, theTupleSpace);
	theTupleSpace->Bind(Parameter(This, MaxFetchPerCycle), "MachineWidth", 4);
	theTupleSpace->Bind(Parameter(This, MaxOutstandingMemoryTransactions), 1);
	theTupleSpace->Bind(Parameter(This, GarbageCollectorSize), NULL, 10000);
	theTupleSpace->EndSubSpace();
	return(true);
	}
	
bool
FetchStage	::	BindStructures(Engine				*aEngine,
							   BranchPredictor		*aBranchPredictor,	
							   BranchTargetBuffer	*aBranchTargetBuffer,
							   Component			*aMemorySystem,
							   Array<ProcessorStage *>	*aAllStages,
							   StoreBuffer				*aStoreBuffer)
	{
	Buffer.Resize(MaxFetchPerCycle*2);
	theEngine = aEngine;
	theBranchPredictor = aBranchPredictor;
	theBranchTargetBuffer = aBranchTargetBuffer;
	theMemorySystem = aMemorySystem;
	AllStages = aAllStages;
	theStoreBuffer = aStoreBuffer;

	AddressMask = ~((uint64)MaxFetchPerCycle * sizeof(uint32) - 1);
	OutstandingMemoryTransactions = 0;
	NextMemoryTransactionId = 0;
	EngineSlack.Resize(GarbageCollectorSize);

	EndSimulation = false;
	Speculative = false;
	Sequence = 0;
	LoadStoreSequence = 0;
	StoreSequence = 0;
	MemoryFetchId = 0;
	GarbageCollector.Resize(GarbageCollectorSize);

	aSourcePort->Connect(this);
	aDestinationPort->Connect(this);

	OutstandingMemoryTransactions = 0;
	OutgoingMemoryRequests.Resize(MaxOutstandingMemoryTransactions);
	OutstandingRequestedAddresses.Resize(MaxOutstandingMemoryTransactions);

	FetchAddress = DecodeAddress = 0;	// will cause a bogus read to occur but thats life.
	return(true);
	}

bool
FetchStage	::	ReceiveMessage (Component *f, Message	*m)
	{
	MemoryTransactionMessage	*mtm = (MemoryTransactionMessage *)m;

	if ( mtm->Command != ReadResponse )
		{
		DELETE ( MemoryTransactionMessage, mtm );
		return ( true );
		}

	if ( mtm->Id3 != (uint64)GetId() )
		{
		DELETE ( MemoryTransactionMessage, mtm );
		return ( true );
		}
	//printf("" format_dec_int64 ": receiving ", GetDomain()->GetCurrentCycle());mtm->Dump(stdout);
	MemoryTransactions.Push(mtm);
	InDomain->ComponentIsLive(this, InDomain->GetCurrentCycle());
	return ( true );
	}

int
FetchStage	::	AsyncSimulateOneCycle ()
	{

	while ( (Buffer.GetSize() < MaxFetchPerCycle)  && (MemoryTransactions.GetSize()) )
		{
		MemoryTransactionMessage	*mtm;

		mtm = MemoryTransactions.Peek();
		//printf("" format_dec_int64 ": processing ", GetDomain()->GetCurrentCycle());mtm->Dump(stdout);
		// the cache can make requests come back out of order
		// need to handle them in-order
		if ( mtm->Id == NextMemoryTransactionId )
			{
			// process a line
			mtm = MemoryTransactions.Pop();
			++NextMemoryTransactionId;
			--OutstandingMemoryTransactions;
			uint64	aAddress;
			OutstandingRequestedAddresses.Remove(&aAddress);
		
			uint64	Index = 0;
			int		Added = 0;
			Index = (DecodeAddress & (~AddressMask))>>2;
			while ( Index < (uint64)MaxFetchPerCycle )
				{
				Instruction	*theInstruction = NULL;
				if(!EngineSlack.GetSize())
					{
					theInstruction = NEW(Instruction);
					if(!theEngine->NextInstruction(&theInstruction))
						{
						// out of instructions
						DELETE(Instruction, theInstruction);
						EndSimulation = 1;
						break;
						}
					}
				else
					{
					EngineSlack.Remove(&theInstruction);
					}


				//////////////////////////////////////////////////////////////////////////////////////////////
				// examine theInstruction -- it's address should be part of the mtm

				if ( (mtm->Address & AddressMask) != (theInstruction->ProgramCounter & AddressMask) )
					{
					EngineSlack.PushBack(theInstruction);	// wowa, undo this and push this instruction back onto the engine slack
					// check to see if we have actually requested the correct address
					int	Index2 = 0;
					while ( Index2 < OutstandingRequestedAddresses.GetSize() )
						{
						OutstandingRequestedAddresses.PeekAt(&aAddress, Index2);
						if ( (aAddress & AddressMask) == (theInstruction->ProgramCounter & AddressMask) )
							break;
						++Index2;
						}

					if ( Index2 >= OutstandingRequestedAddresses.GetSize() )
						{
						FetchAddress = theInstruction->ProgramCounter;
						}
					break;
					}
				if ( (mtm->Address & AddressMask) != (DecodeAddress & AddressMask) )
					{
					EngineSlack.PushBack(theInstruction);	// wowa, undo this and push this instruction back onto the engine slack
					//printf("Discarding - mtm->Address (" format_hex_uint64 ") != DecodeAddress (" format_hex_uint64 ") ", mtm->Address, DecodeAddress);
					FetchAddress = theInstruction->ProgramCounter;
					DecodeAddress = FetchAddress;
					break;
					}
					

				//////////////////////////////////////////////////////////////////////////////////////////////
				// add the instruction into the buffer for sending into later pipeline stages.

				theInstruction->Sequence = Sequence;
				theInstruction->StoreSequence = StoreSequence;
				theInstruction->LoadStoreSequence = LoadStoreSequence;

				theInstruction->MisSpeculatedInstruction = Speculative;
				theInstruction->RollBackOnThisInstruction = false;

				++Sequence;
				if ( theInstruction->Operation == OpStore )
					{
					++StoreSequence;
					++LoadStoreSequence;
					}
				if ( theInstruction->Operation == OpLoad )
					++LoadStoreSequence;

				Buffer.Insert(theInstruction);		
				if ( ! GarbageCollector.Insert(theInstruction) )
					{
					FatalError("No room in instruction garbage collector");
					}
				theInstruction->CurrentProcessorStage = this;

				if ( (theInstruction->ProgramCounter + sizeof(uint32))== theInstruction->NextProgramCounter )
					theInstruction->RealBranchOutcome = NotTaken;
				else
					theInstruction->RealBranchOutcome = Taken;

				//////////////////////////////////////////////////////////////////////////////////////////////
				// figure out what the branch predictor says
				if ( (theInstruction->Operation == OpBranch) || (theInstruction->Operation == OpIndirectBranch) )
					{
					uint64	PredictedPC	=	0;

					/// comment from here down to to make perfect branch prediction
					if ( theInstruction->Operation == OpBranch )
						{
						// what does our predictor say?
						theInstruction->PredictedBranchOutcome = theBranchPredictor->MakePrediction(theInstruction);
						if ( theInstruction->PredictedBranchOutcome == Taken )
							PredictedPC = theInstruction->ProgramCounter + sizeof(uint32) + theInstruction->theAlphaInstruction.BranchDisplacement;
						else
							PredictedPC = theInstruction->ProgramCounter + sizeof(uint32);
						}
					else
						{	// OpIndirectBranch
						PredictedPC = theBranchTargetBuffer->AccessBranchTargetBuffer(theInstruction->ProgramCounter, theInstruction);

						if ( ! PredictedPC )
							{
							theInstruction->PredictedBranchOutcome = NotTaken;
							PredictedPC = theInstruction->ProgramCounter + sizeof(uint32);
							}
						}

					if ( PredictedPC != theInstruction->NextProgramCounter )
						{
						// from now on, we go speculative!
						// first step is to inject a few NOP's.  These NOP's are really inserted by the
						// decode stage (where this branch prediction stuff lies in hardware.. but its
						// easier to model it here in fetch
				
						// punt on this for now.  It depends on our I-Cache bandwidth
						// switch into speculative mode (possibly)
						if ( Speculative == false )
							{
							// tell everyone that we are going speculative and that after this instruction you had better
							// be prepared to roll back
							int	Index = 0;
							while ( Index < AllStages->GetSize() )
								{
								AllStages->ElementAt(Index)->EnterSpeculativeState(Sequence);
								++Index;
								}
							theStoreBuffer->EnterSpeculativeState(LoadStoreSequence);
							theBranchPredictor->EnterSpeculativeState(Sequence);
							// tell the engine to go speculative
							theEngine->EnterSpeculativeMode(PredictedPC);
							nscLoadStoreSequence = LoadStoreSequence;
							nscStoreSequence = StoreSequence;
							Speculative = true;
							theInstruction->RollBackOnThisInstruction = true;	
							}
						// else
						//	we mis-speculated while speculative (no loss)
							theEngine->EnterSpeculativeMode(PredictedPC);
						}

					/// down to here for perfect branch prediction and uncomment the next line
					//PredictedPC = theInstruction->NextProgramCounter;
					DecodeAddress = PredictedPC;
					if ( DecodeAddress != (theInstruction->ProgramCounter + sizeof(uint32) ) )
						{
						int	Index2 = 0;
						while ( Index2 < OutstandingRequestedAddresses.GetSize() )
							{
							OutstandingRequestedAddresses.PeekAt(&aAddress, Index2);
							if ( (aAddress & AddressMask) == (PredictedPC & AddressMask) )
								break;
							++Index2;
							}

						if ( Index2 >= OutstandingRequestedAddresses.GetSize() )
							{
							FetchAddress = PredictedPC;
							}
						}
					}
				else
					DecodeAddress = DecodeAddress + sizeof(uint32);
				++Index;
				++Added;
				}
			DELETE ( MemoryTransactionMessage, mtm );
			}
		else
			{
			//printf("Out of order response Wanted " format_dec_int64 " (next fetch = " format_dec_int64 ")\n", NextMemoryTransactionId, MemoryFetchId);
			break;
			}
		}
	//printf("BufferSize: %d\n", Buffer.GetSize());
	return ( 1 );
	}

int
FetchStage	::	SimulateOneCycle()
	{
	// possibly initiate a transaction
	MemoryTransactionMessage	*mtm;

	if ( OutstandingMemoryTransactions < MaxOutstandingMemoryTransactions )
		{

		++OutstandingMemoryTransactions; 

		// generate a request for the nextfetchpc
		MemoryTransactionMessage	*mtm = NEW ( MemoryTransactionMessage );

		mtm->Address = FetchAddress & AddressMask;
		mtm->Command = Read;
		mtm->Width = sizeof(uint32) * MaxFetchPerCycle;
		mtm->Id = MemoryFetchId;
		mtm->Id2 = 0;
		mtm->Id3 = GetId();
		mtm->Speculative = false;
		mtm->SetSourceComponentId(GetId());

		OutstandingRequestedAddresses.Insert(mtm->Address);
		//printf("" format_dec_int64 ": forming ", GetDomain()->GetCurrentCycle());mtm->Dump(stdout);

		++MemoryFetchId;

		if ( ! OutgoingMemoryRequests.Insert(mtm) )
			FatalError("Fetch transaction initiation rate exceeds max?");

		FetchAddress = FetchAddress + (MaxFetchPerCycle * sizeof(uint32));
		}

	while ( OutgoingMemoryRequests.GetSize() )
		{
		OutgoingMemoryRequests.Peek(&mtm);
		//printf("trying to send " format_dec_int64 ": issueing ", GetDomain()->GetCurrentCycle());mtm->Dump(stdout);
		if ( !SendMessage(theMemorySystem, mtm) )
			break;
		//printf("Sent!\n");
		OutgoingMemoryRequests.Remove(&mtm);
		}

	return ( 1 );
	}

bool
FetchStage	::	AcquireFrom(
	Instruction	**theInstruction,
	int			OfType)
	{
	if ( Buffer.CurrentSize() == 0 )
		return ( false );
	if ( OfType != InstructionTypeAll )
		{
		Instruction	*theTempInstruction;

		Buffer.PeekAt(&theTempInstruction);
		if ( theTempInstruction->theInstructionType != OfType )
			return ( false );
		}
	Buffer.Remove(theInstruction);
	return(true);
	}

bool
FetchStage	::	PeekAt(
	Instruction	**theInstruction,
	int			OfType)
	{
	if ( Buffer.CurrentSize() == 0 )
		return ( false );
	if ( OfType != InstructionTypeAll )
		{
		Instruction	*theTempInstruction;

		Buffer.PeekAt(&theTempInstruction);
		if ( theTempInstruction->theInstructionType != OfType )
			return ( false );
		}
	Buffer.PeekAt(theInstruction);
	return(true);
	}

bool
FetchStage	::	FreeInstruction(
	Instruction	*aInstruction)
	{
	Instruction	*theInstruction = NULL;

	GarbageCollector.Remove(&theInstruction);

	if ( theInstruction != aInstruction )
		{
		FatalError("Garbage collector mis-match.  Instructions must be completed/squashed in fetch order");
		}

	if ( theInstruction->Operation == OpBranch )
		theBranchPredictor->UpdateState(theInstruction);
	if ( theInstruction->Operation == OpIndirectBranch )
		theBranchTargetBuffer->UpdateBranchTargetBuffer(theInstruction->ProgramCounter, theInstruction->NextProgramCounter, theInstruction);

	if ( theInstruction->RollBackOnThisInstruction )
		{
		// need to roll back!
		int	Index = 0;
		while ( Index < AllStages->GetSize() )
			{
			AllStages->ElementAt(Index)->FlushSpeculativeState();
			++Index;
			}
		theStoreBuffer->FlushSpeculativeState();
		theBranchPredictor->FlushSpeculativeState();
		// tell the engine to go non-speculative
		theEngine->ReturnToNonSpeculativeMode();
		Speculative = false;
		FetchAddress = theInstruction->NextProgramCounter;
		}

	DELETE(Instruction, theInstruction);
	return ( true );
	}

Instruction	*
FetchStage	::	LocateInstruction(
	uint64	Sequence)
	{
	int	Size = GarbageCollector.GetSize();
	if ( ! Size )
		return ( NULL );

	Instruction	*theInstruction;

	GarbageCollector.PeekAt(&theInstruction, 0);

	uint64	Head = theInstruction->Sequence;

	int	Index = (int)(Sequence - Head);

	if ( ! GarbageCollector.PeekAt(&theInstruction, Index) )
		return ( NULL );	// hmmm

	if ( theInstruction->Sequence != Sequence )
		return ( NULL );	// hmmm

	return ( theInstruction );	
	}

Instruction	*
FetchStage	::	LocateProducerInstruction (
	int	Number)
	{
	int	Size = GarbageCollector.GetSize();
	if ( ! Size )
		return ( NULL );

	Instruction	*theInstruction;
	int	Index = 0;
	while ( Index < Size )
		{
		if ( !GarbageCollector.PeekAt(&theInstruction, Index) )
			return(NULL);	// hmmm

		if ( theInstruction->Output1 == Number )
			return ( theInstruction );
		if ( theInstruction->Output2 == Number )
			return ( theInstruction );
		++Index;
		}
	return ( NULL );
	}

bool
FetchStage	::	EnterSpeculativeState(
	uint64		MisSpeculatedInstruction)
	{
	nscSequence = MisSpeculatedInstruction;
	return ( true );
	}

bool
FetchStage	::	FlushSpeculativeState ()
	{
	Instruction	*theInstruction;

	Buffer.Clear();

	while ( GarbageCollector.GetSize() )
		{
		GarbageCollector.Remove(&theInstruction);
		DELETE(Instruction, theInstruction);
		}
	Sequence = nscSequence;
	LoadStoreSequence = nscLoadStoreSequence;
	StoreSequence = nscStoreSequence;
	return ( true );
	}


