
class	Engine;
class	StoreBuffer;

class	FetchStage	:
	public	ProcessorStage
	{
	public:
		int						MaxFetchPerCycle;
		int						MaxOutstandingMemoryTransactions;
		int						OutstandingMemoryTransactions;
		uint64					NextMemoryTransactionId;
		uint64					AddressMask;
		Fifo<Instruction	*>	Buffer;
		Engine					*theEngine;
		BranchPredictor			*theBranchPredictor;
		BranchTargetBuffer		*theBranchTargetBuffer;
		Component				*theMemorySystem;
		bool					EndSimulation;
		uint64					Sequence;
		uint64					nscSequence;
		uint64					StoreSequence;
		uint64					nscStoreSequence;
		uint64					LoadStoreSequence;
		uint64					nscLoadStoreSequence;
		uint64					FetchAddress;
		uint64					DecodeAddress;
		bool					Speculative;
		Fifo<Instruction	*>						EngineSlack;
		uint64										MemoryFetchId;
		Fifo<MemoryTransactionMessage			*>	OutgoingMemoryRequests;
		Fifo<uint64>								OutstandingRequestedAddresses;
		PQueueOfPtr<MemoryTransactionMessage	*>	MemoryTransactions;

		Array<ProcessorStage	*>	*AllStages;	
		StoreBuffer				*theStoreBuffer;
		// a garbage collector is a bit over-kill, but it makes for a useful debugging
		// facility, plus mis-speculation roll-back is easier with it.  Unlike a real
		// garbage collector there is no mark and sweep though.. this is just a repository
		// where instruction records are stored as they are fetched and blown away when
		// they are completed.
		int						GarbageCollectorSize;
		Fifo<Instruction	*>	GarbageCollector;

		virtual		~FetchStage() { };

		static	bool	BindParameters(FetchStage	*This, TupleSpace	*theTupleSpace, char	*aName = NULL);
	
		bool		BindStructures(Engine				*aEngine,
								   BranchPredictor		*aBranchPredictor,	
								   BranchTargetBuffer	*aBranchTargetBuffer,
								   Component			*aMemorySystem,
								   Array<ProcessorStage *>	*aAllStages,
								   StoreBuffer				*aStoreBuffer);

		int			AsyncSimulateOneCycle();
		int			SimulateOneCycle();
	
		bool		AcquireFrom(Instruction	**, int	OfType = InstructionTypeAll);
		bool		PeekAt(Instruction	**, int	OfType = InstructionTypeAll);

		bool		FreeInstruction(Instruction *);
		Instruction	*LocateInstruction(uint64);
		Instruction	*LocateProducerInstruction(int);

		bool		EnterSpeculativeState(uint64	MisSpeculatedInstruction);
		bool		FlushSpeculativeState();
	
		bool		SendRequestForAddress(uint64);	
		bool		ReceiveMessage(Component *f, Message	*m);
		int			WhenCanYouTakeThisMessage(Message *, Port *) { return(0); }
	};

