#	include	"Standard.h"



bool
IssueWindowStage	::	BindParameters(IssueWindowStage	*This, TupleSpace	*theTupleSpace, char	*aName )
	{
	theTupleSpace->StartSubSpace(DefaultPass(aName, "IssueWindowStage"));
	SetComponentName(This, theTupleSpace);
	theTupleSpace->Bind(Parameter(This, IssueWindowSize), "PhysicalRegisterFileSize", 128);
	theTupleSpace->Bind(Parameter(This, MaxDispatchPerCycle), "MachineWidth", 4);
	theTupleSpace->EndSubSpace();
	return(true);
	}

bool
IssueWindowStage	::	BindStructures(
			FetchStage	*aFetchStage,
			RenameStage	*aRenameStage,
			StoreBuffer	*aStoreBuffer,
			CompletionStage	*aCompletionStage)	
	{
	PhysicalRegisterFileSize = aRenameStage->PhysicalRegisterFileSize;
	InstructionsInFlight = 0;
	theFetchStage = aFetchStage;
	theRenameStage = aRenameStage;
	theStoreBuffer = aStoreBuffer;
	theCompletionStage = aCompletionStage;
	NotReadyToFire = new Array<Instruction *>[PhysicalRegisterFileSize];
	// update our local (potentially inconsistent) copy of the register state
	int	Index = 0;
	while ( Index < ArchRegisterSize )
		{
		RegisterValidStates.Add(1);
		++Index;
		}
	while ( Index < PhysicalRegisterFileSize )
		{
		RegisterValidStates.Add(0);
		++Index;
		}
	return ( true );
	}


bool
IssueWindowStage	::	PrintDebuggingInformation(FILE	*fp, int	Level)
	{
	PrintValue(HumanFormat, fp, "InstructionsInFlight", InstructionsInFlight);
	PrintValue(HumanFormat, fp, "ReadyToFire-Size", ReadyToFire.GetSize());
	return(true);
	}


int
IssueWindowStage	::	SimulateOneCycle()
	{
	int		Dispatched;
	bool	FireAway;

	Dispatched = 0;
	while ( (Dispatched < MaxDispatchPerCycle) && (InstructionsInFlight < IssueWindowSize) )
		{
		Instruction	*theInstruction;
		if ( ! UpStream->PeekAt(&theInstruction) )
			break;		// there are no instructions to place in issue window

		UpStream->AcquireFrom(&theInstruction);
		++InstructionsInFlight;

		theInstruction->CurrentProcessorStage = this;

		FireAway = true;

		WATCHFOR({printf("Processing ready list: ");Debugger::DumpInstruction(theInstruction, theRenameStage);Debugger::PrintDependencies(2, theInstruction, theFetchStage, theRenameStage);})

		if(!IsRegisterValid(theInstruction->Input1))
			{
			WATCHFOR(printf("Waiting for %d\n", theInstruction->Input1));
			NotReadyToFire[theInstruction->Input1].Add(theInstruction);
			FireAway = false;
			}
		if(!IsRegisterValid(theInstruction->Input2))
			{
			WATCHFOR(printf("Waiting for %d\n", theInstruction->Input2));
			NotReadyToFire[theInstruction->Input2].Add(theInstruction);
			FireAway = false;
			}
		if(!IsRegisterValid(theInstruction->Input3))
			{
			WATCHFOR(printf("Waiting for %d\n", theInstruction->Input3));
			NotReadyToFire[theInstruction->Input3].Add(theInstruction);
			FireAway = false;
			}
		if(FireAway)
			{
			WATCHFOR(printf("NO waiting -- firing\n"));
			theInstruction->InFireQueue = 1;
			ReadyToFire.Push(theInstruction);
			}
		else
			theInstruction->InFireQueue = 0;
		++Dispatched;
		}
	return ( Dispatched );
	}

bool
IssueWindowStage	::	AcquireFrom(
	Instruction	**aInstruction,
	int			OfType)
	{
	(*aInstruction) = NULL;

	if ( ReadyToFire.GetSize() )
		{
		(*aInstruction) = ReadyToFire.Peek();
		if ( ! theCompletionStage->ReserveSpaceFor(*aInstruction) )
			return ( false );
		if ( (*aInstruction)->theInstructionType == InstructionTypeMemory )
		if ( ! theStoreBuffer->ReserveSpaceFor(*aInstruction) )
			return ( false );
		--InstructionsInFlight;
		(*aInstruction) = ReadyToFire.Pop();
		return ( true );
		}
	return(false);
	}

bool
IssueWindowStage	::	PeekAt(
	Instruction	**aInstruction,
	int			OfType)
	{
	(*aInstruction) = NULL;

	if ( ReadyToFire.GetSize() )
		{
		(*aInstruction) = ReadyToFire.Peek();
		
		if ( ! theCompletionStage->ReserveSpaceFor(*aInstruction) )
			return(false);
		if ( (*aInstruction)->theInstructionType == InstructionTypeMemory )
		if ( ! theStoreBuffer->ReserveSpaceFor(*aInstruction) )
			return ( false );
		return ( true );
		}
	return(false);
	}

bool
IssueWindowStage	::	WriteBackRegister(
	int		Number )
	{
	// optimization: decide whether anything in the "not ready" list should get dumped into the 
	// fire away list.
	int		Index = 0;
	Instruction	*theInstruction;
	WATCHFORREG(Number, printf("Validating %d\n", Number));

	if ( (Number == ZeroRegister) || (Number == NoRegister) )
		return(true);

	RegisterValidStates[Number] = 1;
	while ( Index < NotReadyToFire[Number].GetSize() )
		{
		theInstruction = NotReadyToFire[Number].ElementAt(Index);
		if ( theInstruction->InFireQueue == 0 )
			{
			if( IsRegisterValid(theInstruction->Input1) )
			if( IsRegisterValid(theInstruction->Input2) )
			if( IsRegisterValid(theInstruction->Input3) )
				{
				theInstruction->InFireQueue = 1;
				WATCHFOR(printf("Migrated to Firing queue on %d\n", Number));
				//printf("Waking up: %d on %d\n", theInstruction->Sequence, Number);
				ReadyToFire.Push(theInstruction);
				}
			}
		++Index;
		}
	NotReadyToFire[Number].Clear();
	return(1);
	}

/*
bool
IssueWindowStage	::	IsRegisterValid(
	int		Number)
	{
	if ( (Number == ZeroRegister) )		// zero is always valid
		return ( true );
	if ( (Number == NoRegister) )		// the "no register" is always valid
		return ( true );

	if ( RegisterValidStates[Number] )
		return ( true );

	return ( false );
	}
*/

bool
IssueWindowStage	::	NotifyAboutRegisterRename (
	int		Number)
	{
	//WriteBackRegister(Number);
	WATCHFORREG(Number, printf("Invalidating %d\n", Number));
	RegisterValidStates[Number] = 0;
	return(true);
	}

bool
IssueWindowStage	::	EnterSpeculativeState (
	uint64		aMisSpeculatedInstruction)
	{
	MisSpeculatedInstruction = aMisSpeculatedInstruction;
	return(true);
	}

bool
IssueWindowStage	::	FlushSpeculativeState ()
	{
	InstructionsInFlight = 0;

	ReadyToFire.Clear();

	for(int	Index =0; Index < PhysicalRegisterFileSize; Index++ )
		NotReadyToFire[Index].Clear();
	for(int	Index =0; Index < PhysicalRegisterFileSize; Index++ )
		if ( theRenameStage->nscIsRegisterValid(Index) )
			RegisterValidStates[Index] = 1;
		else
			RegisterValidStates[Index] = 0;
	return ( true );
	}


