#	include	"Standard.h"



bool
BypassingRegisterAccessStage	::	BindParameters(BypassingRegisterAccessStage	*This, TupleSpace	*theTupleSpace, char	*aName )
	{
	theTupleSpace->StartSubSpace(DefaultPass(aName, "BypassingRegisterAccessStage"));
	SetComponentName(This, theTupleSpace);
	theTupleSpace->Bind(Parameter(This, ByPassListSize), "MachineWidth", 4);
	theTupleSpace->Bind(Parameter(This, MaxDispatchPerCycle), "MachineWidth", 4);
	theTupleSpace->Bind(Parameter(This, InstructionsInFlight), "MachineWidth", 4);
	theTupleSpace->EndSubSpace();
	return(true);
	}

bool
BypassingRegisterAccessStage	::	BindStructures(
						RenameStage	*aRenameStage,
						IssueWindowStage	*aIssueWindowStage,
						CompletionStage		*aCompletionStage)
	{
	InstructionsInFlight = 0;
	theRenameStage = aRenameStage;
	theIssueWindowStage = aIssueWindowStage;
	theCompletionStage = aCompletionStage;
	// this stage uses the same complex resolving stuff as the issue window.
	// but this is only for efficiency.  In practice, ByPassListSize := MachineWidth,
	// so this is not another issue window...
	NotReadyToFire = new Array<Instruction *>[PhysicalRegisterFileSize];
	return ( true );
	}

bool
BypassingRegisterAccessStage	::	PrintDebuggingInformation(
						FILE	*fp,
						int		Level)
	{
	fprintf(fp, "%s: InstructionsInFlight := %d\n", "RegisterAccessStage", InstructionsInFlight);
	for(int	Index=0;Index<InstructionTypes;Index++)
		fprintf(fp, "%s: ReadyToFire[%d] := %d\n", "RegisterAccessStage", Index, ReadyToFire[Index].GetSize());
	fprintf(fp, "%s: NotificationList := %d\n", "RegisterAccessStage", NotificationList.GetSize());
	
	return(true);
	}

int
BypassingRegisterAccessStage	::	SimulateOneCycle()
	{
	int		Dispatched;
	bool	FireAway;

	Dispatched = 0;
	while ( (Dispatched < MaxDispatchPerCycle) && (InstructionsInFlight < ByPassListSize) )
		{
		Instruction	*theInstruction;
		if ( ! UpStream->AcquireFrom(&theInstruction) )
			break;
		if ( theInstruction->IsNOP() )
			{
			if(theCompletionStage->DispatchTo(theInstruction))
				{
				++Dispatched;
				continue;
				}
			// we can fail.  the Completion buffer may exceed some limit on comletions/cycle.
			// in which case, were just going to dump it into the execution core.  thats the breaks.
			}


		theInstruction->CurrentProcessorStage = this;

		++InstructionsInFlight;

		FireAway = true;
		//printf("Processing ready list: ");//debug.DumpInstruction(theInstruction);
		if(!theRenameStage->IsRegisterValid(theInstruction->Input1))
			{
			//printf("Waiting for %d\n", theInstruction->Input1);
			NotReadyToFire[theInstruction->Input1].Add(theInstruction);
			FireAway = false;
			}
		if(!theRenameStage->IsRegisterValid(theInstruction->Input2))
			{
			//printf("Waiting for %d\n", theInstruction->Input2);
			NotReadyToFire[theInstruction->Input2].Add(theInstruction);
			FireAway = false;
			}
		if(!theRenameStage->IsRegisterValid(theInstruction->Input3))
			{
			//printf("Waiting for %d\n", theInstruction->Input3);
			NotReadyToFire[theInstruction->Input3].Add(theInstruction);
			FireAway = false;
			}
		if(FireAway)
			{
			//printf("NO waiting -- firing\n");
			theInstruction->InFireQueue = 1;
			int	OfType = theInstruction->theInstructionType;
			if ( OfType == InstructionTypeMemory )
				OfType = InstructionTypeInteger;
			ReadyToFire[OfType].Push(theInstruction);
			}
		else
			theInstruction->InFireQueue = 0;

		// best to leave instructiontypes memory to the store buffer to notify.
		if ( theInstruction->theInstructionType != InstructionTypeMemory )
			{
			RegisterWriteUpdate	wru;

			wru.UpdateAt = theInstruction->Delay - 1;
			wru.Register1 = theInstruction->Output1;
			wru.Register2 = theInstruction->Output2;
			NotificationList.Add(wru);
			}

		++Dispatched;
		}
	// now tell the issue window that the registers are available.  This is a lie,
	// since they wont be, but *hopefully* they will be..... essentially this models
	// forwarding networks.
	int	Index;
	Index = NotificationList.GetSize();
	while ( Index > 0 )
		{
		--Index;
		RegisterWriteUpdate	wru = NotificationList[Index];
		if ( wru.UpdateAt <= 0 )	
			{
			theIssueWindowStage->WriteBackRegister(wru.Register1);
			theIssueWindowStage->WriteBackRegister(wru.Register2);
			NotificationList.RemoveAt(Index);
			}
		else
			--NotificationList[Index].UpdateAt;
		}
	return ( Dispatched );
	}

bool
BypassingRegisterAccessStage	::	AcquireFrom(
	Instruction	**aInstruction,
	int			OfType)
	{
	(*aInstruction) = NULL;

	if ( OfType != InstructionTypeAll )
		{
		if ( OfType == InstructionTypeMemory )
			OfType = InstructionTypeInteger;

		if ( ReadyToFire[OfType].GetSize() )
			{
			--InstructionsInFlight;
			(*aInstruction) = ReadyToFire[OfType].Pop();
			return ( true );
			}
		return(false);
		}
	// issue the oldest
	int	Minimum = -1;
	uint64	MinimumSequence = (cuint64(0xffffffff)<<32)|0xffffffff;
	uint64	TestSequence;
	int	Index = 0;
	while ( Index < InstructionTypes )
		{
		if ( ReadyToFire[Index].GetSize() )
			{
			if ( Minimum < 0 )
				{
				Minimum = Index;
				MinimumSequence = ReadyToFire[Index].Peek()->Sequence;
				}
			else
				{
				TestSequence = ReadyToFire[Index].Peek()->Sequence;
				if ( TestSequence < MinimumSequence )
					{
					MinimumSequence = TestSequence;
					Minimum = Index;
					}
				}
			}
		++Index;
		}
	if ( Minimum >= 0 )
		{
		(*aInstruction) = ReadyToFire[Minimum].Pop();
		--InstructionsInFlight;
		//printf("Acquiring returning true!\n");
		return ( true );
		}
	return ( false );
	}

bool
BypassingRegisterAccessStage	::	PeekAt(
	Instruction	**aInstruction,
	int			OfType)
	{
	(*aInstruction) = NULL;

	if ( OfType != InstructionTypeAll )
		{
		if ( OfType == InstructionTypeMemory )
			OfType = InstructionTypeInteger;

		if ( ReadyToFire[OfType].GetSize() )
			{
			(*aInstruction) = ReadyToFire[OfType].Peek();
			return ( true );
			}
		return(false);
		}
	// issue the oldest
	int	Minimum = -1;
	uint64	MinimumSequence = 0;
	uint64	TestSequence;
	int	Index = 0;
	while ( Index < InstructionTypes )
		{
		if ( ReadyToFire[Index].GetSize() )
			{
			if ( Minimum < 0 )
				{
				Minimum = Index;
				MinimumSequence = ReadyToFire[Index].Peek()->Sequence;
				}
			else
				{
				TestSequence = ReadyToFire[Index].Peek()->Sequence;
				if ( TestSequence < MinimumSequence )
					{
					MinimumSequence = TestSequence;
					Minimum = Index;
					}
				}
			}
		++Index;
		}
	if ( Minimum >= 0 )
		{
		(*aInstruction) = ReadyToFire[Minimum].Peek();
		//printf("Peeking returning true!\n");
		return ( true );
		}
	return ( false );
	}

bool
BypassingRegisterAccessStage	::	WriteBackRegister(
	int		Number )
	{
	// optimization: decide whether anything in the "not ready" list should get dumped into the 
	// fire away list.
	int		Index = 0;
	Instruction	*theInstruction;

	//printf("WriteBackRegister %d\n", Number);

	if ( Number == ZeroRegister || Number == NoRegister )
		return(true);

	while ( Index < NotReadyToFire[Number].GetSize() )
		{
		theInstruction = NotReadyToFire[Number].ElementAt(Index);
		if ( theInstruction->InFireQueue == 0 )
			{
			if( theRenameStage->IsRegisterValid(theInstruction->Input1) )
			if( theRenameStage->IsRegisterValid(theInstruction->Input2) )
			if( theRenameStage->IsRegisterValid(theInstruction->Input3) )
				{
				theInstruction->InFireQueue = 1;
				//printf("Waking up: " format_dec_int64 " on %d\n", theInstruction->Sequence, Number);
				int	OfType = theInstruction->theInstructionType;
				if ( OfType == InstructionTypeMemory )
					OfType = InstructionTypeInteger;
				ReadyToFire[OfType].Push(theInstruction);
				}
			}
		++Index;
		}
	NotReadyToFire[Number].Clear();

	return(1);
	}

bool
BypassingRegisterAccessStage	::	FlushSpeculativeState ()
	{
	InstructionsInFlight = 0;
	for(int	Index =0; Index < InstructionTypes; Index++ )
		ReadyToFire[Index].Clear();
	for(int	Index =0; Index < PhysicalRegisterFileSize; Index++ )
		NotReadyToFire[Index].Clear();
	NotificationList.Clear();
	return ( true );
	}



bool
NonBypassingRegisterAccessStage	::	BindParameters(NonBypassingRegisterAccessStage	*This, TupleSpace	*theTupleSpace, char	*aName )
	{
	theTupleSpace->StartSubSpace(DefaultPass(aName, "NonBypassingRegisterAccessStage"));
	SetComponentName(This, theTupleSpace);
	theTupleSpace->Bind(Parameter(This, MaxDispatchPerCycle), "MachineWidth", 4);
	theTupleSpace->Bind(Parameter(This, MaxInstructionsInFlight), "MachineWidth", 4);
	theTupleSpace->EndSubSpace();
	return(true);
	}

bool
NonBypassingRegisterAccessStage	::	BindStructures(
						RenameStage	*aRenameStage,
						IssueWindowStage	*aIssueWindowStage,
						CompletionStage		*aCompletionStage)
	{
	theRenameStage = aRenameStage;
	theIssueWindowStage = aIssueWindowStage;
	theCompletionStage = aCompletionStage;
	InstructionsInFlight = 0;
	for(int Index=0;Index<InstructionTypes;Index++)
		ReadyToFire[Index].Resize(MaxInstructionsInFlight);
	return(true);
	}

int
NonBypassingRegisterAccessStage	::	SimulateOneCycle()
	{
	int	Dispatched = 0;

	while ( (Dispatched < MaxDispatchPerCycle ) && (InstructionsInFlight < MaxInstructionsInFlight) )
		{
		Instruction	*theInstruction;

		if ( ! UpStream->AcquireFrom(&theInstruction) )
			break;
		if ( theInstruction->IsNOP() )
			{
			if(theCompletionStage->DispatchTo(theInstruction))
				{
				++Dispatched;
				continue;
				}
			// we can fail.  the Completion buffer may exceed some limit on completions/cycle.
			// in which case, were just going to dump it into the execution core.  thats the breaks.
			}

		theInstruction->CurrentProcessorStage = this;

		++InstructionsInFlight;

		int	OfType = theInstruction->theInstructionType;
		if ( OfType == InstructionTypeMemory )
			OfType = InstructionTypeInteger;
		ReadyToFire[OfType].Insert(theInstruction);

		if ( theInstruction->theInstructionType != InstructionTypeMemory )
			{
			RegisterWriteUpdate	wru;
			wru.UpdateAt = theInstruction->Delay - 1;
			wru.Register1 = theInstruction->Output1;
			wru.Register2 = theInstruction->Output2;
			NotificationList.Add(wru);
			}

		++Dispatched;
		}
	// now tell the issue window that the registers are available.  This is a lie,
	// since they wont be, but *hopefully* they will be..... essentially this models
	// forwarding networks.
	int	Index;
	Index = NotificationList.GetSize();
	while ( Index > 0 )
		{
		--Index;
		RegisterWriteUpdate	wru = NotificationList[Index];
		if ( wru.UpdateAt <= 0 )	
			{
			theIssueWindowStage->WriteBackRegister(wru.Register1);
			theIssueWindowStage->WriteBackRegister(wru.Register2);
			NotificationList.RemoveAt(Index);
			}
		else
			--NotificationList[Index].UpdateAt;
		}
	return ( Dispatched );
	}


bool
NonBypassingRegisterAccessStage	::	AcquireFrom(
	Instruction	**aInstruction,
	int			OfType)
	{
	(*aInstruction) = NULL;

	if ( OfType != InstructionTypeAll )
		{
		if ( OfType == InstructionTypeMemory )
			OfType = InstructionTypeInteger;
		if ( ReadyToFire[OfType].GetSize() )
			{
			--InstructionsInFlight;
			ReadyToFire[OfType].Remove(aInstruction);
			return ( true );
			}
		return(false);
		}
	// issue the oldest
	int	Minimum = -1;
	uint64	MinimumSequence = 0;
	uint64	TestSequence;
	int	Index = 0;
	while ( Index < InstructionTypes )
		{
		if ( ReadyToFire[Index].GetSize() )
			{
			if ( Minimum < 0 )
				{
				Minimum = Index;
				ReadyToFire[Index].Peek(aInstruction);
				MinimumSequence = (*aInstruction)->Sequence;
				}
			else
				{
				ReadyToFire[Index].Peek(aInstruction);
				TestSequence = (*aInstruction)->Sequence;
				if ( TestSequence < MinimumSequence )
					{
					MinimumSequence = TestSequence;
					Minimum = Index;
					}
				}
			}
		++Index;
		}
	if ( Minimum >= 0 )
		{
		ReadyToFire[Minimum].Remove(aInstruction);
		--InstructionsInFlight;
		//printf("Acquiring returning true!\n");
		return ( true );
		}
	return ( false );
	}

bool
NonBypassingRegisterAccessStage	::	PeekAt(
	Instruction	**aInstruction,
	int			OfType)
	{
	(*aInstruction) = NULL;

	if ( OfType != InstructionTypeAll )
		{
		if ( OfType == InstructionTypeMemory )
			OfType = InstructionTypeInteger;
		if ( ReadyToFire[OfType].GetSize() )
			{
			ReadyToFire[OfType].Peek(aInstruction);
			return ( true );
			}
		return(false);
		}
	// issue the oldest
	int	Minimum = -1;
	uint64	MinimumSequence = 0;
	uint64	TestSequence;
	int	Index = 0;
	while ( Index < InstructionTypes )
		{
		if ( ReadyToFire[Index].GetSize() )
			{
			if ( Minimum < 0 )
				{
				Minimum = Index;
				ReadyToFire[Index].Peek(aInstruction);
				MinimumSequence = (*aInstruction)->Sequence;
				}
			else
				{
				ReadyToFire[Index].Peek(aInstruction);
				TestSequence = (*aInstruction)->Sequence;
				if ( TestSequence < MinimumSequence )
					{
					MinimumSequence = TestSequence;
					Minimum = Index;
					}
				}
			}
		++Index;
		}
	if ( Minimum >= 0 )
		{
		ReadyToFire[Minimum].Peek(aInstruction);
		//printf("Peeking returning true!\n");
		return ( true );
		}
	return ( false );
	}


bool
NonBypassingRegisterAccessStage	::	WriteBackRegister(
	int	Number)
	{
	return(true);
	}


bool
NonBypassingRegisterAccessStage	::	FlushSpeculativeState()
	{
	for(int	Index=0;Index < InstructionTypes;Index++)
		ReadyToFire[Index].Clear();
	InstructionsInFlight = 0;
	NotificationList.Clear();
	return(true);
	}

