#	include	"Standard.h"

DEFINE_ALLOCATOR(PhysicalRegister);



bool
RenameStage	::	BindParameters(RenameStage	*This, TupleSpace	*theTupleSpace, char	*aName )
	{
	theTupleSpace->StartSubSpace(DefaultPass(aName, "RenameStage"));
	SetComponentName(This, theTupleSpace);
	theTupleSpace->Bind(Parameter(This, PhysicalRegisterFileSize), 128);
	theTupleSpace->Bind(Parameter(This, MaxDecodePerCycle), "MachineWidth", 4);
	theTupleSpace->EndSubSpace();
	return(true);
	}

bool
RenameStage	::	BindStructures(FetchStage	*aFetchStage,
							IssueWindowStage	*aIssueWindowStage)
	{
	int		Index = 0;

	theFetchStage = aFetchStage;
	theIssueWindowStage = aIssueWindowStage;

	FreeList.Resize(PhysicalRegisterFileSize);
	DecodedInstructions.Resize(MaxDecodePerCycle);

	WillGoSpeculative = false;
	nscRegisterMap = new int[PhysicalRegisterFileSize];

	while ( Index < ArchRegisterSize )
		{
		PhysicalRegister	*pr = NEW(PhysicalRegister);

		pr->Number = Index;
		pr->Valid = 1;
		pr->PendingUses = 0;
		pr->PartOfArchitectureState = true;
		pr->ArchitectureRegister = Index;
		pr->IsFree = false;
		ArchitectureRegisterFile.Add(pr);
		PhysicalRegisterFile.Add(pr);
		++Index;
		}

	while ( Index < PhysicalRegisterFileSize )
		{
		PhysicalRegister	*pr = NEW(PhysicalRegister);

		pr->Number = Index;
		pr->Valid = 0;
		pr->PendingUses = 0;
		pr->PartOfArchitectureState = false;
		pr->IsFree = true;

		PhysicalRegisterFile.Add(pr);
		FreeList.Insert(pr);

		++Index;
		}
	return(true);
	}

int
RenameStage	::	SimulateOneCycle()
	{
	int			Renamed;

	Renamed = 0;
	while ( Renamed < MaxDecodePerCycle )
		{
		if ( DecodedInstructions.FreeSpace() == 0 )
			break;

		// must always be 2 free output registers
		if ( FreeList.CurrentSize() < 2 )
			break;

		Instruction	*theInstruction;

		if ( !UpStream->AcquireFrom(&theInstruction) )
			break;	// no more instructions

		theInstruction->CurrentProcessorStage = this;
		if ( WillGoSpeculative )
		if ( WatchForThisSpeculativeInstruction == theInstruction->Sequence )
			{
			// this is the first mis-speculated instruction (golden boy)
			// make a copy of our register map
			for(int	Index = 0; Index < PhysicalRegisterFileSize; Index++)
				{
				if (PhysicalRegisterFile[Index]->PartOfArchitectureState )
					nscRegisterMap[Index] = PhysicalRegisterFile[Index]->ArchitectureRegister;
				else
					nscRegisterMap[Index] = -1;
				}
			WillGoSpeculative = false;
			}

		//printf("rename-b: ");debug.DumpInstruction(theInstruction);
		// now rename instruction
		
		// start with inputs
		if ( (theInstruction->Input1 != NoRegister)&&(theInstruction->Input1 != ZeroRegister) )
			{
			theInstruction->Input1 = ArchitectureRegisterFile[theInstruction->Input1]->Number;
			++PhysicalRegisterFile[theInstruction->Input1]->PendingUses;
			}
		if ( (theInstruction->Input2 != NoRegister)&&(theInstruction->Input2 != ZeroRegister) )
			{
			theInstruction->Input2 = ArchitectureRegisterFile[theInstruction->Input2]->Number;
			++PhysicalRegisterFile[theInstruction->Input2]->PendingUses;
			}
		if ( (theInstruction->Input3 != NoRegister)&&(theInstruction->Input3 != ZeroRegister) )
			{
			theInstruction->Input3 = ArchitectureRegisterFile[theInstruction->Input3]->Number;
			++PhysicalRegisterFile[theInstruction->Input3]->PendingUses;
			}

		// rename outputs

		if ( (theInstruction->Output1 != NoRegister)&&(theInstruction->Output1 != ZeroRegister) )
			{
			// acquire something from the free list
			PhysicalRegister	*pr;

			theInstruction->AddToFreeList1 = ArchitectureRegisterFile[theInstruction->Output1]->Number;

			// remove current output name from architectural state
			ArchitectureRegisterFile[theInstruction->Output1]->PartOfArchitectureState = false;

			/* This code is only used for reference counting implementations
			// possibly send it to the free list if no one wants it
			if ( ArchitectureRegisterFile[theInstruction->Output1]->PendingUses == 0 )
				{
				// dump it onto the free list
				ArchitectureRegisterFile[theInstruction->Output1]->Valid = 0;
				if(!ArchitectureRegisterFile[theInstruction->Output1]->IsFree)
					{
					ArchitectureRegisterFile[theInstruction->Output1]->IsFree = true;
					FreeList.Insert(ArchitectureRegisterFile[theInstruction->Output1]);
					}
				}
			*/
			FreeList.Remove(&pr);
			pr->Valid = 0;
			pr->IsFree = false;
			pr->PendingUses = 1;
			// make new register part of architectural state
			ArchitectureRegisterFile[theInstruction->Output1] = pr;
			pr->PartOfArchitectureState = true;
			pr->ArchitectureRegister = theInstruction->Output1;

			theIssueWindowStage->NotifyAboutRegisterRename(pr->Number);
	
			// rename
			theInstruction->Output1 = pr->Number;
			}

		if ( (theInstruction->Output2 != NoRegister)&&(theInstruction->Output2 != ZeroRegister) )
			{
			// acquire something from the free list
			PhysicalRegister	*pr;

			theInstruction->AddToFreeList2 = ArchitectureRegisterFile[theInstruction->Output2]->Number;

			// remove current output name from architectural state
			ArchitectureRegisterFile[theInstruction->Output2]->PartOfArchitectureState = false;

			/* This code is only used for reference counting implementations
			// possibly send it to the free list if no one wants it
			if ( ArchitectureRegisterFile[theInstruction->Output2]->PendingUses == 0 )
				{
				// dump it onto the free list
				ArchitectureRegisterFile[theInstruction->Output2]->Valid = 0;
				if(!ArchitectureRegisterFile[theInstruction->Output2]->IsFree)
					{
					ArchitectureRegisterFile[theInstruction->Output2]->IsFree = true;
					FreeList.Insert(ArchitectureRegisterFile[theInstruction->Output2]);
					}
				}
			*/

			FreeList.Remove(&pr);
			pr->Valid = 0;
			pr->IsFree = false;
			pr->PendingUses = 1;
			// make new register part of architectural state
			ArchitectureRegisterFile[theInstruction->Output2] = pr;
			pr->PartOfArchitectureState = true;
			pr->ArchitectureRegister = theInstruction->Output2;

			theIssueWindowStage->NotifyAboutRegisterRename(pr->Number);
			// rename
			theInstruction->Output2 = pr->Number;
			}

		//printf("rename-a: ");debug.DumpInstruction(theInstruction);
		DecodedInstructions.Insert(theInstruction);

		++Renamed;
		}
	return(Renamed);
	}

bool
RenameStage	::	AcquireFrom(
	Instruction	**theInstruction,
	int			OfType)
	{
	if ( DecodedInstructions.CurrentSize() == 0 )
		return ( false );
	
	DecodedInstructions.Remove(theInstruction);
	return(true);
	}

bool
RenameStage	::	PeekAt(
	Instruction	**theInstruction,
	int			OfType)
	{
	if ( DecodedInstructions.CurrentSize() == 0 )
		return ( false );

	DecodedInstructions.Peek(theInstruction);
	return( true );
	}

bool
RenameStage	::	AddToFreeList(
	int		Number)
	{
	if ( Number >= 0 )
		FreeList.Insert(PhysicalRegisterFile[Number]);
	return(true);
	}

bool
RenameStage	::	WriteBackRegister(
	int		Number)
	{
	// don't care about zero or invalid registers
	if( (Number == ZeroRegister)||(Number == NoRegister))
		return(true);

	// the write back is made valid
	PhysicalRegisterFile[Number]->Valid = 1;
/* Only for reference counting
	// what if no one cares?
	if ( PhysicalRegisterFile[Number]->PendingUses == 0 )			// no one is using it
	if ( ! PhysicalRegisterFile[Number]->PartOfArchitectureState )	// no longer part of architectural state
		{
		// hmm?
		// invalidate the register and send it to the free list
		PhysicalRegisterFile[Number]->Valid = 0;
		if(!PhysicalRegisterFile[Number]->IsFree)
			{
			PhysicalRegisterFile[Number]->IsFree = true;
			FreeList.Insert(PhysicalRegisterFile[Number]);
			}
		}
*/
	return ( true );
	}

bool
RenameStage	::	NotifyAboutRegisterUse(
	int		Number )
	{
	if( (Number == ZeroRegister)||(Number == NoRegister))
		return(true);

	// decrease the usage count
	--PhysicalRegisterFile[Number]->PendingUses;
/* Only for reference counting

	// what if no one cares?
	if ( PhysicalRegisterFile[Number]->PendingUses == 0 )			// no one is using it
	if ( ! PhysicalRegisterFile[Number]->PartOfArchitectureState )	// no longer part of architectural state
		{
		// hmm?
		// invalidate the register and send it to the free list
		PhysicalRegisterFile[Number]->Valid = 0;
		if(!PhysicalRegisterFile[Number]->IsFree)
			{
			PhysicalRegisterFile[Number]->IsFree = true;
			FreeList.Insert(PhysicalRegisterFile[Number]);
			}
		}
*/
	return ( true );
	}
/*
bool
RenameStage	::	IsRegisterValid(
	int		Number )
	{
	if ( (Number == ZeroRegister) )		// zero is always valid
		return ( true );
	if ( (Number == NoRegister) )		// the "no register" is always valid
		return ( true );

	if ( PhysicalRegisterFile[Number]->Valid )	// are we valid?
		return ( true );

	return ( false );
	}
*/

bool
RenameStage	::	nscIsRegisterValid(
	int		Number)
	{
	if ( (Number == ZeroRegister) )		// zero is always valid
		return ( true );
	if ( (Number == NoRegister) )		// the "no register" is always valid
		return ( true );

	if ( nscRegisterMap[Number] >= 0 )
		return ( true );
	return ( false );
	}

bool
RenameStage	::	EnterSpeculativeState(uint64	MisSpeculatedInstruction)
	{
	WillGoSpeculative = true;
	WatchForThisSpeculativeInstruction = MisSpeculatedInstruction;


	return ( true );
	}

bool
RenameStage	::	FlushSpeculativeState()
	{
	if ( WillGoSpeculative )
		{
		// we never saw a bad instruction ... but, our nscMap is now bogus.  Make it!

		// grab the current register map -- just in case we never see a speculative instruction
		// and need to help other stages (i.e. issuewindow) roll back nicely
		for(int	Index = 0; Index < PhysicalRegisterFileSize; Index++)
			{
			if (PhysicalRegisterFile[Index]->PartOfArchitectureState )
				nscRegisterMap[Index] = PhysicalRegisterFile[Index]->ArchitectureRegister;
			else
				nscRegisterMap[Index] = -1;
			}
		WillGoSpeculative = false;
		DecodedInstructions.Clear();
		return ( true );
		}
	WillGoSpeculative = false;

	FreeList.Clear();
	DecodedInstructions.Clear();

	for(int	Index = 0;	Index < PhysicalRegisterFileSize ; Index ++ )
		{
		PhysicalRegister	*pr = PhysicalRegisterFile[Index];
		pr->PendingUses = 0;

		if ( nscRegisterMap[Index] < 0 )
			{
			pr->Valid = 0;
			pr->IsFree = true;
			pr->PartOfArchitectureState = false;
			pr->PendingUses = 0;
			FreeList.Insert(PhysicalRegisterFile[Index]);
			}
		else
			{
			pr->Valid = 1;
			pr->PendingUses = 0;
			pr->IsFree = false;
			pr->PartOfArchitectureState = true;
			pr->ArchitectureRegister = nscRegisterMap[Index];
			ArchitectureRegisterFile[nscRegisterMap[Index]] = pr;
			}
		}
	return ( true );
	}

