#	include	"Standard.h"

DEFINE_ALLOCATOR(StoreBufferEntry);


bool
StoreBuffer	::	BindParameters(StoreBuffer	*This, TupleSpace	*theTupleSpace)
	{
	theTupleSpace->StartSubSpace("StoreBuffer");
	SetComponentName(This, theTupleSpace);
	theTupleSpace->Bind(Parameter(This, StoreBufferSize), "MachineWidth", 4);
	theTupleSpace->Bind(Parameter(This, MaximumDispatchesPerCycle), "MachineWidth", 4);
	theTupleSpace->Bind(Parameter(This, MaximumRetiresPerCycle), "MachineWidth", 4);
	theTupleSpace->Bind(Parameter(This, MaximumIssuesPerCycle), "MachineWidth", 4);
	theTupleSpace->EndSubSpace();
	return(true);
	}

bool
StoreBuffer	::	BindStructures(
					RenameStage			*aRenameStage,
					IssueWindowStage	*aIssueWindowStage,
					RegisterAccessStage	*aRegisterAccessStage,
					CompletionStage		*aCompletionStage,
					Component			*aMemorySystem)
	{
	theRenameStage = aRenameStage;
	theIssueWindowStage = aIssueWindowStage;
	theRegisterAccessStage =aRegisterAccessStage;
	theCompletionStage = aCompletionStage;
	theMemorySystem = aMemorySystem;

	DispatchesThisCycle = RetiresThisCycle = IssuesThisCycle = 0;
	LoadStoreSequence = 0;
	WillGoSpeculative = false;
	WatchForThisSpeculativeLoadStoreSequence = 0;

	for(int	Index = 0; Index < StoreBufferSize ; Index++ )
		OrderingArray.Add(NULL);
	PendingStores = new HashTable(1157, false);	// some random large number (an optimization)
	ReturnBackFromMemorySystem.Resize(StoreBufferSize);
	SendToMemorySystem.Resize(StoreBufferSize);
	return ( true );
	}
		

bool
StoreBuffer	::	PrintDebuggingInformation(FILE	*fp, int	Level)
	{
	PrintValue(HumanFormat, fp, "StoreBufferSize", StoreBufferSize);
	PrintValue(HumanFormat, fp, "LoadStoreSequence", LoadStoreSequence);
	PrintValue(HumanFormat, fp, "DispatchesThisCycle", DispatchesThisCycle);
	PrintValue(HumanFormat, fp, "RetiresThisCycle", RetiresThisCycle);
	PrintValue(HumanFormat, fp, "IssuesThisCycle", IssuesThisCycle);
	PrintValue(HumanFormat, fp, "QueuedForMemory", SendToMemorySystem.GetSize());
	PrintValue(HumanFormat, fp, "WillGoSpeculative", WillGoSpeculative);
	PrintValue(HumanFormat, fp, "WatchForThisSpeculativeLoadStoreSequence", WatchForThisSpeculativeLoadStoreSequence);
	printf("StoreBuffer: ");
	for(int	Index = 0;Index < StoreBufferSize; Index++)
		{
		Instruction	*theInstruction = OrderingArray[(LoadStoreSequence+Index)%StoreBufferSize];
		if(theInstruction)
			printf("(" format_dec_int64 "," format_dec_int64 "),", theInstruction->Sequence, theInstruction->LoadStoreSequence);
		else
			printf("(--?--),");
		}
	printf("\n");
	return(true);
	}

bool
StoreBuffer	::	DispatchTo(
	Instruction	*theInstruction)
	{
	if ( MaximumDispatchesPerCycle )
	if ( DispatchesThisCycle >= MaximumDispatchesPerCycle )
		return ( false );

	if ( theInstruction->LoadStoreSequence >= (LoadStoreSequence + StoreBufferSize))
		return ( false );

	theInstruction->CurrentProcessorStage = this;

/*
	if(!theCompletionStage->DispatchTo(theInstruction))
		return(false);
	theRenameStage->WriteBackRegister(theInstruction->Output1);
	theRenameStage->WriteBackRegister(theInstruction->Output2);
	theRenameStage->NotifyAboutRegisterUse(theInstruction->Output1);
	theRenameStage->NotifyAboutRegisterUse(theInstruction->Output2);
	theIssueWindowStage->WriteBackRegister(theInstruction->Output1);
	theIssueWindowStage->WriteBackRegister(theInstruction->Output2);
	theRegisterAccessStage->WriteBackRegister(theInstruction->Output1);
	theRegisterAccessStage->WriteBackRegister(theInstruction->Output2);
	return(true);
*/

//	printf("insb: ");Debugger::DumpInstruction(theInstruction, theRenameStage);

	// we can take it. if its a load see if it can be satisfied from the
	// load store queue
	StoreBufferEntry	*sbe = NEW(StoreBufferEntry);
	sbe->theInstruction = theInstruction;
	sbe->StrictComparison = false;
	if ( theInstruction->Operation == OpLoad )
		{
		StoreBufferEntry	*match = (StoreBufferEntry*)PendingStores->LocateObject(sbe);
		if ( match )
			{
			// there is a matching store.  This load can be satisfied from the
			// store buffer.  Mark it as executed
			theInstruction->MemoryExecuted = 0;
			theInstruction->MemoryDispatched = 1;
			}
		else
			{
			theInstruction->MemoryExecuted = 0;
			theInstruction->MemoryDispatched = 0;
			}
		DELETE(StoreBufferEntry, sbe);
		}
	else
	if ( theInstruction->Operation == OpStore ) // better be
		{
		// insert it into the pending stores.  this doesn't mean its issued to
		// the memory system.  Only that we know about it
		PendingStores->InsertObject(sbe);
		theInstruction->MemoryExecuted = 0;
		theInstruction->MemoryDispatched = 0;
		}

	// insert it into the store buffer slot
	OrderingArray[theInstruction->LoadStoreSequence % StoreBufferSize ] = theInstruction;

	++DispatchesThisCycle;		
	return ( true );
	}

bool
StoreBuffer	::	ReserveSpaceFor(
	Instruction	*theInstruction)
	{
//	return(true);
	if ( theInstruction->LoadStoreSequence >= (LoadStoreSequence + StoreBufferSize))
		return ( false );
	return ( true );
	}

int
StoreBuffer	::	SimulateOneCycle ()
	{
	////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// attempt to issue any non-dispatched instructions to the memory system
	////////////////////////////////////////////////////////////////////////////////////////////////////////////
	IssuesThisCycle = 0;
	DispatchesThisCycle = 0;
	RetiresThisCycle = 0;

	int	Delta = 0;
	while ( Delta < StoreBufferSize )
		{
		if ( MaximumIssuesPerCycle )
		if ( IssuesThisCycle >= MaximumIssuesPerCycle )
			break;

		Instruction	*theInstruction = OrderingArray[(LoadStoreSequence+Delta)%StoreBufferSize];
		if ( ! theInstruction )
			break;
	
		if ( theInstruction->MemoryExecuted > 0 )
			{
			++Delta;
			continue;
			}
		else
		if ( theInstruction->MemoryDispatched )
			{
			if ( theInstruction->MemoryExecuted == 0 )	// happens when a load is satisfied from a store
				{
				theInstruction->MemoryExecuted = 1;
				++IssuesThisCycle;
				ReturnBackFromMemorySystem.Insert(theInstruction->LoadStoreSequence);

				RegisterWriteUpdate	wru;
				wru.UpdateAt = 0;	// L1 D-Cache access time (hit)-1
				wru.Register1 = theInstruction->Output1;
				wru.Register2 = theInstruction->Output2;
				NotificationList.Add(wru);

				}
			++Delta;
			continue;
			}

		theInstruction->MemoryDispatched = 1;
		theInstruction->MemoryExecuted = 1;
		RegisterWriteUpdate	wru;
		wru.UpdateAt = 0;	// L1 D-Cache access time (hit)-1
		wru.Register1 = theInstruction->Output1;
		wru.Register2 = theInstruction->Output2;
		NotificationList.Add(wru);

		// send the request to the memory system
		MemoryTransactionMessage	*mtm = NEW(MemoryTransactionMessage);
		mtm->Id = theInstruction->LoadStoreSequence;
		mtm->Id2 = theInstruction->RawLoadStoreSequence;
		mtm->Id3 = GetId();
		mtm->Address = theInstruction->MemoryAddress;
		mtm->Width = theInstruction->MemoryTransactionSize;
		mtm->Speculative = theInstruction->MisSpeculatedInstruction;

		if ( theInstruction->Operation == OpStore )
			mtm->Command = Write;
		else
			mtm->Command = Read;
		SendToMemorySystem.Insert(mtm);

		++IssuesThisCycle;
		++Delta;
		}
	////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// process send to memory list
	////////////////////////////////////////////////////////////////////////////////////////////////////////////
	while ( SendToMemorySystem.GetSize() )
		{
		MemoryTransactionMessage	*mtm;
		SendToMemorySystem.Peek(&mtm);
		mtm->SetSourceComponentId(GetId());
		if ( ! SendMessage(theMemorySystem, mtm))
			break;
		Instruction	*theInstruction = OrderingArray[mtm->Id % StoreBufferSize];
		if ( theInstruction->Operation == OpStore )
			ReturnBackFromMemorySystem.Insert(theInstruction->LoadStoreSequence);
#ifdef	ENABLE_MESSAGE_CHECK
		else
			printf("DISPATCHED " format_dec_int64 " AT " format_dec_int64 "\n", theInstruction->LoadStoreSequence, theOutputPort->GetDomain()->GetCurrentCycle());
#endif
		SendToMemorySystem.Remove(&mtm);
		}
	////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// process notification list
	////////////////////////////////////////////////////////////////////////////////////////////////////////////
	int	Index = NotificationList.GetSize();
	while ( Index > 0 )
		{
		--Index;
		RegisterWriteUpdate	wru = NotificationList[Index];
		if ( wru.UpdateAt <= 0 )	
			{
			theIssueWindowStage->WriteBackRegister(wru.Register1);
			theIssueWindowStage->WriteBackRegister(wru.Register2);
			NotificationList.RemoveAt(Index);
			}
		else
			--NotificationList[Index].UpdateAt;
		}
	return ( SimulateOneCyclePart2() );
	}

int
StoreBuffer	::	SimulateOneCyclePart2()
	{
	////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// process any memory requests that came back
	////////////////////////////////////////////////////////////////////////////////////////////////////////////
	while ( ReturnBackFromMemorySystem.GetSize() )
		{
		uint64	Request;

		ReturnBackFromMemorySystem.Remove(&Request);
		Instruction	*theInstruction = OrderingArray[Request%StoreBufferSize];
		if ( ! theInstruction )
			{
			printf("Doh! Request = " format_dec_int64 "\n", Request);
			Warning("Memory system responded with an invalid instruction");
			continue;
			}
#ifdef	ENABLE_MESSAGE_CHECK
		printf("PROCESSED RETURN BACK FOR " format_dec_int64 " AT " format_dec_int64 "\n",theInstruction->LoadStoreSequence, theOutputPort->GetDomain()->GetCurrentCycle());
#endif
		theInstruction->MemoryExecuted = 2;
		}

	////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// flush the head of the store buffer of any instructions sent to the completion stage
	////////////////////////////////////////////////////////////////////////////////////////////////////////////
	while ( true )
		{
		if ( MaximumRetiresPerCycle )
		if ( RetiresThisCycle >= MaximumRetiresPerCycle )
			break;

		Instruction	*theInstruction = OrderingArray[LoadStoreSequence%StoreBufferSize];
		if ( ! theInstruction )
			break;

		if ( theInstruction->MemoryExecuted == 2 )
			{
			if ( !theCompletionStage->DispatchTo(theInstruction) )
				break;	// no go
//			printf("4444: ");Debugger::DumpInstruction(theInstruction, theRenameStage);

			theRenameStage->WriteBackRegister(theInstruction->Output1);
			theRenameStage->WriteBackRegister(theInstruction->Output2);
			theRenameStage->NotifyAboutRegisterUse(theInstruction->Output1);
			theRenameStage->NotifyAboutRegisterUse(theInstruction->Output2);
			theRegisterAccessStage->WriteBackRegister(theInstruction->Output1);
			theRegisterAccessStage->WriteBackRegister(theInstruction->Output2);

			OrderingArray[LoadStoreSequence%StoreBufferSize] = NULL;
			++LoadStoreSequence;

			if ( theInstruction->Operation == OpStore )
				{
				// need to remove it from pendingstores
				StoreBufferEntry	*sbe = NEW(StoreBufferEntry);
				sbe->theInstruction = theInstruction;
				sbe->StrictComparison = true;
				StoreBufferEntry	*match = (StoreBufferEntry*)PendingStores->LocateObject(sbe);
				if ( ! match )
					FatalError("Store buffer problem");

				PendingStores->RemoveObject(sbe);
				DELETE(StoreBufferEntry, sbe);
				DELETE(StoreBufferEntry, match);
				}
			
			}
		else
			break;
		++RetiresThisCycle;
		}

	return(1);
	}

bool
StoreBuffer	::	ReceiveMessage(
	Component	*f,
	Message	*m)
	{
	MemoryTransactionMessage	*mtm = (MemoryTransactionMessage*)m;

	if ( mtm->Command != ReadResponse )
		{
		//printf("SB - NOT A ReadResponse\n");
		//m->Dump(stdout);
		DELETE(MemoryTransactionMessage, mtm);
		return(true);
		}

	if ( mtm->Id3 != (uint64)GetId() )	// filter out requests we did not send (happens when busses are shared)
		{
#ifdef	ENABLE_MESSAGE_CHECK
		printf("THIS WAS NOT MEANT FOR ME: %d\n", GetId());
		m->Dump(stdout);
#endif
		DELETE(MemoryTransactionMessage, mtm);
		return(true);
		}
#ifdef	ENABLE_MESSAGE_CHECK
	printf("RECEIVING RESPONSE FOR " format_dec_int64 " AT " format_dec_int64 "\n", mtm->Id, theOutputPort->GetDomain()->GetCurrentCycle());
	m->Dump(stdout);
#endif
	// receive a message.. filter out requests that originated while in a speculative state
	if ( mtm->Speculative )
		{
		if ( OrderingArray[mtm->Id % StoreBufferSize] )
		if ( OrderingArray[mtm->Id % StoreBufferSize]->MisSpeculatedInstruction )
		if ( OrderingArray[mtm->Id % StoreBufferSize]->LoadStoreSequence == mtm->Id )
		if ( OrderingArray[mtm->Id % StoreBufferSize]->RawLoadStoreSequence == mtm->Id2 )
			{
			ReturnBackFromMemorySystem.Insert(mtm->Id);
			DELETE(MemoryTransactionMessage, mtm);
			SimulateOneCyclePart2();
			return(true);
			}
		DELETE(MemoryTransactionMessage, mtm);
		return(true);
		}
		
	if ( OrderingArray[mtm->Id % StoreBufferSize] )
	if ( !OrderingArray[mtm->Id % StoreBufferSize]->MisSpeculatedInstruction )
	if ( OrderingArray[mtm->Id % StoreBufferSize]->LoadStoreSequence == mtm->Id )
	if ( OrderingArray[mtm->Id % StoreBufferSize]->RawLoadStoreSequence == mtm->Id2 )
		{
		ReturnBackFromMemorySystem.Insert(mtm->Id);
		DELETE(MemoryTransactionMessage, mtm);
		SimulateOneCyclePart2();
		return(true);
		}

	DELETE(MemoryTransactionMessage, mtm);
	return(true);
	}

bool
StoreBuffer	::	FlushSpeculativeState ()
	{

	// INFORM THE DOWN STREAM MEMORY SYSTEM TO FLUSH

	StoreBufferEntry	*sbe = NEW(StoreBufferEntry);
	for(int	Index = 0;Index < StoreBufferSize; Index++)
		{
		Instruction	*theInstruction;
		theInstruction = OrderingArray[Index];

		if (theInstruction )
		if ( theInstruction->Operation == OpStore )
			{
			sbe->theInstruction = theInstruction;
			sbe->StrictComparison = true;
			StoreBufferEntry	*match = (StoreBufferEntry*)PendingStores->LocateObject(sbe);
			if ( ! match )
				FatalError("Store buffer problem");

			PendingStores->RemoveObject(sbe);
			DELETE(StoreBufferEntry, match);
			}
		OrderingArray[Index] = NULL;
		}
	DELETE(StoreBufferEntry, sbe);

	ReturnBackFromMemorySystem.Clear();

	NotificationList.Clear();

	LoadStoreSequence = WatchForThisSpeculativeLoadStoreSequence;
	
	return ( true );
	}

bool
StoreBuffer	::	EnterSpeculativeState(
	uint64	MisSpeculatedLoadStoreSequence)
	{
	WatchForThisSpeculativeLoadStoreSequence = MisSpeculatedLoadStoreSequence;
	return ( true );
	}


