#	include	"Standard.h"

SuperScalarProcessor	::	SuperScalarProcessor()
	{
	theEngine = NULL;
	theFetchStage = NULL;
	theDecodeStage = NULL;
	theRenameStage = NULL;
	theIssueWindowStage = NULL;
	theRegisterAccessStage = NULL;
	theExecutionStage = NULL;
	theCompletionStage = NULL;
	theStoreBuffer = NULL;
	theBranchPredictor = NULL;
	theBranchTargetBuffer = NULL;
	}

SuperScalarProcessor	::	~SuperScalarProcessor()
	{
	SafeDelete(theFetchStage);
	SafeDelete(theDecodeStage);
	SafeDelete(theRenameStage);
	SafeDelete(theIssueWindowStage);
	SafeDelete(theRegisterAccessStage);
	SafeDelete(theExecutionStage);
	SafeDelete(theCompletionStage);
	SafeDelete(theStoreBuffer);
	SafeDelete(theBranchPredictor);
	SafeDelete(theBranchTargetBuffer);
	}

bool
SuperScalarProcessor	::	PrintStatistics (	
	OutputFormat	theFormat,
	FILE			*fp)
	{
	theFetchStage->PrintStatistics(theFormat, fp);
	theDecodeStage->PrintStatistics(theFormat, fp);
	theRenameStage->PrintStatistics(theFormat, fp);
	theRegisterAccessStage->PrintStatistics(theFormat, fp);
	theExecutionStage->PrintStatistics(theFormat, fp);
	theCompletionStage->PrintStatistics(theFormat, fp);
	theStoreBuffer->PrintStatistics(theFormat, fp);
	theBranchPredictor->PrintStatistics(theFormat, fp);
	theBranchTargetBuffer->PrintStatistics(theFormat, fp);
	return(true);
	}

bool
SuperScalarProcessor	::	BindParameters(SuperScalarProcessor	*This, TupleSpace	*theTupleSpace, char	*aName )
	{
	theTupleSpace->StartSubSpace(DefaultPass(aName, "SuperScalarProcessor"));

	SetComponentName(This, theTupleSpace);

	theTupleSpace->Bind(Parameter(This, MachineWidth), 4);
	theTupleSpace->Bind(Parameter(This, WindowSize), 128);

	FetchStage::BindParameters(StaticPass(This, theFetchStage), theTupleSpace);	
	RenameStage::BindParameters(StaticPass(This, theRenameStage), theTupleSpace);	
	IssueWindowStage::BindParameters(StaticPass(This, theIssueWindowStage), theTupleSpace);
	NonBypassingRegisterAccessStage::BindParameters(StaticPass(This, theRegisterAccessStage), theTupleSpace);
	ExecutionStage::BindParameters(StaticPass(This, theExecutionStage), theTupleSpace);
	CompletionStage::BindParameters(StaticPass(This, theCompletionStage), theTupleSpace);
	StoreBuffer::BindParameters(StaticPass(This, theStoreBuffer), theTupleSpace);
	SetAssociativeBranchTargetBuffer::BindParameters(StaticPass(This, theBranchTargetBuffer), theTupleSpace);
	GShareBranchPredictor::BindParameters(StaticPass(This, theBranchPredictor), theTupleSpace);
	theTupleSpace->EndSubSpace();
	return(true);
	}

bool
SuperScalarProcessor	::	Create()
	{
	theBranchPredictor = new GShareBranchPredictor;
	theBranchTargetBuffer = new SetAssociativeBranchTargetBuffer;

	theFetchStage = new FetchStage;
	theDecodeStage = new DelayStage;
	theRenameStage = new RenameStage;
	theIssueWindowStage = new IssueWindowStage;	
	theRegisterAccessStage = new NonBypassingRegisterAccessStage;	
	theExecutionStage = new ExecutionStage;	
	theCompletionStage = new CompletionStage;
	theStoreBuffer = new StoreBuffer;
		
	theBranchPredictor->Create();
	theBranchTargetBuffer->Create();
	theFetchStage->Create();
	theDecodeStage->Create();
	theIssueWindowStage->Create();
	theRegisterAccessStage->Create();
	theExecutionStage->Create();
	theCompletionStage->Create();
	theStoreBuffer->Create();

	AddComponent(theBranchPredictor);
	AddComponent(theBranchTargetBuffer);
	AddComponent(theFetchStage);
	AddComponent(theDecodeStage);
	AddComponent(theIssueWindowStage);
	AddComponent(theRegisterAccessStage);
	AddComponent(theExecutionStage);
	AddComponent(theCompletionStage);
	AddComponent(theStoreBuffer);
	return(true);
	}

bool
SuperScalarProcessor	::	BindStructures (
					TupleSpace	*theTupleSpace,
					Engine		*aEngine,
					Component	*aInstructionMemorySystem,
					Component	*aDataMemorySystem)
	{
	theEngine = aEngine;

	ProcessorStageList.Add(theFetchStage);
	ProcessorStageList.Add(theDecodeStage);
	ProcessorStageList.Add(theRenameStage);
	ProcessorStageList.Add(theIssueWindowStage);
	ProcessorStageList.Add(theRegisterAccessStage);
	ProcessorStageList.Add(theExecutionStage);
	ProcessorStageList.Add(theCompletionStage);

	theInstructionMemorySystem = aInstructionMemorySystem;
	theDataMemorySystem = aDataMemorySystem;

	// initialize the components
	theBranchPredictor->BindStructures();
	theBranchTargetBuffer->BindStructures();

	theFetchStage->BindStructures(theEngine, theBranchPredictor,
								  theBranchTargetBuffer, theInstructionMemorySystem,
								  &ProcessorStageList, theStoreBuffer); 
	theDecodeStage->Initialize(1, MachineWidth, true, false);
	theRenameStage->BindStructures(theFetchStage, theIssueWindowStage );
	theIssueWindowStage->BindStructures(theFetchStage, theRenameStage, theStoreBuffer, theCompletionStage );
	theRegisterAccessStage->BindStructures (theRenameStage, theIssueWindowStage, theCompletionStage );
	theExecutionStage->BindStructures (theFetchStage, theRenameStage, theRegisterAccessStage,
		theStoreBuffer, theCompletionStage );
	theCompletionStage->BindStructures(theFetchStage, theRenameStage);

	theStoreBuffer->BindStructures(theRenameStage, theIssueWindowStage, theRegisterAccessStage,
								   theCompletionStage, theDataMemorySystem);

	// link the processor stages together
	theFetchStage->Connect(NULL, theDecodeStage);
	theDecodeStage->Connect(theFetchStage, theRenameStage);
	theRenameStage->Connect(theDecodeStage, theIssueWindowStage);
	theIssueWindowStage->Connect(theRenameStage, theRegisterAccessStage);
	theRegisterAccessStage->Connect(theIssueWindowStage, theExecutionStage);
	theExecutionStage->Connect(theRegisterAccessStage, theCompletionStage);
	theCompletionStage->Connect(theExecutionStage, NULL);

	return ( true );
	}

int
SuperScalarProcessor	::	SyncSimulateOneCycle()
	{
	theCompletionStage->SimulateOneCycle();
	theExecutionStage->SimulateOneCycle();
	theStoreBuffer->SimulateOneCycle();
	theRegisterAccessStage->SimulateOneCycle();
	theIssueWindowStage->SimulateOneCycle();
	theRenameStage->SimulateOneCycle();
	theDecodeStage->SimulateOneCycle();
	theFetchStage->SimulateOneCycle();
	return(1);
	}

bool
SuperScalarProcessor	::	PrintDebuggingInformation(
			FILE	*fp,
			int		Level)
	{
	theFetchStage->PrintDebuggingInformation(fp, Level);
	theDecodeStage->PrintDebuggingInformation(fp, Level);
	theRenameStage->PrintDebuggingInformation(fp, Level);
	theIssueWindowStage->PrintDebuggingInformation(fp, Level);
	theRegisterAccessStage->PrintDebuggingInformation(fp, Level);
	theExecutionStage->PrintDebuggingInformation(fp, Level);
	theCompletionStage->PrintDebuggingInformation(fp, Level);
	theStoreBuffer->PrintDebuggingInformation(fp, Level);
	printf("Completion buffer is waiting for instruction " format_dec_int64 "\n", theCompletionStage->SequenceNumber);
	Instruction	*theInstruction = theFetchStage->LocateInstruction(theCompletionStage->SequenceNumber);
	if ( ! theInstruction )
		printf("Which I cannot find!!\n");
	else
		{
		Debugger::DumpInstruction(theInstruction, theRenameStage);
		Debugger::PrintDependencies(2, theInstruction, theFetchStage, theRenameStage);
		}

	return(true);
	}
