#	include	"Standard.h"


bool
SingleProcessorSystemBoard	::	BindParameters(SingleProcessorSystemBoard	*This, TupleSpace	*theTupleSpace, char	*aName )
	{
	theTupleSpace->StartSubSpace(DefaultPass(aName, "System"));
	SetComponentName(This, theTupleSpace);
	theTupleSpace->Bind(Parameter(This, MasterClockFrequency), cuint64(1E9));
	theTupleSpace->Bind(Parameter(This, FastForwardAmount), cuint64(0));
	theTupleSpace->Bind(Parameter(This, WarmUpAmount), cuint64(0));
	theTupleSpace->Bind(Parameter(This, SimulateAmount), cuint64(0));
	Domain::BindParameters(StaticPass(This, theDomain), theTupleSpace);
	SuperScalarProcessor::BindParameters(StaticPass(This, theProcessor), theTupleSpace);
	SimpleCacheMemorySystem::BindParameters(StaticPass(This, theMemory), theTupleSpace);
	theTupleSpace->EndSubSpace();
	return(true);
	}

bool
SingleProcessorSystemBoard	::	PrintStatistics (
	OutputFormat	theFormat,
	FILE			*fp)
	{
	PrintValue(theFormat, fp, "CycleCount", theDomain->GetStatCurrentCycle());
	theProcessor->PrintStatistics(theFormat, fp);
	theMemory->PrintStatistics(theFormat, fp);
	return(true);
	}

bool
SingleProcessorSystemBoard	::	Create()
	{
	theProcessor = new SuperScalarProcessor;
	theMemory = new SimpleCacheMemorySystem;
	theDomain = new Domain;
	theProcessor->Create();
	theMemory->Create();
	theDomain->Create();
	AddComponent(theMemory);
	AddComponent(theProcessor);
	return(true);
	}

bool
SingleProcessorSystemBoard	::	BindStructures(TupleSpace	*theTupleSpace, Engine	*aEngine)
	{
	theEngine = aEngine;
	Join(theDomain);
	theProcessor->BindStructures(theTupleSpace, aEngine,
		theMemory->GetOutputInstructionPort(),	// source
		theMemory->GetInputInstructionPort(),	// destination
		theMemory->GetOutputDataPort(),			// source
		theMemory->GetInputDataPort()			// destination
		);
	theMemory->BindStructures(theTupleSpace);
	return(true);
	}

bool
SingleProcessorSystemBoard	::	Initialize(TupleSpace	*theTupleSpace, Engine	*aEngine)
	{
	Create();
	BindParameters(this, theTupleSpace);
	BindStructures(theTupleSpace, aEngine);
	return(true);
	}

bool
SingleProcessorSystemBoard	::	FastForward(
	uint64	nInstructions)
	{
	uint64	Index;
	uint64	Count = 0;
	uint64	Div = nInstructions / 100;
	uint64	Percent = 0;
	Instruction	*theInstruction	=	NEW(Instruction);
	Index = 0;

	theDomain->SetSimulationMode(FastForwarding);
	while(Index < nInstructions)
		{
		if(!theEngine->NextInstruction(&theInstruction))
			{
			DELETE(Instruction, theInstruction);
			return(false);
			}
		++Index;
		++Count;
		if ( Count >= Div )
			{
			++Percent;
			printf("Fast forward " format_dec_uint64 " of " format_dec_uint64 " (" format_dec_uint64 " percent)     \r",
				Index, nInstructions, Percent );
			fflush(stdout);
			Count = 0;
			}
		}
	printf("Fast forward complete                                             \n");
	DELETE(Instruction, theInstruction);
	return(true);
	}

bool
SingleProcessorSystemBoard	::	WarmUp(
	uint64	nInstructions)
	{
	uint64	StartSeq = theProcessor->theCompletionStage->SequenceNumber;
	uint64	Count = 0;
	uint64	Div = nInstructions / 100;
	uint64	Percent = 0;
	uint64	Index = 0;
	uint64	CountStartSeq = theProcessor->theCompletionStage->SequenceNumber;
	uint64	DeadlockDetectSeq = theProcessor->theCompletionStage->SequenceNumber;

	theDomain->SetSimulationMode(WarmingUp);
	while ( (theProcessor->theCompletionStage->SequenceNumber - StartSeq) < nInstructions )
		{
		if (theProcessor->theFetchStage->EndSimulation)
			break;

		theDomain->AsyncSimulateOneCycle();
		Count = theProcessor->theCompletionStage->SequenceNumber - CountStartSeq;
		if ( Count >= Div )
			{
			++Percent;
			printf("Warmup " format_dec_uint64 " of " format_dec_uint64 " (" format_dec_uint64 " percent)     \r",
				theProcessor->theCompletionStage->SequenceNumber - StartSeq, nInstructions, Percent );
			fflush(stdout);
			Count = 0;
			CountStartSeq = theProcessor->theCompletionStage->SequenceNumber;
			}
		++Index;
		if ( Index > 10000 )
			{
			if (theProcessor->theCompletionStage->SequenceNumber == DeadlockDetectSeq )
				{
				FatalError("Machine deadlock during warm up -- check configuration!\n");
				}
			
			DeadlockDetectSeq = theProcessor->theCompletionStage->SequenceNumber;
			Index = 0;
			}
				
		}
	printf("Warmup complete                                             \n");

	theDomain->ResetStatistics();	
	return(true);
	}

bool
SingleProcessorSystemBoard	::	Simulate(
	uint64	nInstructions)
	{
	uint64	LastInstruction = 0;
	uint64	StartInstruction = theProcessor->theCompletionStage->SequenceNumber;
	uint64	EndInstruction = StartInstruction + nInstructions;
	uint64	Count = 0;

	theDomain->SetSimulationMode(Simulating);
	while ( (nInstructions == (0)) || (theProcessor->theCompletionStage->SequenceNumber < EndInstruction) )
		{
		if (theProcessor->theFetchStage->EndSimulation)
			break;

		theDomain->AsyncSimulateOneCycle();

		++Count;
		if ( Count >= 10000 )
			{
			Count = 0;
			printf("Simulate [" format_dec_int64 "] Completed: " format_dec_int64 " IPC = %f         \r",
				theDomain->GetCurrentCycle(), theProcessor->theCompletionStage->SequenceNumber,
				(((double)theProcessor->theCompletionStage->SequenceNumber)/
				((double)theDomain->GetCurrentCycle())) );
#ifdef	ENABLE_LIBRARY_CHECKS
			theProcessor->PrintDebuggingInformation(stdout, 3);
#endif
			// dead lock detection.  It's a good thing.
			if ( theProcessor->theCompletionStage->SequenceNumber == LastInstruction )
				{
				FatalError("Machine deadlock -- probably an invalid configuration.  Check buffer spaces");
				}
			LastInstruction = theProcessor->theCompletionStage->SequenceNumber;
			}
		}
#ifdef	ENABLE_LIBRARY_CHECKS

	printf("END [" format_dec_int64 "] Completed: " format_dec_int64 " IPC = %f                  \n",
			theDomain->GetCurrentCycle(), theProcessor->theCompletionStage->SequenceNumber,
				(((double)theProcessor->theCompletionStage->SequenceNumber)/
				((double)theDomain->GetCurrentCycle())) );

#endif

	return(true);
	};

bool
SingleProcessorSystemBoard	::	DoSimulation()
	{
	FastForward(FastForwardAmount);
	WarmUp(WarmUpAmount);
	Simulate(SimulateAmount);

	return(true);
	}


