#	include	"Standard.h"

DEFINE_ALLOCATOR(AlphaInstruction);


// NOP is bis $31, $31, $31
#	define	AlphaNOPInstruction	((0x11<<26)|(0x20<<5)|(31<<21)|(31<<16)|(31<<0))


/****************************************** AXP Decoding Engine *************************************/

// if you modify AlphaOp's structure, you probably will need to mess around with these.
#	define	NoLink		{-1, -1, -1}, {-1, -1, -1, -1}, NULL,	NULL, NULL
#	define	Link(xxx)	{-1, -1, -1}, {-1, -1, -1, -1}, xxx##Table, xxx##Decode, NULL
#	define	Code(xxx)	xxx

#	define	DirectCode(code, format, type, name, outputs, inputs)	{	#name, name, code, format, type, 	NoLink	},
#	define	LinkCode(code, name)									{	#name, name, code, UnknownFormat, OpUnknown, 	Link(name)	},
#	define	TableEnd												{	NULL,	NoCode, 0, UnknownFormat, 0, NoLink },


// don't change these unless you know what your doing.
#	define	oprDecoding(zzz)	((zzz>>5)&0x7f)
#	define	jmpDecoding(zzz)	((zzz>>14)&0x3)
#	define	fpDecoding(zzz)		((zzz>>5)&0x7ff)
#	define	mfcDecoding(zzz)	((zzz>>8)&0xff)
#	define	opDecoding(zzz)		((zzz>>25)&0x3f)

// The following tables, which are defined by the macros in various table files
// are "compressed" op-code listings.  These are decompressed at run-time into an
// efficient form for parsing alpha opcodes.  The decompression step just saves you
// from hacking to precisely enter blank entries for all table space.

static	uint32 intaDecode(uint32	xxx) { return( oprDecoding(xxx) ); }
static	AlphaOp	intaTable[]	=
	{
#	include	"Tables/inta.table"
	TableEnd
	};

static	uint32	intlDecode(uint32	xxx)	{	return ( oprDecoding(xxx) ); }
static	AlphaOp	intlTable[]	=
	{
#	include	"Tables/intl.table"
	TableEnd
	};

static	uint32	intsDecode(uint32	xxx)	{	return ( oprDecoding(xxx) ); }
static	AlphaOp	intsTable[]	=
	{
#	include	"Tables/ints.table"
	TableEnd
	};

static	uint32	intmDecode(uint32	xxx)	{	return ( oprDecoding(xxx) ); }
static	AlphaOp	intmTable[]	=
	{
#	include	"Tables/intm.table"
	TableEnd
	};

static	uint32	itfpDecode(uint32	xxx)	{	return ( fpDecoding(xxx) ); }
static	AlphaOp	itfpTable[]	=
	{
#	include	"Tables/itfp.table"
	TableEnd
	};

static	uint32	fltvDecode(uint32	xxx)	{	return ( fpDecoding(xxx) ); }
static	AlphaOp	fltvTable[]	=
	{
#	include	"Tables/fltv.table"
	TableEnd
	};

static	uint32	fltiDecode(uint32	xxx)	{	return ( fpDecoding(xxx) ); }
static	AlphaOp	fltiTable[]	=	
	{
#	include	"Tables/flti.table"
	TableEnd
	};

static	uint32	fltlDecode(uint32	xxx)	{	return ( fpDecoding(xxx) ); }
static	AlphaOp	fltlTable[]	=
	{
#	include	"Tables/fltl.table"
	TableEnd
	};

static	uint32	miscDecode(uint32	xxx)	{	return ( mfcDecoding(xxx) ); }
static	AlphaOp	miscTable[]	=
	{
#	include	"Tables/misc.table"
	TableEnd
	};

static	uint32	jmplinkDecode(uint32	xxx)	{	return(jmpDecoding(xxx) );	}
static	AlphaOp	jmplinkTable[]	=
	{
#	include	"Tables/jmplink.table"
	TableEnd
	};

static	uint32	fptiDecode(uint32		xxx)	{	return(oprDecoding(xxx) ); }
static	AlphaOp	fptiTable[]	=
	{
#	include	"Tables/fpti.table"
	TableEnd
	};

static	uint32	TopLevelDecode(uint32	xxx) { return( opDecoding(xxx) ); }
static	AlphaOp	compressedTopLevelTable[]	=
	{
#	include	"Tables/TopLevel.table"
	TableEnd
	};

#	undef	DirectCode
#	undef	LinkCode

static	AlphaOp	*TopLevelTable;
static	AlphaOp	*ReverseMapping [ InstructionCodes ];


static
bool
ConstructReverseMapping (
	AlphaOp	*ReverseMapTable[InstructionCodes],
	AlphaOp	*OpCodeTable)
	{
	int	Index = 0;
	while ( OpCodeTable[Index].Name )
		{
		ReverseMapTable[OpCodeTable[Index].OpCode] = &OpCodeTable[Index];
		if ( OpCodeTable[Index].TableLink )
			ConstructReverseMapping(ReverseMapTable, OpCodeTable[Index].TableLink);
		++Index;
		}
	return ( true );
	}


static
AlphaOp	*
ConstructForwardMapping(
	AlphaOp	*compressedTable,
	uint32	UncompressedSize)
	{
	AlphaOp	*uncompressedTable = new AlphaOp[UncompressedSize];

	memset(uncompressedTable, 0, sizeof(AlphaOp)*UncompressedSize);

	int	Index = 0;
	while ( compressedTable[Index].Name )
		{
		if ( compressedTable[Index].BitCode >= UncompressedSize )
			{
			fprintf(stderr, "Error in Alpha AXP opcode table for opcode %s -- exceeds mask\n",
				compressedTable[Index].Name);
			}	
		else
			{
			if(uncompressedTable[compressedTable[Index].BitCode].Name)
				{
				fprintf(stderr, "Warning: duplicate opcode: %s conflicts with %s\n",
					compressedTable[Index].Name,
					uncompressedTable[compressedTable[Index].BitCode].Name);
				exit(1);
				}
			uncompressedTable[compressedTable[Index].BitCode] = compressedTable[Index];
			if ( uncompressedTable[compressedTable[Index].BitCode].TableLink )
				uncompressedTable[compressedTable[Index].BitCode].TableLink = ConstructForwardMapping(
				compressedTable[Index].TableLink, compressedTable[Index].DecodeLink(0xffffffff)+1);
			}	
		++Index;
		}
	return ( uncompressedTable );
	}

// Decoding needs to only occur once.  It can multiple times (no harm done, just wasted memory)
bool	AlphaProcessor	::	DecodingEngineIsInitialized = false;

bool
AlphaProcessor	::	InitializeAlphaDecodingEngine ()
	{
	if ( DecodingEngineIsInitialized )	return ( true );

	// construct opcode -> AlphaOp mapping
	ConstructReverseMapping(ReverseMapping, compressedTopLevelTable);

	// construct raw instruction bits -> AlphaOp mapping
	TopLevelTable = ConstructForwardMapping(compressedTopLevelTable, TopLevelDecode(0xffffffff)+1);
						
	
	DecodingEngineIsInitialized = true;
	return(true);
	}

char	*AlphaRegisterNames[]	=	
	{	"v0", "r1", "r2", "r3", "r4", "r5", "r6", "err", "r8", "r9", "r10", "r11", "r12",
		"r13", "r14", "fp", "a0", "a1", "a2", "a3", "a4", "a5", "r22", "r12", "r24", "r25",
		"ra", "r27", "r28", "gp", "sp", "zero"};

/************************************ AXP Execution Engine *************************************/


// General support functions used by the op-code simulation...

#	define	GenMaskForBit(zzz)			(((uint64)1)<<(zzz))
#	define	Make64(hhhh,llll)			((((uint64)(hhhh))<<32)|((uint64)(llll)))
#	define	Maxuint64					Make64(0xffffffff, 0xffffffff)
#	define	MaskLower32(zzz)			((zzz) & Make64(0x00000000, 0xffffffff))
#	define	MaskUpper32(zzz)			((zzz) & Make64(0xffffffff, 0x00000000))
#	define	SignExtend32To64(zzz)		(((zzz) & GenMaskForBit(31)) ? ((zzz)|Make64(0xffffffff, 0x00000000)) : (zzz))
#	define	SignExtend21To64(zzz)		(((zzz) & GenMaskForBit(20)) ? ((zzz)|Make64(0xffffffff, 0xffe00000)) : (zzz))
#	define	SignExtend16To64(zzz)		(((zzz) & GenMaskForBit(15)) ? ((zzz)|Make64(0xffffffff, 0xffff0000)) : (zzz))
#	define	SignExtend8To64(zzz)		(((zzz) & GenMaskForBit(7)) ? ((zzz)|Make64(0xffffffff, 0xffffff00)) : (zzz))

#	define	Maskb0(zzz)					((zzz)&Make64(0, 0x000000ff))
#	define	Maskb1(zzz)					((zzz)&Make64(0, 0x0000ff00))
#	define	Maskb2(zzz)					((zzz)&Make64(0, 0x00ff0000))
#	define	Maskb3(zzz)					((zzz)&Make64(0, 0xff000000))
#	define	Maskb4(zzz)					((zzz)&Make64(0x000000ff, 0))
#	define	Maskb5(zzz)					((zzz)&Make64(0x0000ff00, 0))
#	define	Maskb6(zzz)					((zzz)&Make64(0x00ff0000, 0))
#	define	Maskb7(zzz)					((zzz)&Make64(0xff000000, 0))
#	define	b0(zzz)						(((zzz)>>0) & Make64(0, 0xff))
#	define	b1(zzz)						(((zzz)>>8) & Make64(0, 0xff))
#	define	b2(zzz)						(((zzz)>>16) & Make64(0, 0xff))
#	define	b3(zzz)						(((zzz)>>24) & Make64(0, 0xff))
#	define	b4(zzz)						(((zzz)>>32) & Make64(0, 0xff))
#	define	b5(zzz)						(((zzz)>>40) & Make64(0, 0xff))
#	define	b6(zzz)						(((zzz)>>48) & Make64(0, 0xff))
#	define	b7(zzz)						(((zzz)>>56) & Make64(0, 0xff))
#	define	w0(zzz)						((zzz) & Make64(0, 0xffff))
#	define	w1(zzz)						(((zzz)>>16) & Make64(0, 0xffff))
#	define	w2(zzz)						(((zzz)>>32) & Make64(0, 0xffff))
#	define	w3(zzz)						(((zzz)>>48) & Make64(0, 0xffff))
#	define	l0(zzz)						((zzz) & Make64(0, 0xffffffff))
#	define	l1(zzz)						(((zzz)>>32) & Make64(0, 0xffffffff))
#	define	q0(zzz)						(zzz)

#	define	KeepMaskN(yyy,nnn,sss)		(((yyy)&(nnn)) ? (cuint64(0xff)<<(sss)) : 0)
#	define	KeepMask(yyy)				( KeepMaskN(yyy, 0x01, 0)|KeepMaskN(yyy, 0x02, 8)|KeepMaskN(yyy, 0x04, 16)|\
										  KeepMaskN(yyy, 0x08, 24)|KeepMaskN(yyy, 0x10, 32)|KeepMaskN(yyy, 0x20, 40)|\
										  KeepMaskN(yyy, 0x40, 48)|KeepMaskN(yyy, 0x80, 56) )
#	define	ByteZap(zzz,yyy)			((zzz) & KeepMask((~(yyy))))
#	define	ByteKeep(zzz,yyy)			((zzz) & KeepMask((yyy)))

// Specific support for S (float) and T (double) type formats
#	define	FloatingPointSignBitMask	GenMaskForBit(63)

#	define	MAP_S(yyy)					(	((((yyy) & 0xff) == 0xff) ? 0xfff : 	\
											((((yyy) & 0xff) == 0x00) ? 0x000 :	\
											((((yyy) & 0x80) == 0x80) ? (0x400|((yyy)&0x7f))	:	\
											(0x380|((yyy)&0x7f)) )))	)

#	define	LoadFromSFormat_(xxx)		(	((((xxx) >> 31)&0x01)<<63)	|	\
											((MAP_S(((xxx)>>23)&0xff)) << 52)			|	\
											(((xxx) & 0x7fffff) << 29)		|	\
											(0)	)

#	define	LoadFromSFormat(xxx)		LoadFromSFormat_(cuint64(xxx))
#	define	StoreToSFormat(xxx)			(	((((xxx) >> 62)&0x03)	<< 30)		|	((xxx >> 29)&0x3fffffff)	)



// support functions for op-code parsing

#	define	OpCodeField(zzz)					(((zzz) >> 26) & 0x3f)
#	define	RegisterField_Ra(zzz)				(((zzz) >> 21) & 0x1f)
#	define	RegisterField_Rb(zzz)				(((zzz) >> 16) & 0x1f)
#	define	RegisterField_Rc(zzz)				(((zzz) >> 0) & 0x1f)
#	define	MemoryDisplacementField(zzz)		SignExtend16To64((((zzz) >> 0) & 0xffff))
#	define	FunctionField(zzz)					(((zzz) >> 0) & 0xffff)
#	define	OpFunctionField(zzz)				(((zzz) >> 5) & 0x7f)
#	define	BranchDisplacementField(zzz)		cint64(4*SignExtend21To64((((zzz) >> 0) & 0x1fffff)))
#	define	LiteralOrNotField(zzz)				(((zzz) >> 12) & 0x01)
#	define	LiteralField(zzz)					cuint64(((zzz) >> 13) & 0xff)
#	define	PALCodeField(zzz)					(((zzz) >> 0) & 0x3ffffff)



/*************************************** Implementors **************************************/

// What happens is all of the Implementor/*.use files declare lots of functions, such as
// addq_Implementation(...).  These functions look like their corresponding Implementor/<class>.form.
// Essentially, think of the .form files as declaring mini environments, and the .use files use
// those environments to implement the opcode functionality.  The environment does routine things
// for you, like reading the registers and writing the result.  You do not have to use a .form to
// implement an opcode.  The opcode execution engine will invoke an function with the corresponding
// arguments (ImplementorArguments).  So if you want to hijack an opcode that is how it will be done....
// A short function is provided for this: HighJackOpCode(uint32	topLevelCode, uint32 SecondLevelCode, void *functionptr)


// this is just for bit-bashing purposes.
typedef	union	_fpReg
	{
	uint64	Asuint64;
	int64	Asint64;
	double	Asdouble;
	uint32	Asuint32;
	int32	Asint32;
	float	Asfloat;
	}	fpReg;

#	define	SimulationFunction(name, foo)	static	void	 name##_Implementation(ImplementorArguments) { foo; }
#	define	BindFunction(name, list)


// Support functions that map implementor forms into the simulated AlphaProcessor class.  If you want
// to define AlphaProcessor differently, then you may need to redefine these macros.

#	define	ReadFloatingPointRegister_Ra				Proc->FloatingPointRegisters[RegisterField_Ra(RawInstruction)]
#	define	ReadFloatingPointRegister_Rb				Proc->FloatingPointRegisters[RegisterField_Rb(RawInstruction)]
#	define	ReadFloatingPointRegister_Rc				Proc->FloatingPointRegisters[RegisterField_Rc(RawInstruction)]
#	define	WriteFloatingPointRegister_Rc(xxx)			(Proc->FloatingPointRegisters[RegisterField_Rc(RawInstruction)] = xxx)
#	define	WriteFloatingPointRegister_Ra(xxx)			(Proc->FloatingPointRegisters[RegisterField_Ra(RawInstruction)] = xxx)
#	define	WriteFloatingPointControlRegister(xxx)		Proc->FloatingPointControlRegister = xxx
#	define	ReadFloatingPointControlRegister			Proc->FloatingPointControlRegister

#	define	ReadIntegerRegister_Ra						Proc->GeneralPurposeRegisters[RegisterField_Ra(RawInstruction)]
#	define	ReadIntegerRegister_Rb						Proc->GeneralPurposeRegisters[RegisterField_Rb(RawInstruction)]
#	define	ReadIntegerRegister_Rc						Proc->GeneralPurposeRegisters[RegisterField_Rc(RawInstruction)]
#	define	ReadIntegerRegister_Rb_Or_Literal			(LiteralOrNotField(RawInstruction) ? LiteralField(RawInstruction) : ReadIntegerRegister_Rb)
#	define	WriteIntegerRegister_Rc(xxx)				(Proc->GeneralPurposeRegisters[RegisterField_Rc(RawInstruction)] = xxx)
#	define	WriteIntegerRegister_Ra(xxx)				(Proc->GeneralPurposeRegisters[RegisterField_Ra(RawInstruction)] = xxx)
#	define	AccessMemory								(Proc->theMemory)
#	define	SetNextProgramCounter(xxx)					(Proc->NextProgramCounter = xxx)
#	define	GetCurrentProgramCounter					Proc->ProgramCounter



// Right here is where all the foo_Implementation(...) functions get created.
#	define	DefineForms
#	include	"AllForms.h"
#	include	"AllUses.h"
#	undef	DefineForms
#	include	"AllForms.h"
#	undef	SimulationFunction
#	undef	BindFunction


/*********************** Binding implementors into decoding tables ********************/

// This section binds the previously created functions into the decoding table.  We then
// invoke these functions via function pointers.
#	define	ALL_FPT_MODES(xxx)	xxx, xxx##_c, xxx##_d, xxx##_u, xxx##_ud, xxx##_suc, xxx##_sum, xxx##_su, xxx##_um, \
								xxx##_uc, xxx##_sud, xxx##_suic, xxx##_suim, xxx##_sui, xxx##_suid, xxx##_m

#	define	ALL_FPS_MODES(xxx)	xxx, xxx##_c, xxx##_d, xxx##_u, xxx##_ud, xxx##_suc, xxx##_sum, xxx##_su, xxx##_um, \
								xxx##_uc, xxx##_sud, xxx##_suic, xxx##_suim, xxx##_sui, xxx##_suid, xxx##_m

bool	AlphaProcessor	::	ExecutionEngineIsInitialized	=	false;

// hack: this is a nasty preprocessing hack.
// don't try this at home.  This takes a variable size list (x, ...) and packages it into
// an array.  It is there to fool cpp/C++ into multi-argument macros (but not really).  it's
// a gross hack, I admit.  If you do -Wall on some systems it *will* complain about this hack.
static	int	y[256];
static	int	FormArray(int	x0) {y[0]=x0;return(1);}					// handle common case
static	int	FormArray(int	x0, int	x1) {y[0]=x0;y[1]=x1;return(2);}	// handle common case
static	int	FormArray(int	x0, int	x1, int	x2) {y[0]=x0;y[1]=x1;y[2]=x2;return(3);}	// handle common case
// this handles the rest.. note: while you could return the size of 16 and break on -1
// in calls that use this, that *really* slows down compilation times.  
static	int	FormArray(
	int x0, int x1, int x2, int x3,
	int x4=-1, int x5=-1, int x6=-1, int x7=-1,
	int x8=-1, int x9=-1, int x10=-1, int x11=-1,
	int x12=-1, int x13=-1, int x14=-1, int x15=-1	)
	{
	y[0]=x0; y[1]=x1; y[2]=x2; y[3]=x3;
	y[4]=x4; y[5]=x5; y[6]=x6; y[7]=x7;
	y[8]=x8; y[9]=x9; y[10]=x10; y[11]=x11;
	y[12]=x12; y[13]=x13; y[14]=x14; y[15]=x15;
	int	Index = 0;while(Index<16) {if(y[Index]==-1) return(Index);++Index;}
	return(Index);
	};
static	int	FormArray(
	int x0, int x1, int x2, int x3,
	int x4, int x5, int x6, int x7,
	int x8, int x9, int x10, int x11,
	int x12, int x13, int x14, int x15,
	int x16, int x17=-1, int x18=-1, int x19=-1,
	int x20=-1, int x21=-1, int x22=-1, int x23=-1,
	int x24=-1, int x25=-1, int x26=-1, int x27=-1,
	int x28=-1, int x29=-1, int x30=-1, int x31=-1)
	{
	y[0]=x0; y[1]=x1; y[2]=x2; y[3]=x3;
	y[4]=x4; y[5]=x5; y[6]=x6; y[7]=x7;
	y[8]=x8; y[9]=x9; y[10]=x10; y[11]=x11;
	y[12]=x12; y[13]=x13; y[14]=x14; y[15]=x15;
	y[16]=x16; y[17]=x17; y[18]=x18; y[19]=x19;
	y[20]=x20; y[21]=x21; y[22]=x22; y[23]=x23;
	y[24]=x24; y[25]=x25; y[26]=x26; y[27]=x27;
	y[28]=x28; y[29]=x29; y[30]=x30; y[31]=x31;
	int	Index = 0;while(Index<32) {if(y[Index]==-1) return(Index);++Index;}
	return(Index);
	};


bool
AlphaProcessor	::	InitializeAlphaExecutionEngine()
	{
	ExecutionEngineIsInitialized = true;

#	define	SimulationFunction(name, foo)
#	define	BindFunction(name, list) {\
	int	BindSize = FormArray##list;\
	for(int	i=0;i<BindSize;i++) ReverseMapping[y[i]]->ImplementationFunction = name##_Implementation;}


// Here is where the op-codes actually get bound into their implementation functions.
// what you get is a bunch of loops like the above.

#	define	DefineForms
#	include	"AllForms.h"
#	include	"AllUses.h"
#	undef	DefineForms
#	include	"AllForms.h"
#	undef	SimulationFunction
#	undef	BindFunction


// also bind the register inputs and outputs for decoding purposes

// bind outputs
#	define	DirectCode(code, format, type, name, outputs, inputs)	{\
	int	BindSize = FormArray##outputs;\
	for(int	i=0;i<BindSize;i++) ReverseMapping[name]->Outputs[i] = y[i];}
#	define	LinkCode(code, name)
#	include	"AllTables.h"
#	undef	DirectCode
#	undef	LinkCode

// bind inputs
#	define	DirectCode(code, format, type, name, outputs, inputs)	{\
	int	BindSize = FormArray##inputs;\
	for(int	i=0;i<BindSize;i++) ReverseMapping[name]->Inputs[i] = y[i];}
#	define	LinkCode(code, name)
#	include	"AllTables.h"
#	undef	DirectCode
#	undef	LinkCode


	return ( true );
	}

// This function is used to find the AlphaOp record pertaining to an opcode (which may not
// exist in the Alpha ISA, but there is space within the decoding tables for it.
// you can then modify the AlphaOp structure to change its decoding charectoristics, and/or
// provide a new implementation function (or a modified implementation function).  No error
// checking is done with this function -- you need to know what you are doing.  An example
// call might be HighJackOpCode(0x12, 0x72).  This returns the AlphaOp record for the MSKQH
// instruction.  Note: the TopLevel, and NextLevel codes are precisely as the Alpha AXP
// reference manual, except for the 0x18 (misc) instructions.  These are shifted right by 8 bits
// for the NextLevel.  I.e., HighJackOpCode(0x18, 0x04) is the EXCB instruction (as opposed to
// (0x18, 0x0400).  This is to save table space.  All other opcodes are as they are in the book.
// Final note: you can create whole new decoding tables with this (off of the top level table).
// To do this, use multiple calls to this function.  The first call will return the TopLevel
// entry.  Modify that by making a new table (declare it, initialize it, etc).  You can then
// modify that new table with this same function.

AlphaOp	*
AlphaProcessor	::	HighJackOpCode(
	uint32		TopLevel,
	uint32		NextLevel)
	{
	// you must know what you doing when you call this....
	AlphaOp	*ptr = &TopLevelTable [ TopLevel ];

	if ( ptr->TableLink )
		{
		ptr = & ptr->TableLink [ NextLevel ];
		}
	return(ptr);
	}

/************************* Alpha Processor Simulator Support functions *******************/

// These just move bytes to and from the simulated memory space and the simulator memory space
// They are used by the fake OS

bool
AlphaProcessor	::	CopyFromHost (
	uint64	Destination,
	void	*Source,
	uint32	Bytes)
	{
	uint8	*BytePtr = (uint8*)Source;
	uint32	*WordPtr = (uint32*)Source;
	uint64	bDestination = (uint64) Destination;

	/*
	while(Bytes>=4)
		{
		theMemory->WriteMemory32((uint64)bDestination, (uint32)*WordPtr);
		++WordPtr;
		bDestination += 4;
		BytePtr += 4;
		Bytes -= 4;
		}
	*/	
	while(Bytes)
		{
		theMemory->WriteMemory8((uint64)bDestination, (uint8)*BytePtr);
		--Bytes;
		++BytePtr;
		++bDestination;
		}
	return ( true );
	}

bool
AlphaProcessor	::	CopyToHost (
	void	*Destination,
	uint64	Source,
	uint32	Bytes)
	{
	uint64	BytePtr = (uint64)Source;
	uint8	*dBytePtr = (uint8*)Destination;
	uint32	*dWordPtr = (uint32*)Destination;
	uint32	Word;
	uint8	Byte;

/*	
	while(Bytes>=4)
		{
		Word = theMemory->ReadMemory32((uint64)BytePtr);
		(*dWordPtr) = Word;
		++dWordPtr;
		BytePtr += 4;dBytePtr += 4;
		Bytes -= 4;
		}
*/		
	while(Bytes)
		{
		Byte = theMemory->ReadMemory8((uint64)BytePtr);
		(*dBytePtr) = Byte;
		++dBytePtr;
		++BytePtr;
		--Bytes;
		}
	return ( true );
	}


bool
AlphaProcessor	::	CopyStringFromHost(
	uint64	Destination,
	void	*Source,
	uint32	MaxBytes)
	{
	uint8	*BytePtr = (uint8*)Source;
	uint64	dDestination = (uint64)Destination;

	while(MaxBytes)
		{
		theMemory->WriteMemory8((uint64)dDestination, (uint8)*BytePtr);
		if(!(*BytePtr))
			break;
		--MaxBytes;
		++BytePtr;
		++dDestination;
		}
	return ( true );
	}

bool
AlphaProcessor	::	CopyStringToHost(
	void	*Destination,
	uint64	Source,
	uint32	MaxBytes)
	{
	uint64	BytePtr = (uint64)Source;
	uint8	*dBytePtr = (uint8*)Destination;
	uint8	Byte;

	while(MaxBytes)
		{
		Byte = theMemory->ReadMemory8((uint64)BytePtr);
		(*dBytePtr) = Byte;
		if(!Byte)
			break;
		++dBytePtr;
		++BytePtr;
		--MaxBytes;
		}
	return ( true );
	}






/******************************** Start of the actual execution engine *****************/

bool
AlphaProcessor	::	Initialize(
	char	*filename,
	char	**argv,
	char	**argp)
	{
	int	Index;

	InitializeAlphaDecodingEngine();
	InitializeAlphaExecutionEngine();

	// zero all the registers
	for(Index = 0;Index < 32; Index++) GeneralPurposeRegisters[Index] = 0;
	for(Index = 0;Index < 32; Index++) FloatingPointRegisters[Index] = 0;
	ProgramCounter = 0;
	FloatingPointControlRegister = 0;

	// initialize the memory space
	theMemory = new Sparse64BitMemorySpace;
	theMemory->Initialize (16, 16, 16, 16);

	if ( ! Loader::LoadFromDevice(filename, this, argv, argp) )
		{
		printf("Error loading %s\n", filename);
		return ( false );
		}

	// initialize the system call interface
	//theSystemTrapHandler.Initialize(true);
	theSystemTrapHandler.Initialize(false);

	IsSpeculative = false;

	InstructionSequence = 0;
	LoadStoreSequence = 0;
	StoreSequence = 0;

	EndSimulation = false;
	CycleCounter = 0;
	return ( true );
	}

// this just fills in theInstruction with information
bool
AlphaProcessor	::	DecodeInstruction (
	uint32				RawInstruction,
	AlphaInstruction	*theInstruction)
	{

	theInstruction->RawInstruction = RawInstruction;

	uint32	OpCode1 = OpCodeField(RawInstruction);
	uint32	OpCode2;

	AlphaOp			*theOp = &TopLevelTable[OpCode1];

	if ( theOp->OpCode == NoCode )
		{
		theInstruction->theOpCode = NoCode;
		if ( ! IsSpeculative )
			{
			printf("unknown opcode: %x\n", OpCode1);
			return(false);
			}
		return ( false );
		}
	// in the general case, this would be a while loop, however, with Alpha only
	// 2 layers are requird (at most).  So, it's a litte simplier.

	if ( theOp->TableLink )
		{
		theOp = &theOp->TableLink[ OpCode2 = theOp->DecodeLink( RawInstruction ) ];
		if ( theOp->OpCode == NoCode )
			{
			if ( ! IsSpeculative )
				{
				printf("unknown opcode %x.%x\n", (uint32)OpCode1, (uint32) OpCode2 );
				return(false);
				}
			return ( false );
			}
		}

	// at this point theOp points to a known instruction (but not one that is assured to be implemented)
	theInstruction->theFormat = theOp->Format;
	theInstruction->theOpCode = theOp->OpCode;
	theInstruction->Type = theOp->Type;
	theInstruction->Ra = RegisterField_Ra(RawInstruction);
	theInstruction->Rb = RegisterField_Rb(RawInstruction);
	theInstruction->Rc = RegisterField_Rc(RawInstruction);
	theInstruction->MemoryDisplacement = MemoryDisplacementField(RawInstruction);
	theInstruction->Function = FunctionField(RawInstruction);
	theInstruction->OpFunction = OpFunctionField(RawInstruction);
	theInstruction->BranchDisplacement = BranchDisplacementField(RawInstruction);
	theInstruction->PALCode = PALCodeField(RawInstruction);
	theInstruction->ReadOrWrite = 0;
	theInstruction->MemoryAddress = 0;

	if ( LiteralOrNotField(RawInstruction) )
		theInstruction->LiteralOrNot = true;
	else
		theInstruction->LiteralOrNot = false;
	theInstruction->Literal = LiteralField(RawInstruction);
	return ( true );
	}

int
AlphaProcessor	::	SetOutputRegister(
	AlphaInstruction	*theInstruction,
	int		RegisterField)
	{
	switch ( RegisterField )
		{
		case	None:	return(NoRegister);
		case	Ra:	return(theInstruction->Ra);
		case	Rb:	return(theInstruction->Rb);
		case	Rbl:	if ( theInstruction->LiteralOrNot ) return(NoRegister); /* else */ return(theInstruction->Rb);
		case	Rc:	return(theInstruction->Rc);
		case	Fa:	return(theInstruction->Ra + 32);
		case	Fb:	return(theInstruction->Rc + 32);
		case	Fc:	return(theInstruction->Rc + 32);
		case	LockFlag:	return(NoRegister);
		case	Fpcr:	return(FlagRegister);
		}
	return(NoRegister);
	}

bool
AlphaProcessor	::	DecodeInstruction (
	uint32			RawInstruction,
	Instruction		*theInstruction )
	{
//	if ( ! DecodeInstruction(RawInstruction, &theInstruction->theAlphaInstruction) )
//		return ( false );

	theInstruction->Operation = theInstruction->theAlphaInstruction.Type;

	if ( ReverseMapping[theInstruction->theAlphaInstruction.theOpCode] )
		{
		theInstruction->rOutput1=theInstruction->Output1=
			SetOutputRegister(&theInstruction->theAlphaInstruction,ReverseMapping[theInstruction->theAlphaInstruction.theOpCode]->Outputs[0]);
		theInstruction->rOutput2=theInstruction->Output2=
			SetOutputRegister(&theInstruction->theAlphaInstruction,ReverseMapping[theInstruction->theAlphaInstruction.theOpCode]->Outputs[1]);
		theInstruction->rInput1=theInstruction->Input1=
			SetOutputRegister(&theInstruction->theAlphaInstruction,ReverseMapping[theInstruction->theAlphaInstruction.theOpCode]->Inputs[0]);
		theInstruction->rInput2=theInstruction->Input2=
			SetOutputRegister(&theInstruction->theAlphaInstruction,ReverseMapping[theInstruction->theAlphaInstruction.theOpCode]->Inputs[1]);
		theInstruction->rInput3=theInstruction->Input3=
			SetOutputRegister(&theInstruction->theAlphaInstruction,ReverseMapping[theInstruction->theAlphaInstruction.theOpCode]->Inputs[2]);
		}
	else
		{
		theInstruction->rOutput1=theInstruction->Output1=NoRegister;
		theInstruction->rOutput2=theInstruction->Output2=NoRegister;
		theInstruction->rInput1=theInstruction->Input1=NoRegister;
		theInstruction->rInput2=theInstruction->Input2=NoRegister;
		theInstruction->rInput3=theInstruction->Input3=NoRegister;
		}

	theInstruction->MemoryAddress = theMemory->GetLastTransactionAddress();
	theInstruction->MemoryTransactionSize = theMemory->GetLastTransactionSize();

	theInstruction->Finalize();
	
	}


bool
AlphaProcessor	::	DumpInstruction (
	AlphaInstruction	*theInstruction,
	FILE				*Device)
	{
	fpReg	fpRegs[3];

	if ( theInstruction->theOpCode == NoCode )
		{
		fprintf(Device, "\t-- invalid instruction: %8.8x\n", theInstruction->RawInstruction);
		return(false);
		}
	fprintf(Device, "" format_hex_uint64 ": %8.8x:\t%10s\t\t", theInstruction->ProgramCounter, theInstruction->RawInstruction,
			ReverseMapping[ theInstruction->theOpCode ]->Name );


	switch ( theInstruction->theFormat )
		{
		case	UnknownFormat:
			break;
		case	MemoryInstruction:
			fprintf(Device, "%s, %s + #" format_dec_int64 "\t: %s=" format_hex_uint64 ", %s=" format_hex_uint64 ":\tmem_addr=" format_hex_uint64 "",
				AlphaRegisterNames[theInstruction->Ra],
				AlphaRegisterNames[theInstruction->Rb],
				theInstruction->MemoryDisplacement,
				AlphaRegisterNames[theInstruction->Ra], GeneralPurposeRegisters[theInstruction->Ra],
				AlphaRegisterNames[theInstruction->Rb], GeneralPurposeRegisters[theInstruction->Rb],
				theInstruction->MemoryAddress);
			break;
		case	FloatingPointMemoryInstruction:
			fpRegs[0].Asuint64 = FloatingPointRegisters[theInstruction->Ra];
			fprintf(Device, "f%d, %s + #" format_dec_int64 "\t: f%d=" format_hex_uint64 " (%f,%f), %s=" format_hex_uint64 ":\tmem_addr=" format_hex_uint64 "",
				theInstruction->Ra,
				AlphaRegisterNames[theInstruction->Rb],
				theInstruction->MemoryDisplacement,
				theInstruction->Ra, fpRegs[0].Asuint64, fpRegs[0].Asdouble, fpRegs[0].Asfloat,
				AlphaRegisterNames[theInstruction->Rb], GeneralPurposeRegisters[theInstruction->Rb],
				theInstruction->MemoryAddress);
			break;
		case	MemoryWithFunctionInstruction:
			fprintf(Device, "%s, %s\t: %s=" format_hex_uint64 ", %s=" format_hex_uint64 "",
				AlphaRegisterNames[theInstruction->Ra],
				AlphaRegisterNames[theInstruction->Rb],
				AlphaRegisterNames[theInstruction->Ra], GeneralPurposeRegisters[theInstruction->Ra],
				AlphaRegisterNames[theInstruction->Rb], GeneralPurposeRegisters[theInstruction->Rb]);
			break;
		case	BranchInstruction:
			fprintf(Device, "%s, #" format_dec_int64 "\t: %s=" format_hex_uint64 " (" format_dec_int64 ")",
				AlphaRegisterNames[theInstruction->Ra],
				theInstruction->BranchDisplacement,
				AlphaRegisterNames[theInstruction->Ra],
				GeneralPurposeRegisters[theInstruction->Ra],
				GeneralPurposeRegisters[theInstruction->Ra]);
			break;
		case	FloatingPointBranchInstruction:
			fpRegs[0].Asuint64 = FloatingPointRegisters[theInstruction->Ra]; 
			fprintf(Device, "f%d, #" format_dec_int64 "\t: f%d=%f (%f)",
				theInstruction->Ra,
				theInstruction->BranchDisplacement,
				theInstruction->Ra,
				fpRegs[0].Asdouble, fpRegs[0].Asfloat);
			break;
		case	OperateInstruction:
			if ( theInstruction->LiteralOrNot )
				fprintf(Device, "%s <- %s, #" format_hex_uint64 " (" format_dec_uint64 ")\t: %s=" format_hex_uint64 " (" format_dec_int64 "), %s=" format_hex_uint64 " (" format_dec_int64 ")",
					AlphaRegisterNames[theInstruction->Rc],
					AlphaRegisterNames[theInstruction->Ra],
					theInstruction->Literal,
					theInstruction->Literal,
					AlphaRegisterNames[theInstruction->Rc],
						GeneralPurposeRegisters[theInstruction->Rc],
						GeneralPurposeRegisters[theInstruction->Rc],
					AlphaRegisterNames[theInstruction->Ra],
						GeneralPurposeRegisters[theInstruction->Ra],
						GeneralPurposeRegisters[theInstruction->Ra]);
			else
				fprintf(Device, "%s <- %s, %s\t: %s=" format_hex_uint64 " (" format_dec_int64 "), %s=" format_hex_uint64 " (" format_dec_int64 "), %s=" format_hex_uint64 " (" format_dec_int64 ")",
					AlphaRegisterNames[theInstruction->Rc],
					AlphaRegisterNames[theInstruction->Ra],
					AlphaRegisterNames[theInstruction->Rb],
					AlphaRegisterNames[theInstruction->Rc],
						GeneralPurposeRegisters[theInstruction->Rc],
						GeneralPurposeRegisters[theInstruction->Rc],
					AlphaRegisterNames[theInstruction->Ra],
						GeneralPurposeRegisters[theInstruction->Ra],
						GeneralPurposeRegisters[theInstruction->Ra],
					AlphaRegisterNames[theInstruction->Rb],
						GeneralPurposeRegisters[theInstruction->Rb],
						GeneralPurposeRegisters[theInstruction->Rb]);
			break;
		case	FloatingPointInstruction:
			fpRegs[0].Asuint64 = FloatingPointRegisters[theInstruction->Ra]; 
			fpRegs[1].Asuint64 = FloatingPointRegisters[theInstruction->Rb]; 
			fpRegs[2].Asuint64 = FloatingPointRegisters[theInstruction->Rc]; 

			fprintf(Device, "f%d <- f%d, f%d\t: f%d=" format_hex_uint64 " (%f,%f), f%d=" format_hex_uint64 " (%f,%f) f%d=" format_hex_uint64 " (%f,%f)",
					theInstruction->Rc,
					theInstruction->Ra,
					theInstruction->Rb,
					theInstruction->Rc, fpRegs[2].Asuint64, fpRegs[2].Asdouble, fpRegs[2].Asfloat,
					theInstruction->Ra, fpRegs[0].Asuint64, fpRegs[0].Asdouble, fpRegs[0].Asfloat,
					theInstruction->Rb, fpRegs[1].Asuint64, fpRegs[1].Asdouble, fpRegs[1].Asfloat);
			break;
		case	PALInstruction:
			fprintf(Device, "#%d",
				theInstruction->PALCode);
			break;
		}

	fprintf(Device, "\n");
	return ( true );
	}

bool
AlphaProcessor	::	HandleCallPAL(AlphaInstruction	*ai)
	{
	if( IsSpeculative )
		return(false);

	if ( ai->PALCode == 131 )
		theSystemTrapHandler.HandleSystemTrap(this, ai);
	return(true);
	}

bool
AlphaProcessor	::	EnterSpeculativeMode(
	uint64	AtThisProgramCounter )
	{
	if ( ! IsSpeculative )
		{
		memcpy(nscGeneralPurposeRegisters, GeneralPurposeRegisters, 32*sizeof(uint64));
		memcpy(nscFloatingPointRegisters, FloatingPointRegisters, 32*sizeof(uint64)); 
		nscProgramCounter = ProgramCounter;
		nscNextProgramCounter = NextProgramCounter;
		nscFloatingPointControlRegister = FloatingPointControlRegister;
		nscInstructionSequence = InstructionSequence;
		nscLoadStoreSequence = LoadStoreSequence;
		nscStoreSequence = StoreSequence;
		}
	theMemory->IgnoreWrites();
	IsSpeculative = true;

	ProgramCounter = AtThisProgramCounter;

	return ( true );
	}

bool
AlphaProcessor	::	ReturnToNonSpeculativeMode()
	{
	if ( ! IsSpeculative )	return(false);	// ?

	memcpy(GeneralPurposeRegisters, nscGeneralPurposeRegisters, 32*sizeof(uint64));
	memcpy(FloatingPointRegisters, nscFloatingPointRegisters, 32*sizeof(uint64)); 
	ProgramCounter = nscProgramCounter;
	NextProgramCounter = nscNextProgramCounter;
	FloatingPointControlRegister = nscFloatingPointControlRegister;
	InstructionSequence = nscInstructionSequence;
	LoadStoreSequence = nscLoadStoreSequence;
	StoreSequence = nscStoreSequence;

	theMemory->DoNotIgnoreWrites();
	IsSpeculative = false;
	return ( true );
	}

bool
AlphaProcessor	::	Test(bool	Debug)
	{
	uint32				RawInstruction;
	uint64				InstructionsExecuted = 0;
	AlphaInstruction	ai;

	CycleCounter = 0;
	EndSimulation = false;
	while ( ! EndSimulation )
		{
		RawInstruction = theMemory->ReadMemory32(ProgramCounter);
//		ai.theOpCode = NoCode;
		memset(&ai, 0, sizeof(AlphaInstruction));

		if ( ! DecodeInstruction ( RawInstruction, &ai) )
			{
			printf("Decoding instruction failed\n");
			return(false);
			}
		ai.ProgramCounter = ProgramCounter;
		NextProgramCounter = ai.NextProgramCounter = ProgramCounter + 4;

		if(Debug)	
			DumpInstruction	( &ai, stdout );

		if ( ! ReverseMapping[ ai.theOpCode ] )
			{
			DumpInstruction	( &ai, stdout );
			printf("-- error in decoding\n");
			return(false);
			}
		if ( ! ReverseMapping[ ai.theOpCode ]->ImplementationFunction )
			{
			printf("-- op code not implemented!\n");
			DumpInstruction	( &ai, stdout );
			return(false);
			}
		GeneralPurposeRegisters[zero] = 0;
		FloatingPointRegisters[zero] = 0;
		ReverseMapping[ ai.theOpCode ]->ImplementationFunction(this, RawInstruction, &ai);
		ProgramCounter = NextProgramCounter;

		if(Debug)
			{
			printf("-->");
			DumpInstruction	( &ai, stdout );
			}
		++InstructionsExecuted;
		++CycleCounter;
		}
//	printf("InstructionsExecuted = " format_dec_int64 "\n", InstructionsExecuted);
	return(true);
	}

bool
AlphaProcessor	::	NextInstruction(
	Instruction	**theInstructionPtr)
	{
	if ( EndSimulation )
		return ( false );

    (*theInstructionPtr)->RawSequence = InstructionSequence;
    (*theInstructionPtr)->RawLoadStoreSequence = LoadStoreSequence;
    (*theInstructionPtr)->RawStoreSequence = StoreSequence;
    ++InstructionSequence;

	uint32				RawInstruction;
	RawInstruction = theMemory->ReadMemory32(ProgramCounter);
	AlphaInstruction	*aip = &(*theInstructionPtr)->theAlphaInstruction;

	memset(aip, 0, sizeof(AlphaInstruction));

	if ( ! DecodeInstruction ( RawInstruction, aip) )
		{
		if ( ! IsSpeculative )
			{
			printf("Decoding instruction failed\n");
			return(false);
			}
		else
			{
			RawInstruction = AlphaNOPInstruction;
			DecodeInstruction(RawInstruction, aip);
			}
		}
	aip->ProgramCounter = ProgramCounter;
	NextProgramCounter = aip->NextProgramCounter = ProgramCounter + 4;

	if ( ! ReverseMapping[ aip->theOpCode ] )
		{
		DumpInstruction	( aip, stdout );
		printf("-- error in decoding\n");
		return(false);
		}
	if ( ! ReverseMapping[ aip->theOpCode ]->ImplementationFunction )
		{
		if ( ! IsSpeculative )
			{
			printf("-- op code not implemented!\n");
			DumpInstruction	( aip, stdout );
			return(false);
			}
		else
			{
			DecodeInstruction ( AlphaNOPInstruction, aip);
			DecodeInstruction(AlphaNOPInstruction, (*theInstructionPtr));
			}
		}
	GeneralPurposeRegisters[zero] = 0;
	FloatingPointRegisters[zero] = 0;
	ReverseMapping[ aip->theOpCode ]->ImplementationFunction(this, RawInstruction, aip);
	GeneralPurposeRegisters[zero] = 0;
	FloatingPointRegisters[zero] = 0;


	if ( (*theInstructionPtr)->Operation == OpStore )
		++StoreSequence;
	if ( ((*theInstructionPtr)->Operation == OpStore) || ((*theInstructionPtr)->Operation == OpLoad) )
		++LoadStoreSequence;

	DecodeInstruction(RawInstruction, (*theInstructionPtr));
	(*theInstructionPtr)->ProgramCounter = ProgramCounter;
	(*theInstructionPtr)->NextProgramCounter = NextProgramCounter;

	ProgramCounter = NextProgramCounter;

	// hack
	if ( (*theInstructionPtr)->Operation == OpSystem )
		(*theInstructionPtr)->Operation = OpInteger;

	++CycleCounter;
	return(true);
	}


