enum	AlphaRegister
	{
	v0 = 0,
	r1,r2,r3,r4,r5,r6,
	err = 7,
	r8,r9,r10,r11,r12,r13,r14,
	fp = 15,
	a0, a1, a2, a3, a4, a5,
	r22,r23,r24,r25,
	ra = 26,
	r27,r28,
	gp = 29,
	sp = 30,
	zero = 31
	};

extern	char	*AlphaRegisterNames[];

class	AlphaInstruction;
struct	_AlphaOp;
	
class	AlphaProcessor	:
	public	Engine
	{
	public:
		static	bool			DecodingEngineIsInitialized;
		static	bool			ExecutionEngineIsInitialized;
	
	public:
		bool					EndSimulation;
		uint64					CycleCounter;

		// Alpha AXP processor state
		uint64					GeneralPurposeRegisters[32];
		uint64					FloatingPointRegisters[32];
		uint64					ProgramCounter;
		uint64					NextProgramCounter;
		uint64					FloatingPointControlRegister;
		uint64					InstructionSequence;
		uint64					LoadStoreSequence;
		uint64					StoreSequence;

		bool					IsSpeculative;

		// A Non-speculative copy of it.
		uint64					nscGeneralPurposeRegisters[32];
		uint64					nscFloatingPointRegisters[32];
		uint64					nscProgramCounter;
		uint64					nscNextProgramCounter;
		uint64					nscFloatingPointControlRegister;
		uint64					nscInstructionSequence;
		uint64					nscLoadStoreSequence;
		uint64					nscStoreSequence;

		// program memory
		Sparse64BitMemorySpace	*theMemory;

		// the "operating system"
		OSFSystemTrapHandler	theSystemTrapHandler;

		virtual			~AlphaProcessor() { };
		// support functions for accessing the memory space
		bool			CopyFromHost(uint64	SimulatedTargetAddress, void	*data, uint32	Size);
		bool			CopyToHost(void *, uint64, uint32);
		bool			CopyStringFromHost(uint64, void *, uint32);
		bool			CopyStringToHost(void *, uint64, uint32);

		// Used by the program loader
		inline	uint64	*GetIntegerRegisterSet() { return ( GeneralPurposeRegisters ); }
		inline	uint64	*GetProgramCounter() { return(&ProgramCounter); }

		static	bool			InitializeAlphaDecodingEngine();
		static	bool			InitializeAlphaExecutionEngine();

		bool			Initialize(char	*filename, char	**argv, char	**argp);

		int				SetOutputRegister(AlphaInstruction	*, int);
		bool			DecodeInstruction(uint32	RawInstruction, Instruction	*);
		bool			DecodeInstruction(uint32	RawInstruction, AlphaInstruction	*);

		bool			DumpInstruction(AlphaInstruction	*, FILE	*fp);

		bool			HandleCallPAL(AlphaInstruction	*);

		bool			Halt()	{ EndSimulation = true; return ( true ); }
		uint64			GetCycleCounter()	{	return(CycleCounter);	}

		struct	_AlphaOp	*HighJackOpCode(uint32	TopLevel, uint32	NextLevel);

		bool			Test(bool	DebugMode);

		bool			EnterSpeculativeMode(uint64	AtThisProgramCounter);
		bool			ReturnToNonSpeculativeMode();
		bool			NextInstruction(Instruction	**theInstruction);

	};


enum	AlphaInstructionFormat
	{
	UnknownFormat		=	0,
	MemoryInstruction,
	FloatingPointMemoryInstruction,
	MemoryWithFunctionInstruction,
	BranchInstruction,
	FloatingPointBranchInstruction,
	OperateInstruction,
	FloatingPointInstruction,
	PALInstruction
	};

// these are used to map input and output registers for decoding
enum	AlphaRegisterDecodeNames
	{
	None = -1,
	Ra = 1, Rb, Rbl, Rc, Fa, Fb, Fc, LockFlag, Fpcr
	};

// opcode table
enum	AlphaOpCode
	{
	NoCode	=	0,
#	define	DirectCode(code, format, type, name, outputs, inputs)	name ,
#	define	LinkCode(code, name)									name ,

#	include	"AllTables.h"

#	undef	DirectCode
#	undef	LinkCode
	InstructionCodes
	};

class	AlphaInstruction
	{
	public:
		uint32					RawInstruction;
		uint64					ProgramCounter;
		uint64					NextProgramCounter;
		AlphaInstructionFormat	theFormat;
		AlphaOpCode				theOpCode;
		int						Type;
	
		int						Ra;
		int						Rb;
		int						Rc;
		int64					MemoryDisplacement;
		uint32					Function;
		uint32					OpFunction;
		int64					BranchDisplacement;
		bool					LiteralOrNot;
		uint64					Literal;
		uint32					PALCode;
		uint64					MemoryAddress;
		int						ReadOrWrite;
	};

typedef	uint32	DecodeFunction(uint32);
#	define	ImplementorArguments	AlphaProcessor	*Proc, uint32	RawInstruction, AlphaInstruction *ai
typedef	void	ImplementorFunction(ImplementorArguments);

typedef	struct	_AlphaOp
	{
	char		*Name;
	AlphaOpCode	OpCode;
	unsigned	int	BitCode;
	AlphaInstructionFormat	Format;
	unsigned	int	Type;
	int			Outputs[3];	// yes 3 (only 2, but last = -1)
	int			Inputs[4];	// see above
	struct		_AlphaOp	*TableLink;
	DecodeFunction			*DecodeLink;
	ImplementorFunction		*ImplementationFunction;
	}	AlphaOp;


