class	Sparse64BitMemorySpace
	{
	public:
		uint64		Spaces;
		uint64		Worlds;
		uint64		Pages;
		uint64		PageSize;
		uint64		SpaceBits;
		uint64		WorldBits;
		uint64		PageBits;
		uint64		PageSizeBits;

		uint64		SpaceShift;
		uint64		WorldShift;
		uint64		PageShift;
		uint64		SpaceMask;
		uint64		WorldMask;
		uint64		PageMask;
		uint64		PageSizeMask;
		Counter		Accesses;
		Counter		BytesRead;
		Counter		BytesWritten;
		uint8		****SpaceArray;
		bool		theIgnoreWrites;

		uint64		AllocatedSize;

		uint		LastTransactionSize;
		uint64		LastTransactionAddress;

	public:
		Sparse64BitMemorySpace()
			{
			Spaces = Worlds = Pages = PageSize = 0;
			SpaceArray = NULL;
			theIgnoreWrites = false;
			AllocatedSize = 0;
			}

		~Sparse64BitMemorySpace()
			{
			Clear();
			}

		bool	Clear()
			{
			uint64	Index;

			AllocatedSize = 0;
			if(SpaceArray)
				{
				for(Index=0;Index<Spaces;Index++)
					{
					if(SpaceArray[Index])
						{
						uint64	Index2;
						for(Index2 = 0;Index2<Worlds;Index2++)
							{
							uint64	Index3;
							for(Index3 = 0;Index3<Pages;Index3++)
								{
								if ( SpaceArray[Index][Index2][Index3] )
									{
									delete SpaceArray[Index][Index2][Index3];
									SpaceArray[Index][Index2][Index2] = NULL;
									}
								}
							delete SpaceArray[Index][Index2];
							}
						}
					SpaceArray[Index] = NULL;
					}
				delete SpaceArray;
				SpaceArray = NULL;
				}
			Spaces = Worlds = Pages = PageSize = 0;
			SpaceBits = WorldBits = PageBits = PageSizeBits = 0;

			BytesRead = 0;
			BytesWritten = 0;
			Accesses = 0;

			return ( true );	
			}

		bool	Initialize(
			uint64	aSpaceBits,
			uint64	aWorldBits,
			uint64	aPageBits,
			uint64	aPageSizeBits)
			{
			uint64	Index;
	
			Clear();

			if ( (aSpaceBits + aWorldBits + aPageBits + aPageSizeBits) != 64 )
				{
				printf("Error " format_dec_uint64 ":" format_dec_uint64 ":" format_dec_uint64 ":" format_dec_uint64 " != 64 bits\n", aSpaceBits, aWorldBits, aPageBits, aPageSizeBits);
				return(false);
				}
			SpaceBits = aSpaceBits;
			WorldBits = aWorldBits;
			PageBits = aPageBits;
			PageSizeBits = aPageSizeBits;

			Pages = (1<<(PageBits));
			Worlds = (1<<(WorldBits));
			Spaces = (1<<(SpaceBits));
			
			PageSize = (1<<(PageSizeBits));
		
			SpaceShift = WorldBits + PageBits + PageSizeBits;
			SpaceMask = (1<<SpaceBits)-1;
			PageShift = PageSizeBits;
			PageMask = (1<<PageBits)-1;
			WorldShift = PageBits + PageSizeBits;
			WorldMask = (1<<WorldBits)-1;
			PageSizeMask = (1<<PageSizeBits)-1;

			//printf("SpaceBits:    " format_dec_int64 " SpaceShift " format_dec_int64 " SpaceMask: " format_hex_uint64 "\n", SpaceBits, SpaceShift, SpaceMask);
			//printf("WorldBits:    " format_dec_int64 " WorldShift " format_dec_int64 " WorldMask: " format_hex_uint64 "\n", WorldBits, WorldShift, WorldMask);
			//printf("PageBits:     " format_dec_int64 "  PageShift " format_dec_int64 "  PageMask: " format_hex_uint64 "\n", PageBits, PageShift, PageMask);
			//printf("PageSizeBits: " format_dec_int64 "            PageSizeMask: " format_hex_uint64 "\n", PageSizeBits, PageSizeMask);

			SpaceArray = new uint8 ***[Spaces];
			Index = 0;
			for(Index=0;Index<Spaces;Index++)
				SpaceArray[Index] = NULL;

			return ( true );
			}

		bool		IgnoreWrites()
			{
			theIgnoreWrites = true;
			return(true);
			}
		bool		DoNotIgnoreWrites()
			{
			theIgnoreWrites = false;
			return(true);
			}

		uint8	*	GetPage (uint64	sa)
			{
			uint64	tIndex = (sa >> SpaceShift) & SpaceMask;
			uint64	tIndex2 = (sa >> WorldShift) & WorldMask;
			uint64	tIndex3 = (sa >> PageShift) & PageMask;
			uint64	Index2, Index3;

			if(theIgnoreWrites)
				return(NULL);	// speculative mode -- bad address

			if(!SpaceArray[tIndex])
				{
				SpaceArray[tIndex] = new uint8 **[Worlds];
				for(Index2=0;Index2<Worlds;Index2++)
					SpaceArray[tIndex][Index2] = NULL;
				}

			if (!SpaceArray[tIndex][tIndex2])
				{
				SpaceArray[tIndex][tIndex2] = new uint8 *[Pages];
				for(Index3=0;Index3<Pages;Index3++)
					SpaceArray[tIndex][tIndex2][Index3] = NULL;
				}
			if ( !SpaceArray[tIndex][tIndex2][tIndex3] )
				{
				SpaceArray[tIndex][tIndex2][tIndex3] = new uint8 [PageSize];
				memset(SpaceArray[tIndex][tIndex2][tIndex3], 0x00, PageSize);
				AllocatedSize = AllocatedSize + PageSize;
				if ( AllocatedSize > 0x10000000 )
					{
					printf("Allocated size > " format_dec_int64 "KB (limit set in %s at line %d)\n", AllocatedSize / 1024, __FILE__, __LINE__ - 2);
					exit(1);
					}
				}

			return(SpaceArray[tIndex][tIndex2][tIndex3]);
			}
		uint		GetLastTransactionSize()
			{
			return ( LastTransactionSize );
			}
		uint64		GetLastTransactionAddress()
			{
			return ( LastTransactionAddress );
			}
	
		uint8		ReadMemory8(uint64	sa)
			{
			LastTransactionSize = 1;
			LastTransactionAddress = sa;
			uint8	*Page = GetPage(sa);
			if(!Page)	return(0);

			return(Page[sa&PageSizeMask]);
			}

		uint16		ReadMemory16(uint64	sa)
			{
			LastTransactionSize = 2;
			LastTransactionAddress = sa;
			if(!theIgnoreWrites)
			if(sa & 0x1)	Warning("Alignment error");

			uint8	*Page = GetPage(sa);
			if(!Page)	return(0);

			Page = &Page[sa&PageSizeMask];
			le_uint16	*Temp = (le_uint16*)Page;
			return(*Temp);
			}

		uint32		ReadMemory32(uint64	sa)
			{
			LastTransactionSize = 4;
			LastTransactionAddress = sa;
			if(!theIgnoreWrites)
			if(sa & 0x3)
				{
				//Warning("Alignment error");
				printf("Alignment error in ReadMemory32(address = " format_hex_uint64 ")\n", sa);
				}

			uint8	*Page = GetPage(sa);
			if(!Page)	return(0);

			Page = &Page[sa&PageSizeMask];
			le_uint32	*Temp = (le_uint32*)Page;
			return(*Temp);
			}

		uint64		ReadMemory64(uint64	sa)
			{
			LastTransactionSize = 8;
			LastTransactionAddress = sa;
			if(!theIgnoreWrites)
			if(sa & 0x7)	Warning("Alignment error");

			uint8	*Page = GetPage(sa);
			if(!Page)	return(0);

			Page = &Page[sa&PageSizeMask];
			le_uint64	*Temp = (le_uint64*)Page;
			//printf("Reading address " format_hex_uint64 " value=" format_hex_uint64 "\n", sa, *Temp);
			return(*Temp);
			}

		bool		WriteMemory8(uint64	sa, uint8		Data)
			{
			LastTransactionSize = 1;
			LastTransactionAddress = sa;
			if(theIgnoreWrites)	return(true);
			uint8	*Page = GetPage(sa);
			if(!Page)	return(false);

			Page[sa&PageSizeMask] = Data;
			return ( true );
			} 

		bool		WriteMemory16(uint64	sa, uint16		Data)
			{
			LastTransactionSize = 2;
			LastTransactionAddress = sa;
			if(theIgnoreWrites)	return(true);
			if(sa & 0x1)	Warning("Alignment error");

			uint8	*Page = GetPage(sa);
			if(!Page)	return(false);

			Page = &Page[sa&PageSizeMask];
			le_uint16	*Temp = (le_uint16*)Page;
			(*Temp) = Data;
			return ( true );
			}

		bool		WriteMemory32(uint64	sa, uint32		Data)
			{
			LastTransactionSize = 4;
			LastTransactionAddress = sa;
			if(theIgnoreWrites)	return(true);
			if(sa & 0x3)	Warning("Alignment error");

			uint8	*Page = GetPage(sa);
			if(!Page)	return(false);

			Page = &Page[sa&PageSizeMask];
			le_uint32	*Temp = (le_uint32*)Page;
			(*Temp) = Data;
			return ( true );
			}

		bool		WriteMemory64(uint64	sa, uint64		Data)
			{
			LastTransactionSize = 8;
			LastTransactionAddress = sa;
			if(theIgnoreWrites)	return(true);
			if(sa & 0x7)	Warning("Alignment error");

			uint8	*Page = GetPage(sa);
			if(!Page)	return(false);
			
			Page = &Page[sa&PageSizeMask];
			le_uint64	*Temp = (le_uint64*)Page;
			(*Temp) = Data;
			return ( true );
			}
	};
