#	include	"Standard.h"

static	char
UpperCase(char	ch)
	{
	if((ch>='a')&&(ch<='z'))
		return(ch-'a'+'A');
	return(ch);
	}

static	bool
AreEqualPostfix(
	char	*s1,
	char	*s2)
	{
	if ( (!s1)||(!s2) )	return(false);

	int	i1 = strlen(s1);
	int	i2 = strlen(s2);
	while( (i1>=0) && (i2>=0) )
		{
		if(UpperCase(s1[i1])!=UpperCase(s2[i2]))
			return(false);
		--i1;
		--i2;
		}
	return(true);
	}

static	bool
AreEqual(
	char	*s1,
	char	*s2,
	int		PostFixOnly = 0)
	{
	if ( PostFixOnly )
		return ( AreEqualPostfix(s1, s2) );

	while((*s1)&&(*s2))
		{
		if(UpperCase(*s1)!=UpperCase(*s2))
			return(false);
		++s1;
		++s2;
		}
	if(*s1)
		return(false);
	if(*s2)
		return(false);
	return(true);
	}

bool
PrintHelp(
	CommandLineArgument	*cla)
	{
	printf(
"\n"
"Parameters are set by using -Parameter:value strings.  The \"Parameter\" can be the entire parameter\n"
"name or just the (unique) postfix, i.e. \"-root.foobar:1\" can be abbreviated with \"-bar:1\"\n"
"\n"
"The value field can be in any reasonable format (decimal, hex, true/false).\n"
"\n"
);

	int		Index = 0;
	char	s[128];

	int		*iType;
	uint	*uType;
	bool	*bType;
	char	*sType;
	double	*dType;
	int64	*llType;
	uint64	*ulType;

	sprintf(s, "  %80s:<filename>\n", "-Source");
	printf("%s", s);
	while ( cla[Index].Name )
		{
		sprintf(s, "-%-80s", cla[Index].Name);
		printf(" %-80s:", s);
		switch(cla[Index].theType)
			{
			case	typeInteger:		printf("number   ");break;
			case	typeuInteger:	printf("unumber  ");break;
			case	typeBool:		printf("bool     ");break;
			case	typeString:		printf("string   ");break;
			case	typeDouble:		printf("double   ");break;
			case	typeInt64:		printf("int64    ");break;
			case	typeuInt64:		printf("uint64   ");break;
			default:
				printf("error");
			}
		printf("default=");
		iType = (int*) cla[Index].Ptr;
		uType = (uint*) cla[Index].Ptr;
		bType = (bool*) cla[Index].Ptr;
		sType = (char*) cla[Index].Ptr;
		dType = (double*) cla[Index].Ptr;
		llType = (int64 *) cla[Index].Ptr;
		ulType = (uint64 *) cla[Index].Ptr;

		switch(cla[Index].theType)
			{
			case	typeInteger:	printf("%10d  (0x%8.8x)", *iType, *iType);break;
			case	typeuInteger:printf("%10u  (0x%8.8x)", *uType, *uType);break;
			case	typeBool:
				if(*bType)
					printf("true");
				else
					printf("false");
				break;
			case	typeString: printf("%s", sType); break;
			case	typeDouble: printf("%f", (double) *dType);break;
			case	typeInt64: printf("" format_dec_int64 " (0x" format_hex_uint64 ")", (int64) *llType, (int64)*llType);break;
			case	typeuInt64: printf("" format_dec_uint64 " (0x" format_hex_uint64 ")", (uint64) *ulType, (uint64)*ulType);break;
			default:
				printf("error");
			}
		printf("\n");
		++Index;
		}
	return(true);
	}

static	bool
LocateAndSetArgument(
	CommandLineArgument	*cla,
	char				*Parameter,
	char				*Value,
	int					PostFixOnly = 0,
	int					OnErrorFlag = 0)
	{
	int		Index = 0;

	int		*iType;
	uint	*uType;
	bool	*bType;
	char	*sType;
	double	*dType;
	int64	*llType;
	uint64	*ulType;

	bool	SetSomething = false;
	while ( cla[Index].Name )
		{
		if(AreEqual(Parameter, cla[Index].Name, PostFixOnly))
			{
			iType = (int*) cla[Index].Ptr;
			uType = (uint*) cla[Index].Ptr;
			bType = (bool*) cla[Index].Ptr;
			sType = (char*) cla[Index].Ptr;
			dType = (double*) cla[Index].Ptr;
			llType = (int64*) cla[Index].Ptr;
			ulType = (uint64*) cla[Index].Ptr;

			if(cla[Index].SetOrDefault)
				(*cla[Index].SetOrDefault) = true;
			switch(cla[Index].theType)
				{
				case	typeInteger:
					if((Value[1]=='x')||(Value[1]=='X'))
						(*iType) = (int)strtol(Value, NULL, 16);
					else
						(*iType) = (int)strtol(Value, NULL, 10);
					break;
				case	typeuInteger:
					if((Value[1]=='x')||(Value[1]=='X'))
						(*uType) = (uint)strtoul(Value, NULL, 16);
					else
						(*uType) = (uint)strtoul(Value, NULL, 10);
					break;
				case	typeBool:
					if(AreEqual(Value, "false"))
						(*bType) = false;
					else
					if(AreEqual(Value, "true"))
						(*bType) = true;
					else
					if(AreEqual(Value, "no"))
						(*bType) = false;
					else
					if(AreEqual(Value, "yes"))
						(*bType) = true;
					else
					if(Value[0] == '0')
						(*bType) = false;
					else
					if(Value[0] == '1')
						(*bType) = true;
					else
						{
						if ( OnErrorFlag < 2 )
							printf("Error decoding boolean: \"%s\"\n", Value);
						if ( OnErrorFlag < 1 )
							return(false);
						}
					break;
				case	typeString:
					strcpy(sType, Value);
					break;
				case	typeDouble:
					(*dType) = atof(Value);
					break;
				case	typeInt64:
					if((Value[1]=='x')||(Value[1]=='X'))
						(*llType) = (int64)strtoll(Value, NULL, 16);
					else
						(*llType) = (int64)strtoll(Value, NULL, 10);
					break;
				case	typeuInt64:
					if((Value[1]=='x')||(Value[1]=='X'))
						(*ulType) = (uint64)strtoull(Value, NULL, 16);
					else
						(*ulType) = (uint64)strtoull(Value, NULL, 10);
					break;
				default:
					if ( OnErrorFlag < 2 )
						printf("error: mal-formed CLA %s\n", cla[Index].Name);
					if ( OnErrorFlag < 1 )
						return(false);
				}
			SetSomething = true;
			if ( !PostFixOnly )
				return(true);
			}
		++Index;
		}
	if ( SetSomething ) 
		return ( true );
	if ( !PostFixOnly )
		return ( LocateAndSetArgument(cla, Parameter, Value, 1, OnErrorFlag) );
	else
		{
		if ( OnErrorFlag < 2 )
			printf("Parameter %s is not found try -h\n", Parameter);
		if ( OnErrorFlag < 1 )
			return(false);
		}
	return ( true );
	}

static	int
SourceInputFile(
	CommandLineArgument	*cla,
	char	*filename,
	int		OnErrorFlag)
	{
	FILE	*fp = fopen(filename, "r");
	char	s[1024];

	if(!fp)
		{
		if ( OnErrorFlag < 2 )
			printf("Could not open source input: %s\n", filename);
		if ( OnErrorFlag < 1 )
			return(1);
		return(0);
		}
	fgets(s, 1024, fp);
	while(!feof(fp))
		{
		if(s[0]=='#')
			{
			fgets(s, 1024, fp);
			continue;
			}
		while(s[0])
			{
			if((s[0]==' ')||(s[0]=='\t')||(s[0]=='-'))
				{
				memcpy(&s[0], &s[1], strlen(s));
				continue;
				}
			break;
			}
		if(strlen(s)<=1)
			{
			fgets(s, 1024, fp);
			continue;
			}
		s[strlen(s)-1] = '\0';

		char	*field = strchr(s, ':');
		if(!field)
			{
			if ( OnErrorFlag < 2 )
				printf("Malformed argument: %s\n", s);
			if ( OnErrorFlag < 1 )
				{
				fclose(fp);
				return(-1);
				}
			fgets(s, 1024, fp);
			continue;
			}
		(*field) = '\0';
		++field;
		if(!(*field))
			{
			if ( OnErrorFlag < 2 )
				printf("Malformed argument: %s\n", s);
			if ( OnErrorFlag < 1 )
				{
				fclose(fp);
				return(-1);
				}
			fgets(s, 1024, fp);
			continue;
			}
		int	Index =0;
		while(s[Index])
			{
			if((s[Index]==' ')||(s[Index]=='\t'))
				{
				s[Index] = '\0';
				break;
				}
			++Index;
			}
		while(field[0])
			{
			if((field[0]==' ')||(field[0]=='\t'))
				{
				memcpy(&field[0], &field[1], strlen(field));
				continue;
				}
			break;
			}
		if(!strcmp(s, "Source"))
			{
			SourceInputFile(cla, field, OnErrorFlag);
			fgets(s, 1024, fp);
			continue;
			}
		if(!LocateAndSetArgument(cla, s, field, 0, OnErrorFlag))
			{
			if ( OnErrorFlag < 2 )
				printf("Notfound/Malformed argument: %s:%s\n", s, field);
			if ( OnErrorFlag < 1 )
				{
				fclose(fp);
				return(-1);
				}
			}
		
		fgets(s, 1024, fp);
		}
	fclose(fp);
	return(0);
	}

int
ParseCommandLine(
	CommandLineArgument	*cla,
	int					argc,
	char				*argv[],
	int					OnErrorFlag)
	{
	int		vargc = 1;
	char	s[1024];
	while (vargc < argc)
		{
		if((argv[vargc][0]=='-')||(argv[vargc][1]=='/'))
			{
			if(AreEqual(&argv[vargc][1], "h"))
				{
				PrintHelp(cla);
				exit(1);
				return(-1);
				}
			strcpy(s, &argv[vargc][1]);

			char	*field = strchr(s, ':');
			if(!field)
				{
				if ( OnErrorFlag < 2 )
					printf("Malformed argument: %s\n", argv[vargc]);
				if ( OnErrorFlag < 1 )
					{
					exit(2);
					return(-1);
					}
				++vargc;
				continue;
				}
			(*field) = '\0';
			++field;
			if(!(*field))
				{
				if ( OnErrorFlag < 2 )
					printf("Malformed argument: %s\n", argv[vargc]);
				if ( OnErrorFlag < 1 )
					{
					exit(2);
					return(-1);
					}
				++vargc;
				continue;
				}
			if(!strcmp(s, "Source"))
				{
				SourceInputFile(cla, field, OnErrorFlag);
				++vargc;
				continue;
				}
			if(!LocateAndSetArgument(cla, s, field, 0, OnErrorFlag))
				{
				if ( OnErrorFlag < 2 )
					printf("Notfound/Malformed argument: %s\n", argv[vargc]);
				if ( OnErrorFlag < 1 )
					{
					exit(2);
					return(-1);
					}
				}
			++vargc;
			}
		else
			break;
			
		}
	return(vargc);
	}

int
ParseString(
	CommandLineArgument	*cla,
	char				*theString,
	int					OnErrorFlag)
	{
	// hack the string up into new lines
	char	TempString[1024];

	char	*s1 = theString;
	char	*s;
	
	while(true)
		{
		int	sIndex = 0;	
		if(!(*s1))	return(1);
		while((*s1)!='\n')
			{
			if(!(*s1))	break;
			TempString[sIndex] = (*s1);

			++s1;
			++sIndex;
			}
		if(*s1)	++s1;
		TempString[sIndex] = '\n';
		++sIndex;
		TempString[sIndex] = '\0';

		s = &TempString[0];


		if(s[0]=='#')
			{
			continue;
			}
		while(s[0])
			{
			if((s[0]==' ')||(s[0]=='\t')||(s[0]=='-'))
				{
				memcpy(&s[0], &s[1], strlen(s));
				continue;
				}
			break;
			}
		if(strlen(s)<=1)
			{
			continue;
			}
		s[strlen(s)-1] = '\0';

		char	*field = strchr(s, ':');
		if(!field)
			{
			if ( OnErrorFlag < 2 )
				printf("Malformed argument: %s\n", s);
			if ( OnErrorFlag < 1 )
				return(-1);
			continue;
			}
		(*field) = '\0';
		++field;
		if(!(*field))
			{
			if ( OnErrorFlag < 2 )
				printf("Malformed argument: %s\n", s);
			if ( OnErrorFlag < 1 )
				return(-1);
			continue;
			}
		int	Index =0;
		while(s[Index])
			{
			if((s[Index]==' ')||(s[Index]=='\t'))
				{
				s[Index] = '\0';
				break;
				}
			++Index;
			}
		while(field[0])
			{
			if((field[0]==' ')||(field[0]=='\t'))
				{
				memcpy(&field[0], &field[1], strlen(field));
				continue;
				}
			break;
			}
		if(!strcmp(s, "Source"))
			{
			SourceInputFile(cla, field, OnErrorFlag);
			continue;
			}
		if(!LocateAndSetArgument(cla, s, field, 0, OnErrorFlag))
			{
			if ( OnErrorFlag < 2 )
				printf("Notfound/Malformed argument: %s:%s\n", s, field);
			if ( OnErrorFlag < 1 )
				return(-1);
			continue;
			}
		
		}
	return(true);
	}

bool
DumpParameters(
	CommandLineArgument		*cla,
	FILE					*fp)
	{
	int		Index = 0;
	char	s[128];

	int		*iType;
	uint	*uType;
	bool	*bType;
	char	*sType;
	double	*dType;
	int64	*llType;
	uint64	*ulType;

	while ( cla[Index].Name )
		{
		sprintf(s, "-%s", cla[Index].Name);
		fprintf(fp, "  %-80s:", s);
		iType = (int*) cla[Index].Ptr;
		uType = (uint*) cla[Index].Ptr;
		bType = (bool*) cla[Index].Ptr;
		sType = (char*) cla[Index].Ptr;
		dType = (double*) cla[Index].Ptr;
		llType = (int64 *) cla[Index].Ptr;
		ulType = (uint64 *) cla[Index].Ptr;

		switch(cla[Index].theType)
			{
			case	typeInteger:	fprintf(fp, "%d", *iType);break;
			case	typeuInteger:fprintf(fp, "%u", *uType);break;
			case	typeBool:
				if(*bType)
					fprintf(fp, "true");
				else
					fprintf(fp, "false");
				break;
			case	typeString: fprintf(fp, "%s", sType); break;
			case	typeDouble: fprintf(fp, "%f", (double) *dType);break;
			case	typeInt64: fprintf(fp, "" format_dec_int64 "", (int64) *llType);break;
			case	typeuInt64: fprintf(fp, "" format_dec_uint64 "", (uint64) *ulType);break;
			default:
				printf("error");
			}
		fprintf(fp, "\n");
		++Index;
		}
	return(true);
	}

void
Banner(FILE	*fp)
	{
	fprintf(fp,
"TrivialScalar Alpha AXP Processor Architecture Application Simulator                        v1.0\n"
"by Steven Swanson and Mark Oskin                    (c) Dept. of Computer Science and Engineering\n"
"swanson@cs.washington.edu, oskin@cs.washington.edu      University of Washington             2002\n"
"\n"
);
	}

