#	include	"Standard.h"

template	<class	T>
SetAssociativeMemory<T>	*
SetAssociativeMemory<T>	::	Create(
	uint	LineSize,
	uint	BankSize,
	uint	Assoc,
	ReplacementPolicy	Policy)
	{
	SetAssociativeMemory	*theCache;
	uint		Index1, Index2;

	theCache = new SetAssociativeMemory;
	theCache->NumberOfCacheBanks = Assoc;
	theCache->Hits = 0;
	theCache->Misses = 0;
	theCache->WriteBacks = 0;
	theCache->theReplacementPolicy = Policy;
	theCache->CacheBanks = (Bank<T>*)new Bank<T>[Assoc];
	theCache->LineSize = LineSize;
	theCache->LastBankAccess = 0;
	Index1 = 1;
	Index2 = 0;
	while(Index1 < LineSize)
		{
		++Index2;
		Index1 = Index1 << 1;
		}
	theCache->ShiftAmount = Index2;
	Index1 = 1;
	Index2 = 0;
	while(Index1 < BankSize)
		{
		Index2 = (Index2 << 1)|1;
		Index1 = Index1 << 1;
		}
	theCache->EntryMask = Index2;	
	Index1 = 0;
	while(Index1 < Assoc)
		{
		theCache->CacheBanks[Index1].NumberOfCacheLines = BankSize;
		theCache->CacheBanks[Index1].CacheLines =
			(Line<T>*) new Line<T>[BankSize];
		Index2 = 0;
		while(Index2 < BankSize)
			{
			theCache->CacheBanks[Index1].CacheLines[Index2].RealAddress = 0;
			theCache->CacheBanks[Index1].CacheLines[Index2].Dirty = 0;
			theCache->CacheBanks[Index1].CacheLines[Index2].Valid = 0;
			theCache->CacheBanks[Index1].CacheLines[Index2].PolicyField =Index1;
			theCache->CacheBanks[Index1].CacheLines[Index2].SharedField = 0;
			++Index2;
			}
		++Index1;
		}
	return(theCache);
	}

template	<class	T>
void
SetAssociativeMemory<T>	::	UpdatePolicy(
	int				Bank,
	uint			Entry)
	{
	int		OldEntry;
	uint		Index1;

	OldEntry = CacheBanks[Bank].CacheLines[Entry].PolicyField;

	Index1 = 0;
	while ( Index1 < NumberOfCacheBanks)
		{
		if(Index1 == (uint)Bank)
			CacheBanks[Index1].CacheLines[Entry].PolicyField =
				NumberOfCacheBanks-1;
		else
		if(CacheBanks[Index1].CacheLines[Entry].PolicyField > OldEntry)
			--CacheBanks[Index1].CacheLines[Entry].PolicyField;
		++Index1;
		}
	}


template	<class	T>
Line<T>	*
SetAssociativeMemory<T>	::	Locate(
	uint64			aAddress,
	uint			*HitOrMiss,
	uint			*ForcedWriteBack,
	Line<T>			*WrittenCacheLine)
	{
	unsigned	int	Entry;
	uint64		theLine;
	uint			Index1;
	int				Invalid = -1, Lowest = 0;

	Entry = (aAddress >> ShiftAmount)&EntryMask;
	theLine = (aAddress >> ShiftAmount);
	(*ForcedWriteBack) = 0;

	// LastBankAccess is an optimization.  Caches exploit
	// locality to improve performance.  Turns out we can
	// do the same thing with cache simulators.
	if(CacheBanks[LastBankAccess].
		CacheLines[Entry].RealAddress == theLine)
	if(CacheBanks[LastBankAccess].CacheLines[Entry].Valid)
		{
		Index1 = LastBankAccess;
		if((uint)CacheBanks[Index1].CacheLines[Entry].PolicyField !=
			(uint)(NumberOfCacheBanks-1))
			UpdatePolicy(Index1, Entry);
		++Hits;
		(*HitOrMiss) = 1;
		return(&CacheBanks[Index1].CacheLines[Entry]);
		}
	
	Index1 = 0;
	while(Index1 < NumberOfCacheBanks)
		{
		if(CacheBanks[Index1].CacheLines[Entry].RealAddress == theLine)
		if(CacheBanks[Index1].CacheLines[Entry].Valid)
			{
			if((uint)CacheBanks[Index1].CacheLines[Entry].PolicyField !=
				(uint)(NumberOfCacheBanks-1))
				UpdatePolicy(Index1, Entry);
			++Hits;
			(*HitOrMiss) = 1;
			LastBankAccess = Index1;
			return(&CacheBanks[Index1].CacheLines[Entry]);
			}
		if(!CacheBanks[Index1].CacheLines[Entry].Valid)
			Invalid = Index1;
		if(!CacheBanks[Index1].CacheLines[Entry].PolicyField)
			Lowest = Index1;
		++Index1;
		}
	// Not in the cache!	
	++Misses;
	(*HitOrMiss) = 0;

	if(Invalid>=0)
		Index1 = Invalid;
	else
	if(theReplacementPolicy == LRU)
		Index1 = Lowest;
	else
	if(theReplacementPolicy == Random)
		Index1 = rand()%NumberOfCacheBanks;

	if(CacheBanks[Index1].CacheLines[Entry].Dirty)
		{
		(*ForcedWriteBack) = 1;
		memcpy(WrittenCacheLine, &CacheBanks[Index1].CacheLines[Entry],
			sizeof(Line<T>));
		++WriteBacks;
		}

	CacheBanks[Index1].CacheLines[Entry].Dirty = 0;
	CacheBanks[Index1].CacheLines[Entry].Valid = 1;
	CacheBanks[Index1].CacheLines[Entry].RealAddress = theLine;
	if((uint)CacheBanks[Index1].CacheLines[Entry].PolicyField !=
		(uint)(NumberOfCacheBanks-1))
		UpdatePolicy(Index1, Entry);
	LastBankAccess = Index1;
	return(&CacheBanks[Index1].CacheLines[Entry]);
	}

template	<class	T>
Line<T>	*
SetAssociativeMemory<T>	::	IsInCache(
	uint64	aAddress)
	{
	unsigned	int	Entry;
	unsigned	int	theLine;
	uint		Index1;

	Entry = (aAddress >> ShiftAmount)&EntryMask;
	theLine = (aAddress >> ShiftAmount);

	if(CacheBanks[LastBankAccess].
		CacheLines[Entry].RealAddress == theLine)
	if(CacheBanks[LastBankAccess].CacheLines[Entry].Valid)
		{
		Index1 = LastBankAccess;
		return(&CacheBanks[Index1].CacheLines[Entry]);
		}
	
	Index1 = 0;
	while(Index1 < NumberOfCacheBanks)
		{
		if(CacheBanks[Index1].CacheLines[Entry].RealAddress == theLine)
		if(CacheBanks[Index1].CacheLines[Entry].Valid)
			{
			LastBankAccess = Index1;
			return(&CacheBanks[Index1].CacheLines[Entry]);
			}
		++Index1;
		}
	return(NULL);
	}

// unfortunately, you need to add specific instantiations of template classes
// that are used.
template	class	SetAssociativeMemory<int>;
template	class	SetAssociativeMemory<uint64>;


